package utf8;
use strict;
use warnings;
use SelfLoader;

sub DEBUG () { 0 }

sub DESTROY {}

my %Cache;

our (%PropertyAlias, %PA_reverse, %PropValueAlias, %PVA_reverse, %PVA_abbr_map);

sub croak { require Carp; Carp::croak(@_) }

##
## "SWASH" == "SWATCH HASH". A "swatch" is a swatch of the Unicode landscape.
## It's a data structure that encodes a set of Unicode characters.
##

sub SWASHNEW {
    my ($class, $type, $list, $minbits, $none) = @_;
    local $^D = 0 if $^D;

    print STDERR "SWASHNEW @_\n" if DEBUG;

    ##
    ## Get the list of codepoints for the type.
    ## Called from utf8.c
    ##
    ## Given a $type, our goal is to fill $list with the set of codepoint
    ## ranges.
    ##
    ## To make the parsing of $type clear, this code takes the a rather
    ## unorthodox approach of last'ing out of the block once we have the
    ## info we need. Were this to be a subroutine, the 'last' would just
    ## be a 'return'.
    ##
    my $file; ## file to load data from, and also part of the %Cache key.
    my $ListSorted = 0;

    if ($type)
    {
        $type =~ s/^\s+//;
        $type =~ s/\s+$//;

        print "type = $type\n" if DEBUG;

      GETFILE:
        {
	    ##
	    ## It could be a user-defined property.
	    ##

	    my $caller1 = $type =~ s/(.+)::// ? $1 : caller(1);

	    if (defined $caller1 && $type =~ /^(?:\w+)$/) {
		my $prop = "${caller1}::$type";
		if (exists &{$prop}) {
		    no strict 'refs';
		    
		    $list = &{$prop};
		    last GETFILE;
		}
	    }

            my $wasIs;

            ($wasIs = $type =~ s/^Is(?:\s+|[-_])?//i)
              or
            $type =~ s/^(?:(?:General(?:\s+|_)?)?Category|gc)\s*[:=]\s*//i
              or
            $type =~ s/^(?:Script|sc)\s*[:=]\s*//i
              or
            $type =~ s/^Block\s*[:=]\s*/In/i;


	    ##
	    ## See if it's in some enumeration.
	    ##
	    require "unicore/PVA.pl";
	    if ($type =~ /^([\w\s]+)[:=]\s*(.*)/) {
		my ($enum, $val) = (lc $1, lc $2);
		$enum =~ tr/ _-//d;
		$val =~ tr/ _-//d;

		my $pa = $PropertyAlias{$enum} ? $enum : $PA_reverse{$enum};
		my $f = $PropValueAlias{$pa}{$val} ? $val : $PVA_reverse{$pa}{lc $val};

		if ($pa and $f) {
		    $pa = "gc_sc" if $pa eq "gc" or $pa eq "sc";
		    $file = "unicore/lib/$pa/$PVA_abbr_map{$pa}{lc $f}.pl";
		    last GETFILE;
		}
	    }
	    else {
		my $t = lc $type;
		$t =~ tr/ _-//d;

		if ($PropValueAlias{gc}{$t} or $PropValueAlias{sc}{$t}) {
		    $file = "unicore/lib/gc_sc/$PVA_abbr_map{gc_sc}{$t}.pl";
		    last GETFILE;
		}
	    }

            ##
            ## See if it's in the direct mapping table.
            ##
            require "unicore/Exact.pl";
            if (my $base = $utf8::Exact{$type}) {
                $file = "unicore/lib/gc_sc/$base.pl";
                last GETFILE;
            }

            ##
            ## If not there exactly, try the canonical form. The canonical
            ## form is lowercased, with any separators (\s+|[-_]) removed.
            ##
            my $canonical = lc $type;
            $canonical =~ s/(?<=[a-z\d])(?:\s+|[-_])(?=[a-z\d])//g;
            print "canonical = $canonical\n" if DEBUG;

            require "unicore/Canonical.pl";
            if (my $base = ($utf8::Canonical{$canonical} || $utf8::Canonical{ lc $utf8::PropertyAlias{$canonical} })) {
                $file = "unicore/lib/gc_sc/$base.pl";
                last GETFILE;
            }

	    ##
	    ## See if it's a user-level "To".
	    ##

	    my $caller0 = caller(0);

	    if (defined $caller0 && $type =~ /^To(?:\w+)$/) {
		my $map = $caller0 . "::" . $type;

		if (exists &{$map}) {
		    no strict 'refs';
		    
		    $list = &{$map};
		    last GETFILE;
		}
	    }

            ##
            ## Last attempt -- see if it's a standard "To" name
	    ## (e.g. "ToLower")  ToTitle is used by ucfirst().
	    ## The user-level way to access ToDigit() and ToFold()
	    ## is to use Unicode::UCD.
            ##
            if ($type =~ /^To(Digit|Fold|Lower|Title|Upper)$/)
            {
                $file = "unicore/To/$1.pl";
                ## would like to test to see if $file actually exists....
                last GETFILE;
            }

            ##
            ## If we reach this line, it's because we couldn't figure
            ## out what to do with $type. Ouch.
            ##

            return $type;
        }

	if (defined $file) {
	    print "found it (file='$file')\n" if DEBUG;

	    ##
	    ## If we reach here, it was due to a 'last GETFILE' above
	    ## (exception: user-defined properties and mappings), so we
	    ## have a filename, so now we load it if we haven't already.
	    ## If we have, return the cached results. The cache key is the
	    ## file to load.
	    ##
	    if ($Cache{$file} and ref($Cache{$file}) eq $class)
	    {
		print "Returning cached '$file' for \\p{$type}\n" if DEBUG;
		return $Cache{$class, $file};
	    }

	    if($file =~ m!unicore/lib/gc_sc/([^\./]+)\.pl$!) {
		my $class = "unicore::lib::selfloader::$1";
		SelfLoader->load_stubs();
		$list = $class->getstrings();
	    } else {
		$list = do $file;
	    }
	}

        $ListSorted = 1; ## we know that these lists are sorted
    }

    my $extras;
    my $bits = 0;

    my $ORIG = $list;
    if ($list) {
	my @tmp = split(/^/m, $list);
	my %seen;
	no warnings;
	$extras = join '', grep /^[^0-9a-fA-F]/, @tmp;
	$list = join '',
	    map  { $_->[1] }
	    sort { $a->[0] <=> $b->[0] }
	    map  { /^([0-9a-fA-F]+)/; [ hex($1), $_ ] }
	    grep { /^([0-9a-fA-F]+)/ and not $seen{$1}++ } @tmp; # XXX doesn't do ranges right
    }

    if ($none) {
	my $hextra = sprintf "%04x", $none + 1;
	$list =~ s/\tXXXX$/\t$hextra/mg;
    }

    if ($minbits < 32) {
	my $top = 0;
	while ($list =~ /^([0-9a-fA-F]+)(?:[\t]([0-9a-fA-F]+)?)(?:[ \t]([0-9a-fA-F]+))?/mg) {
	    my $min = hex $1;
	    my $max = defined $2 ? hex $2 : $min;
	    my $val = defined $3 ? hex $3 : 0;
	    $val += $max - $min if defined $3;
	    $top = $val if $val > $top;
	}
	$bits =
	    $top > 0xffff ? 32 :
	    $top > 0xff ? 16 :
	    $top > 1 ? 8 : 1
    }
    $bits = $minbits if $bits < $minbits;

    my @extras;
    for my $x ($extras) {
	pos $x = 0;
	while ($x =~ /^([^0-9a-fA-F\n])(.*)/mg) {
	    my $char = $1;
	    my $name = $2;
	    print STDERR "$1 => $2\n" if DEBUG;
	    if ($char =~ /[-+!&]/) {
		my ($c,$t) = split(/::/, $name, 2);	# bogus use of ::, really
		my $subobj;
		if ($c eq 'utf8') {
		    $subobj = utf8->SWASHNEW($t, "", 0, 0, 0);
		}
		elsif (exists &$name) {
		    $subobj = utf8->SWASHNEW($name, "", 0, 0, 0);
		}
		elsif ($c =~ /^([0-9a-fA-F]+)/) {
		    $subobj = utf8->SWASHNEW("", $c, 0, 0, 0);
		}
		return $subobj unless ref $subobj;
		push @extras, $name => $subobj;
		$bits = $subobj->{BITS} if $bits < $subobj->{BITS};
	    }
	}
    }

    print STDERR "CLASS = $class, TYPE => $type, BITS => $bits, NONE => $none\nEXTRAS =>\n$extras\nLIST =>\n$list\n" if DEBUG;

    my $SWASH = bless {
	TYPE => $type,
	BITS => $bits,
	EXTRAS => $extras,
	LIST => $list,
	NONE => $none,
	@extras,
    } => $class;

    if ($file) {
        $Cache{$class, $file} = $SWASH;
    }

    return $SWASH;
}

# NOTE: utf8.c:swash_init() assumes entries are never modified once generated.

sub SWASHGET {
    # See utf8.c:Perl_swash_fetch for problems with this interface.
    my ($self, $start, $len) = @_;
    local $^D = 0 if $^D;
    my $type = $self->{TYPE};
    my $bits = $self->{BITS};
    my $none = $self->{NONE};
    print STDERR "SWASHGET @_ [$type/$bits/$none]\n" if DEBUG;
    my $end = $start + $len;
    my $swatch = "";
    my $key;
    vec($swatch, $len - 1, $bits) = 0;	# Extend to correct length.
    if ($none) {
	for $key (0 .. $len - 1) { vec($swatch, $key, $bits) = $none }
    }

    for ($self->{LIST}) {
	pos $_ = 0;
	if ($bits > 1) {
	  LINE:
	    while (/^([0-9a-fA-F]+)(?:[ \t]([0-9a-fA-F]+)?)?(?:[ \t]([0-9a-fA-F]+))?/mg) {
		chomp;
		my ($a, $b, $c) = ($1, $2, $3);
		croak "$type: illegal mapping '$_'"
		    if $type =~ /^To/ &&
		       !(defined $a && defined $c);
		my $min = hex $a;
		my $max = defined $b ? hex $b : $min;
		my $val = defined $c ? hex $c : 0;
		next if $max < $start;
		print "$min $max $val\n" if DEBUG;
		if ($none) {
		    if ($min < $start) {
			$val += $start - $min if $val < $none;
			$min = $start;
		    }
		    for ($key = $min; $key <= $max; $key++) {
			last LINE if $key >= $end;
			print STDERR "$key => $val\n" if DEBUG;
			vec($swatch, $key - $start, $bits) = $val;
			++$val if $val < $none;
		    }
		}
		else {
		    if ($min < $start) {
			$val += $start - $min;
			$min = $start;
		    }
		    for ($key = $min; $key <= $max; $key++, $val++) {
			last LINE if $key >= $end;
			print STDERR "$key => $val\n" if DEBUG;
			vec($swatch, $key - $start, $bits) = $val;
		    }
		}
	    }
	}
	else {
	  LINE:
	    while (/^([0-9a-fA-F]+)(?:[ \t]+([0-9a-fA-F]+))?/mg) {
		chomp;
		my $min = hex $1;
		my $max = defined $2 ? hex $2 : $min;
		next if $max < $start;
		if ($min < $start) {
		    $min = $start;
		}
		for ($key = $min; $key <= $max; $key++) {
		    last LINE if $key >= $end;
		    print STDERR "$key => 1\n" if DEBUG;
		    vec($swatch, $key - $start, 1) = 1;
		}
	    }
	}
    }
    for my $x ($self->{EXTRAS}) {
	pos $x = 0;
	while ($x =~ /^([-+!&])(.*)/mg) {
	    my $char = $1;
	    my $name = $2;
	    print STDERR "INDIRECT $1 $2\n" if DEBUG;
	    my $otherbits = $self->{$name}->{BITS};
	    croak("SWASHGET size mismatch") if $bits < $otherbits;
	    my $other = $self->{$name}->SWASHGET($start, $len);
	    if ($char eq '+') {
		if ($bits == 1 and $otherbits == 1) {
		    $swatch |= $other;
		}
		else {
		    for ($key = 0; $key < $len; $key++) {
			vec($swatch, $key, $bits) = vec($other, $key, $otherbits);
		    }
		}
	    }
	    elsif ($char eq '!') {
		if ($bits == 1 and $otherbits == 1) {
		    $swatch |= ~$other;
		}
		else {
		    for ($key = 0; $key < $len; $key++) {
			if (!vec($other, $key, $otherbits)) {
			    vec($swatch, $key, $bits) = 1;
			}
		    }
		}
	    }
	    elsif ($char eq '-') {
		if ($bits == 1 and $otherbits == 1) {
		    $swatch &= ~$other;
		}
		else {
		    for ($key = 0; $key < $len; $key++) {
			if (vec($other, $key, $otherbits)) {
			    vec($swatch, $key, $bits) = 0;
			}
		    }
		}
	    }
	    elsif ($char eq '&') {
		if ($bits == 1 and $otherbits == 1) {
		    $swatch &= $other;
		}
		else {
		    for ($key = 0; $key < $len; $key++) {
			if (!vec($other, $key, $otherbits)) {
			    vec($swatch, $key, $bits) = 0;
			}
		    }
		}
	    }
	}
    }
    if (DEBUG) {
	print STDERR "CELLS ";
	for ($key = 0; $key < $len; $key++) {
	    print STDERR vec($swatch, $key, $bits), " ";
	}
	print STDERR "\n";
    }
    $swatch;
}

1;

# lib/unicore/lib/gc_sc/* in here
__DATA__
package unicore::lib::selfloader::_CanonDC;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{_CanonDCIJ}
# 
# Meaning: (for internal casefolding use)
#
return <<'END';
0069	006A	
012F		
1E2D		
1ECB		
END
};
1;

package unicore::lib::selfloader::_CaseIgn;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{_CaseIgnorable}
# 
# Meaning: (for internal casefolding use)
#
return <<'END';
00AD		
0300	036F	
0483	0486	
0591	05B9	
05BB	05BD	
05BF		
05C1	05C2	
05C4	05C5	
05C7		
0610	0615	
064B	065E	
0670		
06D6	06DC	
06DF	06E4	
06E7	06E8	
06EA	06ED	
0711		
0730	074A	
07A6	07B0	
0901	0902	
093C		
0941	0948	
094D		
0951	0954	
0962	0963	
0981		
09BC		
09C1	09C4	
09CD		
09E2	09E3	
0A01	0A02	
0A3C		
0A41	0A42	
0A47	0A48	
0A4B	0A4D	
0A70	0A71	
0A81	0A82	
0ABC		
0AC1	0AC5	
0AC7	0AC8	
0ACD		
0AE2	0AE3	
0B01		
0B3C		
0B3F		
0B41	0B43	
0B4D		
0B56		
0B82		
0BC0		
0BCD		
0C3E	0C40	
0C46	0C48	
0C4A	0C4D	
0C55	0C56	
0CBC		
0CBF		
0CC6		
0CCC	0CCD	
0D41	0D43	
0D4D		
0DCA		
0DD2	0DD4	
0DD6		
0E31		
0E34	0E3A	
0E47	0E4E	
0EB1		
0EB4	0EB9	
0EBB	0EBC	
0EC8	0ECD	
0F18	0F19	
0F35		
0F37		
0F39		
0F71	0F7E	
0F80	0F84	
0F86	0F87	
0F90	0F97	
0F99	0FBC	
0FC6		
102D	1030	
1032		
1036	1037	
1039		
1058	1059	
135F		
1712	1714	
1732	1734	
1752	1753	
1772	1773	
17B7	17BD	
17C6		
17C9	17D3	
17DD		
180B	180D	
18A9		
1920	1922	
1927	1928	
1932		
1939	193B	
1A17	1A18	
1DC0	1DC3	
2010		
20D0	20DC	
20E1		
20E5	20EB	
302A	302F	
3099	309A	
A806		
A80B		
A825	A826	
FB1E		
FE00	FE0F	
FE20	FE23	
10A01	10A03	
10A05	10A06	
10A0C	10A0F	
10A38	10A3A	
10A3F		
1D167	1D169	
1D17B	1D182	
1D185	1D18B	
1D1AA	1D1AD	
1D242	1D244	
E0100	E01EF	
END
};
1;

package unicore::lib::selfloader::_CombAbo;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{_CombAbove}
# 
# Meaning: (for internal casefolding use)
#
return <<'END';
0300	0314	
033D	0344	
0346		
034A	034C	
0350	0352	
0357		
035B		
0363	036F	
0483	0486	
0592	0595	
0597	0599	
059C	05A1	
05A8	05A9	
05AB	05AC	
05AF		
05C4		
0610	0615	
0653	0654	
0657	065B	
065D	065E	
06D6	06DC	
06DF	06E2	
06E4		
06E7	06E8	
06EB	06EC	
0730		
0732	0733	
0735	0736	
073A		
073D		
073F	0741	
0743		
0745		
0747		
0749	074A	
0951		
0953	0954	
0F82	0F83	
0F86	0F87	
135F		
17DD		
193A		
1A17		
1DC0	1DC1	
1DC3		
20D0	20D1	
20D4	20D7	
20DB	20DC	
20E1		
20E7		
20E9		
FE20	FE23	
10A0F		
10A38		
1D185	1D189	
1D1AA	1D1AD	
1D242	1D244	
END
};
1;

package unicore::lib::selfloader::AHex;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'ASCII_Hex_Digit'
#
return <<'END';
0030	0039	ASCII_Hex_Digit
0041	0046	ASCII_Hex_Digit
0061	0066	ASCII_Hex_Digit
END
};
1;

package unicore::lib::selfloader::Alnum;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Alnum}
# 
# Meaning: [[:Alnum:]]
#
return <<'END';
0030	0039	
0041	005A	
0061	007A	
00AA		
00B5		
00BA		
00C0	00D6	
00D8	00F6	
00F8	0241	
0250	02C1	
02C6	02D1	
02E0	02E4	
02EE		
0300	036F	
037A		
0386		
0388	038A	
038C		
038E	03A1	
03A3	03CE	
03D0	03F5	
03F7	0481	
0483	0486	
0488	04CE	
04D0	04F9	
0500	050F	
0531	0556	
0559		
0561	0587	
0591	05B9	
05BB	05BD	
05BF		
05C1	05C2	
05C4	05C5	
05C7		
05D0	05EA	
05F0	05F2	
0610	0615	
0621	063A	
0640	065E	
0660	0669	
066E	06D3	
06D5	06DC	
06DE	06E8	
06EA	06FC	
06FF		
0710	074A	
074D	076D	
0780	07B1	
0901	0939	
093C	094D	
0950	0954	
0958	0963	
0966	096F	
097D		
0981	0983	
0985	098C	
098F	0990	
0993	09A8	
09AA	09B0	
09B2		
09B6	09B9	
09BC	09C4	
09C7	09C8	
09CB	09CE	
09D7		
09DC	09DD	
09DF	09E3	
09E6	09F1	
0A01	0A03	
0A05	0A0A	
0A0F	0A10	
0A13	0A28	
0A2A	0A30	
0A32	0A33	
0A35	0A36	
0A38	0A39	
0A3C		
0A3E	0A42	
0A47	0A48	
0A4B	0A4D	
0A59	0A5C	
0A5E		
0A66	0A74	
0A81	0A83	
0A85	0A8D	
0A8F	0A91	
0A93	0AA8	
0AAA	0AB0	
0AB2	0AB3	
0AB5	0AB9	
0ABC	0AC5	
0AC7	0AC9	
0ACB	0ACD	
0AD0		
0AE0	0AE3	
0AE6	0AEF	
0B01	0B03	
0B05	0B0C	
0B0F	0B10	
0B13	0B28	
0B2A	0B30	
0B32	0B33	
0B35	0B39	
0B3C	0B43	
0B47	0B48	
0B4B	0B4D	
0B56	0B57	
0B5C	0B5D	
0B5F	0B61	
0B66	0B6F	
0B71		
0B82	0B83	
0B85	0B8A	
0B8E	0B90	
0B92	0B95	
0B99	0B9A	
0B9C		
0B9E	0B9F	
0BA3	0BA4	
0BA8	0BAA	
0BAE	0BB9	
0BBE	0BC2	
0BC6	0BC8	
0BCA	0BCD	
0BD7		
0BE6	0BEF	
0C01	0C03	
0C05	0C0C	
0C0E	0C10	
0C12	0C28	
0C2A	0C33	
0C35	0C39	
0C3E	0C44	
0C46	0C48	
0C4A	0C4D	
0C55	0C56	
0C60	0C61	
0C66	0C6F	
0C82	0C83	
0C85	0C8C	
0C8E	0C90	
0C92	0CA8	
0CAA	0CB3	
0CB5	0CB9	
0CBC	0CC4	
0CC6	0CC8	
0CCA	0CCD	
0CD5	0CD6	
0CDE		
0CE0	0CE1	
0CE6	0CEF	
0D02	0D03	
0D05	0D0C	
0D0E	0D10	
0D12	0D28	
0D2A	0D39	
0D3E	0D43	
0D46	0D48	
0D4A	0D4D	
0D57		
0D60	0D61	
0D66	0D6F	
0D82	0D83	
0D85	0D96	
0D9A	0DB1	
0DB3	0DBB	
0DBD		
0DC0	0DC6	
0DCA		
0DCF	0DD4	
0DD6		
0DD8	0DDF	
0DF2	0DF3	
0E01	0E3A	
0E40	0E4E	
0E50	0E59	
0E81	0E82	
0E84		
0E87	0E88	
0E8A		
0E8D		
0E94	0E97	
0E99	0E9F	
0EA1	0EA3	
0EA5		
0EA7		
0EAA	0EAB	
0EAD	0EB9	
0EBB	0EBD	
0EC0	0EC4	
0EC6		
0EC8	0ECD	
0ED0	0ED9	
0EDC	0EDD	
0F00		
0F18	0F19	
0F20	0F29	
0F35		
0F37		
0F39		
0F3E	0F47	
0F49	0F6A	
0F71	0F84	
0F86	0F8B	
0F90	0F97	
0F99	0FBC	
0FC6		
1000	1021	
1023	1027	
1029	102A	
102C	1032	
1036	1039	
1040	1049	
1050	1059	
10A0	10C5	
10D0	10FA	
10FC		
1100	1159	
115F	11A2	
11A8	11F9	
1200	1248	
124A	124D	
1250	1256	
1258		
125A	125D	
1260	1288	
128A	128D	
1290	12B0	
12B2	12B5	
12B8	12BE	
12C0		
12C2	12C5	
12C8	12D6	
12D8	1310	
1312	1315	
1318	135A	
135F		
1380	138F	
13A0	13F4	
1401	166C	
166F	1676	
1681	169A	
16A0	16EA	
1700	170C	
170E	1714	
1720	1734	
1740	1753	
1760	176C	
176E	1770	
1772	1773	
1780	17B3	
17B6	17D3	
17D7		
17DC	17DD	
17E0	17E9	
180B	180D	
1810	1819	
1820	1877	
1880	18A9	
1900	191C	
1920	192B	
1930	193B	
1946	196D	
1970	1974	
1980	19A9	
19B0	19C9	
19D0	19D9	
1A00	1A1B	
1D00	1DC3	
1E00	1E9B	
1EA0	1EF9	
1F00	1F15	
1F18	1F1D	
1F20	1F45	
1F48	1F4D	
1F50	1F57	
1F59		
1F5B		
1F5D		
1F5F	1F7D	
1F80	1FB4	
1FB6	1FBC	
1FBE		
1FC2	1FC4	
1FC6	1FCC	
1FD0	1FD3	
1FD6	1FDB	
1FE0	1FEC	
1FF2	1FF4	
1FF6	1FFC	
2071		
207F		
2090	2094	
20D0	20EB	
2102		
2107		
210A	2113	
2115		
2119	211D	
2124		
2126		
2128		
212A	212D	
212F	2131	
2133	2139	
213C	213F	
2145	2149	
2C00	2C2E	
2C30	2C5E	
2C80	2CE4	
2D00	2D25	
2D30	2D65	
2D6F		
2D80	2D96	
2DA0	2DA6	
2DA8	2DAE	
2DB0	2DB6	
2DB8	2DBE	
2DC0	2DC6	
2DC8	2DCE	
2DD0	2DD6	
2DD8	2DDE	
3005	3006	
302A	302F	
3031	3035	
303B	303C	
3041	3096	
3099	309A	
309D	309F	
30A1	30FA	
30FC	30FF	
3105	312C	
3131	318E	
31A0	31B7	
31F0	31FF	
3400	4DB5	
4E00	9FBB	
A000	A48C	
A800	A827	
AC00	D7A3	
F900	FA2D	
FA30	FA6A	
FA70	FAD9	
FB00	FB06	
FB13	FB17	
FB1D	FB28	
FB2A	FB36	
FB38	FB3C	
FB3E		
FB40	FB41	
FB43	FB44	
FB46	FBB1	
FBD3	FD3D	
FD50	FD8F	
FD92	FDC7	
FDF0	FDFB	
FE00	FE0F	
FE20	FE23	
FE70	FE74	
FE76	FEFC	
FF10	FF19	
FF21	FF3A	
FF41	FF5A	
FF66	FFBE	
FFC2	FFC7	
FFCA	FFCF	
FFD2	FFD7	
FFDA	FFDC	
10000	1000B	
1000D	10026	
10028	1003A	
1003C	1003D	
1003F	1004D	
10050	1005D	
10080	100FA	
10300	1031E	
10330	10349	
10380	1039D	
103A0	103C3	
103C8	103CF	
10400	1049D	
104A0	104A9	
10800	10805	
10808		
1080A	10835	
10837	10838	
1083C		
1083F		
10A00	10A03	
10A05	10A06	
10A0C	10A13	
10A15	10A17	
10A19	10A33	
10A38	10A3A	
10A3F		
1D165	1D169	
1D16D	1D172	
1D17B	1D182	
1D185	1D18B	
1D1AA	1D1AD	
1D242	1D244	
1D400	1D454	
1D456	1D49C	
1D49E	1D49F	
1D4A2		
1D4A5	1D4A6	
1D4A9	1D4AC	
1D4AE	1D4B9	
1D4BB		
1D4BD	1D4C3	
1D4C5	1D505	
1D507	1D50A	
1D50D	1D514	
1D516	1D51C	
1D51E	1D539	
1D53B	1D53E	
1D540	1D544	
1D546		
1D54A	1D550	
1D552	1D6A5	
1D6A8	1D6C0	
1D6C2	1D6DA	
1D6DC	1D6FA	
1D6FC	1D714	
1D716	1D734	
1D736	1D74E	
1D750	1D76E	
1D770	1D788	
1D78A	1D7A8	
1D7AA	1D7C2	
1D7C4	1D7C9	
1D7CE	1D7FF	
20000	2A6D6	
2F800	2FA1D	
E0100	E01EF	
END
};
1;

package unicore::lib::selfloader::Alpha;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Alpha}
# 
# Meaning: [[:Alpha:]]
#
return <<'END';
0041	005A	
0061	007A	
00AA		
00B5		
00BA		
00C0	00D6	
00D8	00F6	
00F8	0241	
0250	02C1	
02C6	02D1	
02E0	02E4	
02EE		
0300	036F	
037A		
0386		
0388	038A	
038C		
038E	03A1	
03A3	03CE	
03D0	03F5	
03F7	0481	
0483	0486	
0488	04CE	
04D0	04F9	
0500	050F	
0531	0556	
0559		
0561	0587	
0591	05B9	
05BB	05BD	
05BF		
05C1	05C2	
05C4	05C5	
05C7		
05D0	05EA	
05F0	05F2	
0610	0615	
0621	063A	
0640	065E	
066E	06D3	
06D5	06DC	
06DE	06E8	
06EA	06EF	
06FA	06FC	
06FF		
0710	074A	
074D	076D	
0780	07B1	
0901	0939	
093C	094D	
0950	0954	
0958	0963	
097D		
0981	0983	
0985	098C	
098F	0990	
0993	09A8	
09AA	09B0	
09B2		
09B6	09B9	
09BC	09C4	
09C7	09C8	
09CB	09CE	
09D7		
09DC	09DD	
09DF	09E3	
09F0	09F1	
0A01	0A03	
0A05	0A0A	
0A0F	0A10	
0A13	0A28	
0A2A	0A30	
0A32	0A33	
0A35	0A36	
0A38	0A39	
0A3C		
0A3E	0A42	
0A47	0A48	
0A4B	0A4D	
0A59	0A5C	
0A5E		
0A70	0A74	
0A81	0A83	
0A85	0A8D	
0A8F	0A91	
0A93	0AA8	
0AAA	0AB0	
0AB2	0AB3	
0AB5	0AB9	
0ABC	0AC5	
0AC7	0AC9	
0ACB	0ACD	
0AD0		
0AE0	0AE3	
0B01	0B03	
0B05	0B0C	
0B0F	0B10	
0B13	0B28	
0B2A	0B30	
0B32	0B33	
0B35	0B39	
0B3C	0B43	
0B47	0B48	
0B4B	0B4D	
0B56	0B57	
0B5C	0B5D	
0B5F	0B61	
0B71		
0B82	0B83	
0B85	0B8A	
0B8E	0B90	
0B92	0B95	
0B99	0B9A	
0B9C		
0B9E	0B9F	
0BA3	0BA4	
0BA8	0BAA	
0BAE	0BB9	
0BBE	0BC2	
0BC6	0BC8	
0BCA	0BCD	
0BD7		
0C01	0C03	
0C05	0C0C	
0C0E	0C10	
0C12	0C28	
0C2A	0C33	
0C35	0C39	
0C3E	0C44	
0C46	0C48	
0C4A	0C4D	
0C55	0C56	
0C60	0C61	
0C82	0C83	
0C85	0C8C	
0C8E	0C90	
0C92	0CA8	
0CAA	0CB3	
0CB5	0CB9	
0CBC	0CC4	
0CC6	0CC8	
0CCA	0CCD	
0CD5	0CD6	
0CDE		
0CE0	0CE1	
0D02	0D03	
0D05	0D0C	
0D0E	0D10	
0D12	0D28	
0D2A	0D39	
0D3E	0D43	
0D46	0D48	
0D4A	0D4D	
0D57		
0D60	0D61	
0D82	0D83	
0D85	0D96	
0D9A	0DB1	
0DB3	0DBB	
0DBD		
0DC0	0DC6	
0DCA		
0DCF	0DD4	
0DD6		
0DD8	0DDF	
0DF2	0DF3	
0E01	0E3A	
0E40	0E4E	
0E81	0E82	
0E84		
0E87	0E88	
0E8A		
0E8D		
0E94	0E97	
0E99	0E9F	
0EA1	0EA3	
0EA5		
0EA7		
0EAA	0EAB	
0EAD	0EB9	
0EBB	0EBD	
0EC0	0EC4	
0EC6		
0EC8	0ECD	
0EDC	0EDD	
0F00		
0F18	0F19	
0F35		
0F37		
0F39		
0F3E	0F47	
0F49	0F6A	
0F71	0F84	
0F86	0F8B	
0F90	0F97	
0F99	0FBC	
0FC6		
1000	1021	
1023	1027	
1029	102A	
102C	1032	
1036	1039	
1050	1059	
10A0	10C5	
10D0	10FA	
10FC		
1100	1159	
115F	11A2	
11A8	11F9	
1200	1248	
124A	124D	
1250	1256	
1258		
125A	125D	
1260	1288	
128A	128D	
1290	12B0	
12B2	12B5	
12B8	12BE	
12C0		
12C2	12C5	
12C8	12D6	
12D8	1310	
1312	1315	
1318	135A	
135F		
1380	138F	
13A0	13F4	
1401	166C	
166F	1676	
1681	169A	
16A0	16EA	
1700	170C	
170E	1714	
1720	1734	
1740	1753	
1760	176C	
176E	1770	
1772	1773	
1780	17B3	
17B6	17D3	
17D7		
17DC	17DD	
180B	180D	
1820	1877	
1880	18A9	
1900	191C	
1920	192B	
1930	193B	
1950	196D	
1970	1974	
1980	19A9	
19B0	19C9	
1A00	1A1B	
1D00	1DC3	
1E00	1E9B	
1EA0	1EF9	
1F00	1F15	
1F18	1F1D	
1F20	1F45	
1F48	1F4D	
1F50	1F57	
1F59		
1F5B		
1F5D		
1F5F	1F7D	
1F80	1FB4	
1FB6	1FBC	
1FBE		
1FC2	1FC4	
1FC6	1FCC	
1FD0	1FD3	
1FD6	1FDB	
1FE0	1FEC	
1FF2	1FF4	
1FF6	1FFC	
2071		
207F		
2090	2094	
20D0	20EB	
2102		
2107		
210A	2113	
2115		
2119	211D	
2124		
2126		
2128		
212A	212D	
212F	2131	
2133	2139	
213C	213F	
2145	2149	
2C00	2C2E	
2C30	2C5E	
2C80	2CE4	
2D00	2D25	
2D30	2D65	
2D6F		
2D80	2D96	
2DA0	2DA6	
2DA8	2DAE	
2DB0	2DB6	
2DB8	2DBE	
2DC0	2DC6	
2DC8	2DCE	
2DD0	2DD6	
2DD8	2DDE	
3005	3006	
302A	302F	
3031	3035	
303B	303C	
3041	3096	
3099	309A	
309D	309F	
30A1	30FA	
30FC	30FF	
3105	312C	
3131	318E	
31A0	31B7	
31F0	31FF	
3400	4DB5	
4E00	9FBB	
A000	A48C	
A800	A827	
AC00	D7A3	
F900	FA2D	
FA30	FA6A	
FA70	FAD9	
FB00	FB06	
FB13	FB17	
FB1D	FB28	
FB2A	FB36	
FB38	FB3C	
FB3E		
FB40	FB41	
FB43	FB44	
FB46	FBB1	
FBD3	FD3D	
FD50	FD8F	
FD92	FDC7	
FDF0	FDFB	
FE00	FE0F	
FE20	FE23	
FE70	FE74	
FE76	FEFC	
FF21	FF3A	
FF41	FF5A	
FF66	FFBE	
FFC2	FFC7	
FFCA	FFCF	
FFD2	FFD7	
FFDA	FFDC	
10000	1000B	
1000D	10026	
10028	1003A	
1003C	1003D	
1003F	1004D	
10050	1005D	
10080	100FA	
10300	1031E	
10330	10349	
10380	1039D	
103A0	103C3	
103C8	103CF	
10400	1049D	
10800	10805	
10808		
1080A	10835	
10837	10838	
1083C		
1083F		
10A00	10A03	
10A05	10A06	
10A0C	10A13	
10A15	10A17	
10A19	10A33	
10A38	10A3A	
10A3F		
1D165	1D169	
1D16D	1D172	
1D17B	1D182	
1D185	1D18B	
1D1AA	1D1AD	
1D242	1D244	
1D400	1D454	
1D456	1D49C	
1D49E	1D49F	
1D4A2		
1D4A5	1D4A6	
1D4A9	1D4AC	
1D4AE	1D4B9	
1D4BB		
1D4BD	1D4C3	
1D4C5	1D505	
1D507	1D50A	
1D50D	1D514	
1D516	1D51C	
1D51E	1D539	
1D53B	1D53E	
1D540	1D544	
1D546		
1D54A	1D550	
1D552	1D6A5	
1D6A8	1D6C0	
1D6C2	1D6DA	
1D6DC	1D6FA	
1D6FC	1D714	
1D716	1D734	
1D736	1D74E	
1D750	1D76E	
1D770	1D788	
1D78A	1D7A8	
1D7AA	1D7C2	
1D7C4	1D7C9	
20000	2A6D6	
2F800	2FA1D	
E0100	E01EF	
END
};
1;

package unicore::lib::selfloader::Alphabet;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Alphabetic} (and fuzzy permutations)
# 
# Meaning: [\p{L}\p{OtherAlphabetic}]
#
return <<'END';
0041	005A	
0061	007A	
00AA		
00B5		
00BA		
00C0	00D6	
00D8	00F6	
00F8	0241	
0250	02C1	
02C6	02D1	
02E0	02E4	
02EE		
0345		
037A		
0386		
0388	038A	
038C		
038E	03A1	
03A3	03CE	
03D0	03F5	
03F7	0481	
048A	04CE	
04D0	04F9	
0500	050F	
0531	0556	
0559		
0561	0587	
05B0	05B9	
05BB	05BD	
05BF		
05C1	05C2	
05C4	05C5	
05C7		
05D0	05EA	
05F0	05F2	
0610	0615	
0621	063A	
0640	0657	
0659	065E	
066E	06D3	
06D5	06DC	
06E1	06E8	
06ED	06EF	
06FA	06FC	
06FF		
0710	073F	
074D	076D	
0780	07B1	
0901	0939	
093D	094C	
0950		
0958	0963	
097D		
0981	0983	
0985	098C	
098F	0990	
0993	09A8	
09AA	09B0	
09B2		
09B6	09B9	
09BD	09C4	
09C7	09C8	
09CB	09CC	
09CE		
09D7		
09DC	09DD	
09DF	09E3	
09F0	09F1	
0A01	0A03	
0A05	0A0A	
0A0F	0A10	
0A13	0A28	
0A2A	0A30	
0A32	0A33	
0A35	0A36	
0A38	0A39	
0A3E	0A42	
0A47	0A48	
0A4B	0A4C	
0A59	0A5C	
0A5E		
0A70	0A74	
0A81	0A83	
0A85	0A8D	
0A8F	0A91	
0A93	0AA8	
0AAA	0AB0	
0AB2	0AB3	
0AB5	0AB9	
0ABD	0AC5	
0AC7	0AC9	
0ACB	0ACC	
0AD0		
0AE0	0AE3	
0B01	0B03	
0B05	0B0C	
0B0F	0B10	
0B13	0B28	
0B2A	0B30	
0B32	0B33	
0B35	0B39	
0B3D	0B43	
0B47	0B48	
0B4B	0B4C	
0B56	0B57	
0B5C	0B5D	
0B5F	0B61	
0B71		
0B82	0B83	
0B85	0B8A	
0B8E	0B90	
0B92	0B95	
0B99	0B9A	
0B9C		
0B9E	0B9F	
0BA3	0BA4	
0BA8	0BAA	
0BAE	0BB9	
0BBE	0BC2	
0BC6	0BC8	
0BCA	0BCC	
0BD7		
0C01	0C03	
0C05	0C0C	
0C0E	0C10	
0C12	0C28	
0C2A	0C33	
0C35	0C39	
0C3E	0C44	
0C46	0C48	
0C4A	0C4C	
0C55	0C56	
0C60	0C61	
0C82	0C83	
0C85	0C8C	
0C8E	0C90	
0C92	0CA8	
0CAA	0CB3	
0CB5	0CB9	
0CBD	0CC4	
0CC6	0CC8	
0CCA	0CCC	
0CD5	0CD6	
0CDE		
0CE0	0CE1	
0D02	0D03	
0D05	0D0C	
0D0E	0D10	
0D12	0D28	
0D2A	0D39	
0D3E	0D43	
0D46	0D48	
0D4A	0D4C	
0D57		
0D60	0D61	
0D82	0D83	
0D85	0D96	
0D9A	0DB1	
0DB3	0DBB	
0DBD		
0DC0	0DC6	
0DCF	0DD4	
0DD6		
0DD8	0DDF	
0DF2	0DF3	
0E01	0E3A	
0E40	0E46	
0E4D		
0E81	0E82	
0E84		
0E87	0E88	
0E8A		
0E8D		
0E94	0E97	
0E99	0E9F	
0EA1	0EA3	
0EA5		
0EA7		
0EAA	0EAB	
0EAD	0EB9	
0EBB	0EBD	
0EC0	0EC4	
0EC6		
0ECD		
0EDC	0EDD	
0F00		
0F40	0F47	
0F49	0F6A	
0F71	0F81	
0F88	0F8B	
0F90	0F97	
0F99	0FBC	
1000	1021	
1023	1027	
1029	102A	
102C	1032	
1036		
1038		
1050	1059	
10A0	10C5	
10D0	10FA	
10FC		
1100	1159	
115F	11A2	
11A8	11F9	
1200	1248	
124A	124D	
1250	1256	
1258		
125A	125D	
1260	1288	
128A	128D	
1290	12B0	
12B2	12B5	
12B8	12BE	
12C0		
12C2	12C5	
12C8	12D6	
12D8	1310	
1312	1315	
1318	135A	
135F		
1380	138F	
13A0	13F4	
1401	166C	
166F	1676	
1681	169A	
16A0	16EA	
1700	170C	
170E	1713	
1720	1733	
1740	1753	
1760	176C	
176E	1770	
1772	1773	
1780	17B3	
17B6	17C8	
17D7		
17DC		
1820	1877	
1880	18A9	
1900	191C	
1920	192B	
1930	1938	
1950	196D	
1970	1974	
1980	19A9	
19B0	19C9	
1A00	1A1B	
1D00	1DBF	
1E00	1E9B	
1EA0	1EF9	
1F00	1F15	
1F18	1F1D	
1F20	1F45	
1F48	1F4D	
1F50	1F57	
1F59		
1F5B		
1F5D		
1F5F	1F7D	
1F80	1FB4	
1FB6	1FBC	
1FBE		
1FC2	1FC4	
1FC6	1FCC	
1FD0	1FD3	
1FD6	1FDB	
1FE0	1FEC	
1FF2	1FF4	
1FF6	1FFC	
2071		
207F		
2090	2094	
2102		
2107		
210A	2113	
2115		
2119	211D	
2124		
2126		
2128		
212A	212D	
212F	2131	
2133	2139	
213C	213F	
2145	2149	
24B6	24E9	
2C00	2C2E	
2C30	2C5E	
2C80	2CE4	
2D00	2D25	
2D30	2D65	
2D6F		
2D80	2D96	
2DA0	2DA6	
2DA8	2DAE	
2DB0	2DB6	
2DB8	2DBE	
2DC0	2DC6	
2DC8	2DCE	
2DD0	2DD6	
2DD8	2DDE	
3005	3006	
3031	3035	
303B	303C	
3041	3096	
309D	309F	
30A1	30FA	
30FC	30FF	
3105	312C	
3131	318E	
31A0	31B7	
31F0	31FF	
3400	4DB5	
4E00	9FBB	
A000	A48C	
A800	A801	
A803	A805	
A807	A80A	
A80C	A827	
AC00	D7A3	
F900	FA2D	
FA30	FA6A	
FA70	FAD9	
FB00	FB06	
FB13	FB17	
FB1D	FB28	
FB2A	FB36	
FB38	FB3C	
FB3E		
FB40	FB41	
FB43	FB44	
FB46	FBB1	
FBD3	FD3D	
FD50	FD8F	
FD92	FDC7	
FDF0	FDFB	
FE70	FE74	
FE76	FEFC	
FF21	FF3A	
FF41	FF5A	
FF66	FFBE	
FFC2	FFC7	
FFCA	FFCF	
FFD2	FFD7	
FFDA	FFDC	
10000	1000B	
1000D	10026	
10028	1003A	
1003C	1003D	
1003F	1004D	
10050	1005D	
10080	100FA	
10300	1031E	
10330	10349	
10380	1039D	
103A0	103C3	
103C8	103CF	
10400	1049D	
10800	10805	
10808		
1080A	10835	
10837	10838	
1083C		
1083F		
10A00	10A03	
10A05	10A06	
10A0C	10A13	
10A15	10A17	
10A19	10A33	
1D400	1D454	
1D456	1D49C	
1D49E	1D49F	
1D4A2		
1D4A5	1D4A6	
1D4A9	1D4AC	
1D4AE	1D4B9	
1D4BB		
1D4BD	1D4C3	
1D4C5	1D505	
1D507	1D50A	
1D50D	1D514	
1D516	1D51C	
1D51E	1D539	
1D53B	1D53E	
1D540	1D544	
1D546		
1D54A	1D550	
1D552	1D6A5	
1D6A8	1D6C0	
1D6C2	1D6DA	
1D6DC	1D6FA	
1D6FC	1D714	
1D716	1D734	
1D736	1D74E	
1D750	1D76E	
1D770	1D788	
1D78A	1D7A8	
1D7AA	1D7C2	
1D7C4	1D7C9	
20000	2A6D6	
2F800	2FA1D	
END
};
1;

package unicore::lib::selfloader::Any;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Any}
# 	\p{Any}
# 
# Meaning: [\x{0000}-\x{10FFFF}]
#
return <<'END';
0000	10FFFF	
END
};
1;

package unicore::lib::selfloader::Arab;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Arabic} (and fuzzy permutations)
# 
# Meaning: Script 'Arabic'
#
return <<'END';
060B		Arabic
060D	0615	Arabic
061E		Arabic
0621	063A	Arabic
0641	064A	Arabic
0656	065E	Arabic
066A	066F	Arabic
0671	06DC	Arabic
06DE	06FF	Arabic
0750	076D	Arabic
FB50	FBB1	Arabic
FBD3	FD3D	Arabic
FD50	FD8F	Arabic
FD92	FDC7	Arabic
FDF0	FDFC	Arabic
FE70	FE74	Arabic
FE76	FEFC	Arabic
END
};
1;

package unicore::lib::selfloader::Armn;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Armenian} (and fuzzy permutations)
# 
# Meaning: Script 'Armenian'
#
return <<'END';
0531	0556	Armenian
0559	055F	Armenian
0561	0587	Armenian
058A		Armenian
FB13	FB17	Armenian
END
};
1;

package unicore::lib::selfloader::ASCII;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{ASCII}
# 
# Meaning: [[:ASCII:]]
#
return <<'END';
0000	007F	
END
};
1;

package unicore::lib::selfloader::AsciiHex;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{AsciiHexDigit} (and fuzzy permutations)
# 
# Meaning: Extended property 'ASCII_Hex_Digit'
#
return <<'END';
0030	0039	ASCII_Hex_Digit
0041	0046	ASCII_Hex_Digit
0061	0066	ASCII_Hex_Digit
END
};
1;

package unicore::lib::selfloader::Assigned;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Assigned}
# 
# Meaning: All assigned code points
#
return <<'END';
0000	0241	
0250	036F	
0374	0375	
037A		
037E		
0384	038A	
038C		
038E	03A1	
03A3	03CE	
03D0	0486	
0488	04CE	
04D0	04F9	
0500	050F	
0531	0556	
0559	055F	
0561	0587	
0589	058A	
0591	05B9	
05BB	05C7	
05D0	05EA	
05F0	05F4	
0600	0603	
060B	0615	
061B		
061E	061F	
0621	063A	
0640	065E	
0660	070D	
070F	074A	
074D	076D	
0780	07B1	
0901	0939	
093C	094D	
0950	0954	
0958	0970	
097D		
0981	0983	
0985	098C	
098F	0990	
0993	09A8	
09AA	09B0	
09B2		
09B6	09B9	
09BC	09C4	
09C7	09C8	
09CB	09CE	
09D7		
09DC	09DD	
09DF	09E3	
09E6	09FA	
0A01	0A03	
0A05	0A0A	
0A0F	0A10	
0A13	0A28	
0A2A	0A30	
0A32	0A33	
0A35	0A36	
0A38	0A39	
0A3C		
0A3E	0A42	
0A47	0A48	
0A4B	0A4D	
0A59	0A5C	
0A5E		
0A66	0A74	
0A81	0A83	
0A85	0A8D	
0A8F	0A91	
0A93	0AA8	
0AAA	0AB0	
0AB2	0AB3	
0AB5	0AB9	
0ABC	0AC5	
0AC7	0AC9	
0ACB	0ACD	
0AD0		
0AE0	0AE3	
0AE6	0AEF	
0AF1		
0B01	0B03	
0B05	0B0C	
0B0F	0B10	
0B13	0B28	
0B2A	0B30	
0B32	0B33	
0B35	0B39	
0B3C	0B43	
0B47	0B48	
0B4B	0B4D	
0B56	0B57	
0B5C	0B5D	
0B5F	0B61	
0B66	0B71	
0B82	0B83	
0B85	0B8A	
0B8E	0B90	
0B92	0B95	
0B99	0B9A	
0B9C		
0B9E	0B9F	
0BA3	0BA4	
0BA8	0BAA	
0BAE	0BB9	
0BBE	0BC2	
0BC6	0BC8	
0BCA	0BCD	
0BD7		
0BE6	0BFA	
0C01	0C03	
0C05	0C0C	
0C0E	0C10	
0C12	0C28	
0C2A	0C33	
0C35	0C39	
0C3E	0C44	
0C46	0C48	
0C4A	0C4D	
0C55	0C56	
0C60	0C61	
0C66	0C6F	
0C82	0C83	
0C85	0C8C	
0C8E	0C90	
0C92	0CA8	
0CAA	0CB3	
0CB5	0CB9	
0CBC	0CC4	
0CC6	0CC8	
0CCA	0CCD	
0CD5	0CD6	
0CDE		
0CE0	0CE1	
0CE6	0CEF	
0D02	0D03	
0D05	0D0C	
0D0E	0D10	
0D12	0D28	
0D2A	0D39	
0D3E	0D43	
0D46	0D48	
0D4A	0D4D	
0D57		
0D60	0D61	
0D66	0D6F	
0D82	0D83	
0D85	0D96	
0D9A	0DB1	
0DB3	0DBB	
0DBD		
0DC0	0DC6	
0DCA		
0DCF	0DD4	
0DD6		
0DD8	0DDF	
0DF2	0DF4	
0E01	0E3A	
0E3F	0E5B	
0E81	0E82	
0E84		
0E87	0E88	
0E8A		
0E8D		
0E94	0E97	
0E99	0E9F	
0EA1	0EA3	
0EA5		
0EA7		
0EAA	0EAB	
0EAD	0EB9	
0EBB	0EBD	
0EC0	0EC4	
0EC6		
0EC8	0ECD	
0ED0	0ED9	
0EDC	0EDD	
0F00	0F47	
0F49	0F6A	
0F71	0F8B	
0F90	0F97	
0F99	0FBC	
0FBE	0FCC	
0FCF	0FD1	
1000	1021	
1023	1027	
1029	102A	
102C	1032	
1036	1039	
1040	1059	
10A0	10C5	
10D0	10FC	
1100	1159	
115F	11A2	
11A8	11F9	
1200	1248	
124A	124D	
1250	1256	
1258		
125A	125D	
1260	1288	
128A	128D	
1290	12B0	
12B2	12B5	
12B8	12BE	
12C0		
12C2	12C5	
12C8	12D6	
12D8	1310	
1312	1315	
1318	135A	
135F	137C	
1380	1399	
13A0	13F4	
1401	1676	
1680	169C	
16A0	16F0	
1700	170C	
170E	1714	
1720	1736	
1740	1753	
1760	176C	
176E	1770	
1772	1773	
1780	17DD	
17E0	17E9	
17F0	17F9	
1800	180E	
1810	1819	
1820	1877	
1880	18A9	
1900	191C	
1920	192B	
1930	193B	
1940		
1944	196D	
1970	1974	
1980	19A9	
19B0	19C9	
19D0	19D9	
19DE	1A1B	
1A1E	1A1F	
1D00	1DC3	
1E00	1E9B	
1EA0	1EF9	
1F00	1F15	
1F18	1F1D	
1F20	1F45	
1F48	1F4D	
1F50	1F57	
1F59		
1F5B		
1F5D		
1F5F	1F7D	
1F80	1FB4	
1FB6	1FC4	
1FC6	1FD3	
1FD6	1FDB	
1FDD	1FEF	
1FF2	1FF4	
1FF6	1FFE	
2000	2063	
206A	2071	
2074	208E	
2090	2094	
20A0	20B5	
20D0	20EB	
2100	214C	
2153	2183	
2190	23DB	
2400	2426	
2440	244A	
2460	269C	
26A0	26B1	
2701	2704	
2706	2709	
270C	2727	
2729	274B	
274D		
274F	2752	
2756		
2758	275E	
2761	2794	
2798	27AF	
27B1	27BE	
27C0	27C6	
27D0	27EB	
27F0	2B13	
2C00	2C2E	
2C30	2C5E	
2C80	2CEA	
2CF9	2D25	
2D30	2D65	
2D6F		
2D80	2D96	
2DA0	2DA6	
2DA8	2DAE	
2DB0	2DB6	
2DB8	2DBE	
2DC0	2DC6	
2DC8	2DCE	
2DD0	2DD6	
2DD8	2DDE	
2E00	2E17	
2E1C	2E1D	
2E80	2E99	
2E9B	2EF3	
2F00	2FD5	
2FF0	2FFB	
3000	303F	
3041	3096	
3099	30FF	
3105	312C	
3131	318E	
3190	31B7	
31C0	31CF	
31F0	321E	
3220	3243	
3250	32FE	
3300	4DB5	
4DC0	9FBB	
A000	A48C	
A490	A4C6	
A700	A716	
A800	A82B	
AC00	D7A3	
D800	FA2D	
FA30	FA6A	
FA70	FAD9	
FB00	FB06	
FB13	FB17	
FB1D	FB36	
FB38	FB3C	
FB3E		
FB40	FB41	
FB43	FB44	
FB46	FBB1	
FBD3	FD3F	
FD50	FD8F	
FD92	FDC7	
FDF0	FDFD	
FE00	FE19	
FE20	FE23	
FE30	FE52	
FE54	FE66	
FE68	FE6B	
FE70	FE74	
FE76	FEFC	
FEFF		
FF01	FFBE	
FFC2	FFC7	
FFCA	FFCF	
FFD2	FFD7	
FFDA	FFDC	
FFE0	FFE6	
FFE8	FFEE	
FFF9	FFFD	
10000	1000B	
1000D	10026	
10028	1003A	
1003C	1003D	
1003F	1004D	
10050	1005D	
10080	100FA	
10100	10102	
10107	10133	
10137	1018A	
10300	1031E	
10320	10323	
10330	1034A	
10380	1039D	
1039F	103C3	
103C8	103D5	
10400	1049D	
104A0	104A9	
10800	10805	
10808		
1080A	10835	
10837	10838	
1083C		
1083F		
10A00	10A03	
10A05	10A06	
10A0C	10A13	
10A15	10A17	
10A19	10A33	
10A38	10A3A	
10A3F	10A47	
10A50	10A58	
1D000	1D0F5	
1D100	1D126	
1D12A	1D1DD	
1D200	1D245	
1D300	1D356	
1D400	1D454	
1D456	1D49C	
1D49E	1D49F	
1D4A2		
1D4A5	1D4A6	
1D4A9	1D4AC	
1D4AE	1D4B9	
1D4BB		
1D4BD	1D4C3	
1D4C5	1D505	
1D507	1D50A	
1D50D	1D514	
1D516	1D51C	
1D51E	1D539	
1D53B	1D53E	
1D540	1D544	
1D546		
1D54A	1D550	
1D552	1D6A5	
1D6A8	1D7C9	
1D7CE	1D7FF	
20000	2A6D6	
2F800	2FA1D	
E0001		
E0020	E007F	
E0100	E01EF	
F0000	FFFFD	
100000	10FFFD	
END
};
1;

package unicore::lib::selfloader::Beng;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Bengali} (and fuzzy permutations)
# 
# Meaning: Script 'Bengali'
#
return <<'END';
0981	0983	Bengali
0985	098C	Bengali
098F	0990	Bengali
0993	09A8	Bengali
09AA	09B0	Bengali
09B2		Bengali
09B6	09B9	Bengali
09BC	09C4	Bengali
09C7	09C8	Bengali
09CB	09CE	Bengali
09D7		Bengali
09DC	09DD	Bengali
09DF	09E3	Bengali
09E6	09FA	Bengali
END
};
1;

package unicore::lib::selfloader::BidiC;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Bidi_Control'
#
return <<'END';
200E	200F	Bidi_Control
202A	202E	Bidi_Control
END
};
1;

package unicore::lib::selfloader::BidiCont;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{BidiControl} (and fuzzy permutations)
# 
# Meaning: Extended property 'Bidi_Control'
#
return <<'END';
200E	200F	Bidi_Control
202A	202E	Bidi_Control
END
};
1;

package unicore::lib::selfloader::Blank;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Blank}
# 
# Meaning: [[:Blank:]]
#
return <<'END';
0009		
0020		
00A0		
1680		
180E		
2000	200A	
202F		
205F		
3000		
END
};
1;

package unicore::lib::selfloader::Bopo;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Bopomofo} (and fuzzy permutations)
# 
# Meaning: Script 'Bopomofo'
#
return <<'END';
3105	312C	Bopomofo
31A0	31B7	Bopomofo
END
};
1;

package unicore::lib::selfloader::Brai;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Braille} (and fuzzy permutations)
# 
# Meaning: Script 'Braille'
#
return <<'END';
2800	28FF	Braille
END
};
1;

package unicore::lib::selfloader::Bugi;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Buginese} (and fuzzy permutations)
# 
# Meaning: Script 'Buginese'
#
return <<'END';
1A00	1A1B	Buginese
1A1E	1A1F	Buginese
END
};
1;

package unicore::lib::selfloader::Buhd;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Buhid} (and fuzzy permutations)
# 
# Meaning: Script 'Buhid'
#
return <<'END';
1740	1753	Buhid
END
};
1;

package unicore::lib::selfloader::C;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{C}
# 	\p{C} (and fuzzy permutations)
# 
# Meaning: Major Category 'C'
#
return <<'END';
0000	001F	
007F	009F	
00AD		
0242	024F	
0370	0373	
0376	0379	
037B	037D	
037F	0383	
038B		
038D		
03A2		
03CF		
0487		
04CF		
04FA	04FF	
0510	0530	
0557	0558	
0560		
0588		
058B	0590	
05BA		
05C8	05CF	
05EB	05EF	
05F5	060A	
0616	061A	
061C	061D	
0620		
063B	063F	
065F		
06DD		
070E	070F	
074B	074C	
076E	077F	
07B2	0900	
093A	093B	
094E	094F	
0955	0957	
0971	097C	
097E	0980	
0984		
098D	098E	
0991	0992	
09A9		
09B1		
09B3	09B5	
09BA	09BB	
09C5	09C6	
09C9	09CA	
09CF	09D6	
09D8	09DB	
09DE		
09E4	09E5	
09FB	0A00	
0A04		
0A0B	0A0E	
0A11	0A12	
0A29		
0A31		
0A34		
0A37		
0A3A	0A3B	
0A3D		
0A43	0A46	
0A49	0A4A	
0A4E	0A58	
0A5D		
0A5F	0A65	
0A75	0A80	
0A84		
0A8E		
0A92		
0AA9		
0AB1		
0AB4		
0ABA	0ABB	
0AC6		
0ACA		
0ACE	0ACF	
0AD1	0ADF	
0AE4	0AE5	
0AF0		
0AF2	0B00	
0B04		
0B0D	0B0E	
0B11	0B12	
0B29		
0B31		
0B34		
0B3A	0B3B	
0B44	0B46	
0B49	0B4A	
0B4E	0B55	
0B58	0B5B	
0B5E		
0B62	0B65	
0B72	0B81	
0B84		
0B8B	0B8D	
0B91		
0B96	0B98	
0B9B		
0B9D		
0BA0	0BA2	
0BA5	0BA7	
0BAB	0BAD	
0BBA	0BBD	
0BC3	0BC5	
0BC9		
0BCE	0BD6	
0BD8	0BE5	
0BFB	0C00	
0C04		
0C0D		
0C11		
0C29		
0C34		
0C3A	0C3D	
0C45		
0C49		
0C4E	0C54	
0C57	0C5F	
0C62	0C65	
0C70	0C81	
0C84		
0C8D		
0C91		
0CA9		
0CB4		
0CBA	0CBB	
0CC5		
0CC9		
0CCE	0CD4	
0CD7	0CDD	
0CDF		
0CE2	0CE5	
0CF0	0D01	
0D04		
0D0D		
0D11		
0D29		
0D3A	0D3D	
0D44	0D45	
0D49		
0D4E	0D56	
0D58	0D5F	
0D62	0D65	
0D70	0D81	
0D84		
0D97	0D99	
0DB2		
0DBC		
0DBE	0DBF	
0DC7	0DC9	
0DCB	0DCE	
0DD5		
0DD7		
0DE0	0DF1	
0DF5	0E00	
0E3B	0E3E	
0E5C	0E80	
0E83		
0E85	0E86	
0E89		
0E8B	0E8C	
0E8E	0E93	
0E98		
0EA0		
0EA4		
0EA6		
0EA8	0EA9	
0EAC		
0EBA		
0EBE	0EBF	
0EC5		
0EC7		
0ECE	0ECF	
0EDA	0EDB	
0EDE	0EFF	
0F48		
0F6B	0F70	
0F8C	0F8F	
0F98		
0FBD		
0FCD	0FCE	
0FD2	0FFF	
1022		
1028		
102B		
1033	1035	
103A	103F	
105A	109F	
10C6	10CF	
10FD	10FF	
115A	115E	
11A3	11A7	
11FA	11FF	
1249		
124E	124F	
1257		
1259		
125E	125F	
1289		
128E	128F	
12B1		
12B6	12B7	
12BF		
12C1		
12C6	12C7	
12D7		
1311		
1316	1317	
135B	135E	
137D	137F	
139A	139F	
13F5	1400	
1677	167F	
169D	169F	
16F1	16FF	
170D		
1715	171F	
1737	173F	
1754	175F	
176D		
1771		
1774	177F	
17B4	17B5	
17DE	17DF	
17EA	17EF	
17FA	17FF	
180F		
181A	181F	
1878	187F	
18AA	18FF	
191D	191F	
192C	192F	
193C	193F	
1941	1943	
196E	196F	
1975	197F	
19AA	19AF	
19CA	19CF	
19DA	19DD	
1A1C	1A1D	
1A20	1CFF	
1DC4	1DFF	
1E9C	1E9F	
1EFA	1EFF	
1F16	1F17	
1F1E	1F1F	
1F46	1F47	
1F4E	1F4F	
1F58		
1F5A		
1F5C		
1F5E		
1F7E	1F7F	
1FB5		
1FC5		
1FD4	1FD5	
1FDC		
1FF0	1FF1	
1FF5		
1FFF		
200B	200F	
202A	202E	
2060	206F	
2072	2073	
208F		
2095	209F	
20B6	20CF	
20EC	20FF	
214D	2152	
2184	218F	
23DC	23FF	
2427	243F	
244B	245F	
269D	269F	
26B2	2700	
2705		
270A	270B	
2728		
274C		
274E		
2753	2755	
2757		
275F	2760	
2795	2797	
27B0		
27BF		
27C7	27CF	
27EC	27EF	
2B14	2BFF	
2C2F		
2C5F	2C7F	
2CEB	2CF8	
2D26	2D2F	
2D66	2D6E	
2D70	2D7F	
2D97	2D9F	
2DA7		
2DAF		
2DB7		
2DBF		
2DC7		
2DCF		
2DD7		
2DDF	2DFF	
2E18	2E1B	
2E1E	2E7F	
2E9A		
2EF4	2EFF	
2FD6	2FEF	
2FFC	2FFF	
3040		
3097	3098	
3100	3104	
312D	3130	
318F		
31B8	31BF	
31D0	31EF	
321F		
3244	324F	
32FF		
4DB6	4DBF	
9FBC	9FFF	
A48D	A48F	
A4C7	A6FF	
A717	A7FF	
A82C	ABFF	
D7A4	F8FF	
FA2E	FA2F	
FA6B	FA6F	
FADA	FAFF	
FB07	FB12	
FB18	FB1C	
FB37		
FB3D		
FB3F		
FB42		
FB45		
FBB2	FBD2	
FD40	FD4F	
FD90	FD91	
FDC8	FDEF	
FDFE	FDFF	
FE1A	FE1F	
FE24	FE2F	
FE53		
FE67		
FE6C	FE6F	
FE75		
FEFD	FF00	
FFBF	FFC1	
FFC8	FFC9	
FFD0	FFD1	
FFD8	FFD9	
FFDD	FFDF	
FFE7		
FFEF	FFFB	
FFFE	FFFF	
1000C		
10027		
1003B		
1003E		
1004E	1004F	
1005E	1007F	
100FB	100FF	
10103	10106	
10134	10136	
1018B	102FF	
1031F		
10324	1032F	
1034B	1037F	
1039E		
103C4	103C7	
103D6	103FF	
1049E	1049F	
104AA	107FF	
10806	10807	
10809		
10836		
10839	1083B	
1083D	1083E	
10840	109FF	
10A04		
10A07	10A0B	
10A14		
10A18		
10A34	10A37	
10A3B	10A3E	
10A48	10A4F	
10A59	1CFFF	
1D0F6	1D0FF	
1D127	1D129	
1D173	1D17A	
1D1DE	1D1FF	
1D246	1D2FF	
1D357	1D3FF	
1D455		
1D49D		
1D4A0	1D4A1	
1D4A3	1D4A4	
1D4A7	1D4A8	
1D4AD		
1D4BA		
1D4BC		
1D4C4		
1D506		
1D50B	1D50C	
1D515		
1D51D		
1D53A		
1D53F		
1D545		
1D547	1D549	
1D551		
1D6A6	1D6A7	
1D7CA	1D7CD	
1D800	1FFFF	
2A6D7	2F7FF	
2FA1E	E00FF	
E01F0	10FFFF	
END
};
1;

package unicore::lib::selfloader::Canadian;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{CanadianAboriginal} (and fuzzy permutations)
# 
# Meaning: Script 'Canadian_Aboriginal'
#
return <<'END';
1401	1676	Canadian_Aboriginal
END
};
1;

package unicore::lib::selfloader::Cc;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Cc}
# 	\p{Cc} (and fuzzy permutations)
# 
# Meaning: General Category 'Cc'
#
return <<'END';
0000	001F	
007F	009F	
END
};
1;

package unicore::lib::selfloader::Cf;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Cf}
# 	\p{Cf} (and fuzzy permutations)
# 
# Meaning: General Category 'Cf'
#
return <<'END';
00AD		
0600	0603	
06DD		
070F		
17B4	17B5	
200B	200F	
202A	202E	
2060	2063	
206A	206F	
FEFF		
FFF9	FFFB	
1D173	1D17A	
E0001		
E0020	E007F	
END
};
1;

package unicore::lib::selfloader::Cher;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Cherokee} (and fuzzy permutations)
# 
# Meaning: Script 'Cherokee'
#
return <<'END';
13A0	13F4	Cherokee
END
};
1;

package unicore::lib::selfloader::Cn;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Cn}
# 	\p{Cn} (and fuzzy permutations)
# 
# Meaning: General Category 'Cn' [not functional in Perl]
#
return <<'END';
0242	024F	
0370	0373	
0376	0379	
037B	037D	
037F	0383	
038B		
038D		
03A2		
03CF		
0487		
04CF		
04FA	04FF	
0510	0530	
0557	0558	
0560		
0588		
058B	0590	
05BA		
05C8	05CF	
05EB	05EF	
05F5	05FF	
0604	060A	
0616	061A	
061C	061D	
0620		
063B	063F	
065F		
070E		
074B	074C	
076E	077F	
07B2	0900	
093A	093B	
094E	094F	
0955	0957	
0971	097C	
097E	0980	
0984		
098D	098E	
0991	0992	
09A9		
09B1		
09B3	09B5	
09BA	09BB	
09C5	09C6	
09C9	09CA	
09CF	09D6	
09D8	09DB	
09DE		
09E4	09E5	
09FB	0A00	
0A04		
0A0B	0A0E	
0A11	0A12	
0A29		
0A31		
0A34		
0A37		
0A3A	0A3B	
0A3D		
0A43	0A46	
0A49	0A4A	
0A4E	0A58	
0A5D		
0A5F	0A65	
0A75	0A80	
0A84		
0A8E		
0A92		
0AA9		
0AB1		
0AB4		
0ABA	0ABB	
0AC6		
0ACA		
0ACE	0ACF	
0AD1	0ADF	
0AE4	0AE5	
0AF0		
0AF2	0B00	
0B04		
0B0D	0B0E	
0B11	0B12	
0B29		
0B31		
0B34		
0B3A	0B3B	
0B44	0B46	
0B49	0B4A	
0B4E	0B55	
0B58	0B5B	
0B5E		
0B62	0B65	
0B72	0B81	
0B84		
0B8B	0B8D	
0B91		
0B96	0B98	
0B9B		
0B9D		
0BA0	0BA2	
0BA5	0BA7	
0BAB	0BAD	
0BBA	0BBD	
0BC3	0BC5	
0BC9		
0BCE	0BD6	
0BD8	0BE5	
0BFB	0C00	
0C04		
0C0D		
0C11		
0C29		
0C34		
0C3A	0C3D	
0C45		
0C49		
0C4E	0C54	
0C57	0C5F	
0C62	0C65	
0C70	0C81	
0C84		
0C8D		
0C91		
0CA9		
0CB4		
0CBA	0CBB	
0CC5		
0CC9		
0CCE	0CD4	
0CD7	0CDD	
0CDF		
0CE2	0CE5	
0CF0	0D01	
0D04		
0D0D		
0D11		
0D29		
0D3A	0D3D	
0D44	0D45	
0D49		
0D4E	0D56	
0D58	0D5F	
0D62	0D65	
0D70	0D81	
0D84		
0D97	0D99	
0DB2		
0DBC		
0DBE	0DBF	
0DC7	0DC9	
0DCB	0DCE	
0DD5		
0DD7		
0DE0	0DF1	
0DF5	0E00	
0E3B	0E3E	
0E5C	0E80	
0E83		
0E85	0E86	
0E89		
0E8B	0E8C	
0E8E	0E93	
0E98		
0EA0		
0EA4		
0EA6		
0EA8	0EA9	
0EAC		
0EBA		
0EBE	0EBF	
0EC5		
0EC7		
0ECE	0ECF	
0EDA	0EDB	
0EDE	0EFF	
0F48		
0F6B	0F70	
0F8C	0F8F	
0F98		
0FBD		
0FCD	0FCE	
0FD2	0FFF	
1022		
1028		
102B		
1033	1035	
103A	103F	
105A	109F	
10C6	10CF	
10FD	10FF	
115A	115E	
11A3	11A7	
11FA	11FF	
1249		
124E	124F	
1257		
1259		
125E	125F	
1289		
128E	128F	
12B1		
12B6	12B7	
12BF		
12C1		
12C6	12C7	
12D7		
1311		
1316	1317	
135B	135E	
137D	137F	
139A	139F	
13F5	1400	
1677	167F	
169D	169F	
16F1	16FF	
170D		
1715	171F	
1737	173F	
1754	175F	
176D		
1771		
1774	177F	
17DE	17DF	
17EA	17EF	
17FA	17FF	
180F		
181A	181F	
1878	187F	
18AA	18FF	
191D	191F	
192C	192F	
193C	193F	
1941	1943	
196E	196F	
1975	197F	
19AA	19AF	
19CA	19CF	
19DA	19DD	
1A1C	1A1D	
1A20	1CFF	
1DC4	1DFF	
1E9C	1E9F	
1EFA	1EFF	
1F16	1F17	
1F1E	1F1F	
1F46	1F47	
1F4E	1F4F	
1F58		
1F5A		
1F5C		
1F5E		
1F7E	1F7F	
1FB5		
1FC5		
1FD4	1FD5	
1FDC		
1FF0	1FF1	
1FF5		
1FFF		
2064	2069	
2072	2073	
208F		
2095	209F	
20B6	20CF	
20EC	20FF	
214D	2152	
2184	218F	
23DC	23FF	
2427	243F	
244B	245F	
269D	269F	
26B2	2700	
2705		
270A	270B	
2728		
274C		
274E		
2753	2755	
2757		
275F	2760	
2795	2797	
27B0		
27BF		
27C7	27CF	
27EC	27EF	
2B14	2BFF	
2C2F		
2C5F	2C7F	
2CEB	2CF8	
2D26	2D2F	
2D66	2D6E	
2D70	2D7F	
2D97	2D9F	
2DA7		
2DAF		
2DB7		
2DBF		
2DC7		
2DCF		
2DD7		
2DDF	2DFF	
2E18	2E1B	
2E1E	2E7F	
2E9A		
2EF4	2EFF	
2FD6	2FEF	
2FFC	2FFF	
3040		
3097	3098	
3100	3104	
312D	3130	
318F		
31B8	31BF	
31D0	31EF	
321F		
3244	324F	
32FF		
4DB6	4DBF	
9FBC	9FFF	
A48D	A48F	
A4C7	A6FF	
A717	A7FF	
A82C	ABFF	
D7A4	D7FF	
FA2E	FA2F	
FA6B	FA6F	
FADA	FAFF	
FB07	FB12	
FB18	FB1C	
FB37		
FB3D		
FB3F		
FB42		
FB45		
FBB2	FBD2	
FD40	FD4F	
FD90	FD91	
FDC8	FDEF	
FDFE	FDFF	
FE1A	FE1F	
FE24	FE2F	
FE53		
FE67		
FE6C	FE6F	
FE75		
FEFD	FEFE	
FF00		
FFBF	FFC1	
FFC8	FFC9	
FFD0	FFD1	
FFD8	FFD9	
FFDD	FFDF	
FFE7		
FFEF	FFF8	
FFFE	FFFF	
1000C		
10027		
1003B		
1003E		
1004E	1004F	
1005E	1007F	
100FB	100FF	
10103	10106	
10134	10136	
1018B	102FF	
1031F		
10324	1032F	
1034B	1037F	
1039E		
103C4	103C7	
103D6	103FF	
1049E	1049F	
104AA	107FF	
10806	10807	
10809		
10836		
10839	1083B	
1083D	1083E	
10840	109FF	
10A04		
10A07	10A0B	
10A14		
10A18		
10A34	10A37	
10A3B	10A3E	
10A48	10A4F	
10A59	1CFFF	
1D0F6	1D0FF	
1D127	1D129	
1D1DE	1D1FF	
1D246	1D2FF	
1D357	1D3FF	
1D455		
1D49D		
1D4A0	1D4A1	
1D4A3	1D4A4	
1D4A7	1D4A8	
1D4AD		
1D4BA		
1D4BC		
1D4C4		
1D506		
1D50B	1D50C	
1D515		
1D51D		
1D53A		
1D53F		
1D545		
1D547	1D549	
1D551		
1D6A6	1D6A7	
1D7CA	1D7CD	
1D800	1FFFF	
2A6D7	2F7FF	
2FA1E	E0000	
E0002	E001F	
E0080	E00FF	
E01F0	EFFFF	
FFFFE	FFFFF	
10FFFE	10FFFF	
END
};
1;

package unicore::lib::selfloader::Cntrl;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Cntrl}
# 
# Meaning: [[:Cntrl:]]
#
return <<'END';
0000	001F	
007F	009F	
00AD		
0600	0603	
06DD		
070F		
17B4	17B5	
200B	200F	
202A	202E	
2060	2063	
206A	206F	
D800	F8FF	
FEFF		
FFF9	FFFB	
1D173	1D17A	
E0001		
E0020	E007F	
F0000	FFFFD	
100000	10FFFD	
END
};
1;

package unicore::lib::selfloader::Co;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Co}
# 	\p{Co} (and fuzzy permutations)
# 
# Meaning: General Category 'Co'
#
return <<'END';
E000	F8FF	
F0000	FFFFD	
100000	10FFFD	
END
};
1;

package unicore::lib::selfloader::Copt;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Coptic} (and fuzzy permutations)
# 
# Meaning: Script 'Coptic'
#
return <<'END';
03E2	03EF	Coptic
2C80	2CEA	Coptic
2CF9	2CFF	Coptic
END
};
1;

package unicore::lib::selfloader::Cprt;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Cypriot} (and fuzzy permutations)
# 
# Meaning: Script 'Cypriot'
#
return <<'END';
10800	10805	Cypriot
10808		Cypriot
1080A	10835	Cypriot
10837	10838	Cypriot
1083C		Cypriot
1083F		Cypriot
END
};
1;

package unicore::lib::selfloader::Cs;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Cs}
# 	\p{Cs} (and fuzzy permutations)
# 
# Meaning: General Category 'Cs'
#
return <<'END';
D800	DFFF	
END
};
1;

package unicore::lib::selfloader::Cyrl;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Cyrillic} (and fuzzy permutations)
# 
# Meaning: Script 'Cyrillic'
#
return <<'END';
0400	0486	Cyrillic
0488	04CE	Cyrillic
04D0	04F9	Cyrillic
0500	050F	Cyrillic
1D2B		Cyrillic
1D78		Cyrillic
END
};
1;

package unicore::lib::selfloader::Dash;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Dash'
#
return <<'END';
002D		Dash
058A		Dash
1806		Dash
2010	2015	Dash
2053		Dash
207B		Dash
208B		Dash
2212		Dash
2E17		Dash
301C		Dash
3030		Dash
30A0		Dash
FE31	FE32	Dash
FE58		Dash
FE63		Dash
FF0D		Dash
END
};
1;

package unicore::lib::selfloader::Dash2;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Dash} (and fuzzy permutations)
# 
# Meaning: Extended property 'Dash'
#
return <<'END';
002D		Dash
058A		Dash
1806		Dash
2010	2015	Dash
2053		Dash
207B		Dash
208B		Dash
2212		Dash
2E17		Dash
301C		Dash
3030		Dash
30A0		Dash
FE31	FE32	Dash
FE58		Dash
FE63		Dash
FF0D		Dash
END
};
1;

package unicore::lib::selfloader::Dep;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Deprecated'
#
return <<'END';
0340	0341	Deprecated
17A3		Deprecated
17D3		Deprecated
206A	206F	Deprecated
END
};
1;

package unicore::lib::selfloader::Deprecat;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Deprecated} (and fuzzy permutations)
# 
# Meaning: Extended property 'Deprecated'
#
return <<'END';
0340	0341	Deprecated
17A3		Deprecated
17D3		Deprecated
206A	206F	Deprecated
END
};
1;

package unicore::lib::selfloader::Deva;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Devanagari} (and fuzzy permutations)
# 
# Meaning: Script 'Devanagari'
#
return <<'END';
0901	0939	Devanagari
093C	094D	Devanagari
0950	0954	Devanagari
0958	0963	Devanagari
0966	096F	Devanagari
097D		Devanagari
END
};
1;

package unicore::lib::selfloader::Dia;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Diacritic'
#
return <<'END';
005E		Diacritic
0060		Diacritic
00A8		Diacritic
00AF		Diacritic
00B4		Diacritic
00B7	00B8	Diacritic
02B0	034E	Diacritic
0350	0357	Diacritic
035D	0362	Diacritic
0374	0375	Diacritic
037A		Diacritic
0384	0385	Diacritic
0483	0486	Diacritic
0559		Diacritic
0591	05A1	Diacritic
05A3	05B9	Diacritic
05BB	05BD	Diacritic
05BF		Diacritic
05C1	05C2	Diacritic
05C4		Diacritic
064B	0652	Diacritic
0657	0658	Diacritic
06DF	06E0	Diacritic
06E5	06E6	Diacritic
06EA	06EC	Diacritic
0730	074A	Diacritic
07A6	07B0	Diacritic
093C		Diacritic
094D		Diacritic
0951	0954	Diacritic
09BC		Diacritic
09CD		Diacritic
0A3C		Diacritic
0A4D		Diacritic
0ABC		Diacritic
0ACD		Diacritic
0B3C		Diacritic
0B4D		Diacritic
0BCD		Diacritic
0C4D		Diacritic
0CBC		Diacritic
0CCD		Diacritic
0D4D		Diacritic
0DCA		Diacritic
0E47	0E4C	Diacritic
0E4E		Diacritic
0EC8	0ECC	Diacritic
0F18	0F19	Diacritic
0F35		Diacritic
0F37		Diacritic
0F39		Diacritic
0F3E	0F3F	Diacritic
0F82	0F84	Diacritic
0F86	0F87	Diacritic
0FC6		Diacritic
1037		Diacritic
1039		Diacritic
17C9	17D3	Diacritic
17DD		Diacritic
1939	193B	Diacritic
1D2C	1D6A	Diacritic
1FBD		Diacritic
1FBF	1FC1	Diacritic
1FCD	1FCF	Diacritic
1FDD	1FDF	Diacritic
1FED	1FEF	Diacritic
1FFD	1FFE	Diacritic
302A	302F	Diacritic
3099	309C	Diacritic
30FC		Diacritic
FB1E		Diacritic
FE20	FE23	Diacritic
FF3E		Diacritic
FF40		Diacritic
FF70		Diacritic
FF9E	FF9F	Diacritic
FFE3		Diacritic
1D167	1D169	Diacritic
1D16D	1D172	Diacritic
1D17B	1D182	Diacritic
1D185	1D18B	Diacritic
1D1AA	1D1AD	Diacritic
END
};
1;

package unicore::lib::selfloader::Diacriti;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Diacritic} (and fuzzy permutations)
# 
# Meaning: Extended property 'Diacritic'
#
return <<'END';
005E		Diacritic
0060		Diacritic
00A8		Diacritic
00AF		Diacritic
00B4		Diacritic
00B7	00B8	Diacritic
02B0	034E	Diacritic
0350	0357	Diacritic
035D	0362	Diacritic
0374	0375	Diacritic
037A		Diacritic
0384	0385	Diacritic
0483	0486	Diacritic
0559		Diacritic
0591	05A1	Diacritic
05A3	05B9	Diacritic
05BB	05BD	Diacritic
05BF		Diacritic
05C1	05C2	Diacritic
05C4		Diacritic
064B	0652	Diacritic
0657	0658	Diacritic
06DF	06E0	Diacritic
06E5	06E6	Diacritic
06EA	06EC	Diacritic
0730	074A	Diacritic
07A6	07B0	Diacritic
093C		Diacritic
094D		Diacritic
0951	0954	Diacritic
09BC		Diacritic
09CD		Diacritic
0A3C		Diacritic
0A4D		Diacritic
0ABC		Diacritic
0ACD		Diacritic
0B3C		Diacritic
0B4D		Diacritic
0BCD		Diacritic
0C4D		Diacritic
0CBC		Diacritic
0CCD		Diacritic
0D4D		Diacritic
0DCA		Diacritic
0E47	0E4C	Diacritic
0E4E		Diacritic
0EC8	0ECC	Diacritic
0F18	0F19	Diacritic
0F35		Diacritic
0F37		Diacritic
0F39		Diacritic
0F3E	0F3F	Diacritic
0F82	0F84	Diacritic
0F86	0F87	Diacritic
0FC6		Diacritic
1037		Diacritic
1039		Diacritic
17C9	17D3	Diacritic
17DD		Diacritic
1939	193B	Diacritic
1D2C	1D6A	Diacritic
1FBD		Diacritic
1FBF	1FC1	Diacritic
1FCD	1FCF	Diacritic
1FDD	1FDF	Diacritic
1FED	1FEF	Diacritic
1FFD	1FFE	Diacritic
302A	302F	Diacritic
3099	309C	Diacritic
30FC		Diacritic
FB1E		Diacritic
FE20	FE23	Diacritic
FF3E		Diacritic
FF40		Diacritic
FF70		Diacritic
FF9E	FF9F	Diacritic
FFE3		Diacritic
1D167	1D169	Diacritic
1D16D	1D172	Diacritic
1D17B	1D182	Diacritic
1D185	1D18B	Diacritic
1D1AA	1D1AD	Diacritic
END
};
1;

package unicore::lib::selfloader::Digit;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Digit}
# 
# Meaning: [[:Digit:]]
#
return <<'END';
0030	0039	
0660	0669	
06F0	06F9	
0966	096F	
09E6	09EF	
0A66	0A6F	
0AE6	0AEF	
0B66	0B6F	
0BE6	0BEF	
0C66	0C6F	
0CE6	0CEF	
0D66	0D6F	
0E50	0E59	
0ED0	0ED9	
0F20	0F29	
1040	1049	
17E0	17E9	
1810	1819	
1946	194F	
19D0	19D9	
FF10	FF19	
104A0	104A9	
1D7CE	1D7FF	
END
};
1;

package unicore::lib::selfloader::Dsrt;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Deseret} (and fuzzy permutations)
# 
# Meaning: Script 'Deseret'
#
return <<'END';
10400	1044F	Deseret
END
};
1;

package unicore::lib::selfloader::Ethi;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Ethiopic} (and fuzzy permutations)
# 
# Meaning: Script 'Ethiopic'
#
return <<'END';
1200	1248	Ethiopic
124A	124D	Ethiopic
1250	1256	Ethiopic
1258		Ethiopic
125A	125D	Ethiopic
1260	1288	Ethiopic
128A	128D	Ethiopic
1290	12B0	Ethiopic
12B2	12B5	Ethiopic
12B8	12BE	Ethiopic
12C0		Ethiopic
12C2	12C5	Ethiopic
12C8	12D6	Ethiopic
12D8	1310	Ethiopic
1312	1315	Ethiopic
1318	135A	Ethiopic
135F	137C	Ethiopic
1380	1399	Ethiopic
2D80	2D96	Ethiopic
2DA0	2DA6	Ethiopic
2DA8	2DAE	Ethiopic
2DB0	2DB6	Ethiopic
2DB8	2DBE	Ethiopic
2DC0	2DC6	Ethiopic
2DC8	2DCE	Ethiopic
2DD0	2DD6	Ethiopic
2DD8	2DDE	Ethiopic
END
};
1;

package unicore::lib::selfloader::Ext;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Extender'
#
return <<'END';
00B7		Extender
02D0	02D1	Extender
0640		Extender
0E46		Extender
0EC6		Extender
1843		Extender
3005		Extender
3031	3035	Extender
309D	309E	Extender
30FC	30FE	Extender
A015		Extender
FF70		Extender
END
};
1;

package unicore::lib::selfloader::Extender;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Extender} (and fuzzy permutations)
# 
# Meaning: Extended property 'Extender'
#
return <<'END';
00B7		Extender
02D0	02D1	Extender
0640		Extender
0E46		Extender
0EC6		Extender
1843		Extender
3005		Extender
3031	3035	Extender
309D	309E	Extender
30FC	30FE	Extender
A015		Extender
FF70		Extender
END
};
1;

package unicore::lib::selfloader::Geor;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Georgian} (and fuzzy permutations)
# 
# Meaning: Script 'Georgian'
#
return <<'END';
10A0	10C5	Georgian
10D0	10FA	Georgian
10FC		Georgian
2D00	2D25	Georgian
END
};
1;

package unicore::lib::selfloader::Glag;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Glagolitic} (and fuzzy permutations)
# 
# Meaning: Script 'Glagolitic'
#
return <<'END';
2C00	2C2E	Glagolitic
2C30	2C5E	Glagolitic
END
};
1;

package unicore::lib::selfloader::Goth;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Gothic} (and fuzzy permutations)
# 
# Meaning: Script 'Gothic'
#
return <<'END';
10330	1034A	Gothic
END
};
1;

package unicore::lib::selfloader::Graph;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Graph}
# 
# Meaning: [[:Graph:]]
#
return <<'END';
0021	007E	
00A1	0241	
0250	036F	
0374	0375	
037A		
037E		
0384	038A	
038C		
038E	03A1	
03A3	03CE	
03D0	0486	
0488	04CE	
04D0	04F9	
0500	050F	
0531	0556	
0559	055F	
0561	0587	
0589	058A	
0591	05B9	
05BB	05C7	
05D0	05EA	
05F0	05F4	
0600	0603	
060B	0615	
061B		
061E	061F	
0621	063A	
0640	065E	
0660	070D	
070F	074A	
074D	076D	
0780	07B1	
0901	0939	
093C	094D	
0950	0954	
0958	0970	
097D		
0981	0983	
0985	098C	
098F	0990	
0993	09A8	
09AA	09B0	
09B2		
09B6	09B9	
09BC	09C4	
09C7	09C8	
09CB	09CE	
09D7		
09DC	09DD	
09DF	09E3	
09E6	09FA	
0A01	0A03	
0A05	0A0A	
0A0F	0A10	
0A13	0A28	
0A2A	0A30	
0A32	0A33	
0A35	0A36	
0A38	0A39	
0A3C		
0A3E	0A42	
0A47	0A48	
0A4B	0A4D	
0A59	0A5C	
0A5E		
0A66	0A74	
0A81	0A83	
0A85	0A8D	
0A8F	0A91	
0A93	0AA8	
0AAA	0AB0	
0AB2	0AB3	
0AB5	0AB9	
0ABC	0AC5	
0AC7	0AC9	
0ACB	0ACD	
0AD0		
0AE0	0AE3	
0AE6	0AEF	
0AF1		
0B01	0B03	
0B05	0B0C	
0B0F	0B10	
0B13	0B28	
0B2A	0B30	
0B32	0B33	
0B35	0B39	
0B3C	0B43	
0B47	0B48	
0B4B	0B4D	
0B56	0B57	
0B5C	0B5D	
0B5F	0B61	
0B66	0B71	
0B82	0B83	
0B85	0B8A	
0B8E	0B90	
0B92	0B95	
0B99	0B9A	
0B9C		
0B9E	0B9F	
0BA3	0BA4	
0BA8	0BAA	
0BAE	0BB9	
0BBE	0BC2	
0BC6	0BC8	
0BCA	0BCD	
0BD7		
0BE6	0BFA	
0C01	0C03	
0C05	0C0C	
0C0E	0C10	
0C12	0C28	
0C2A	0C33	
0C35	0C39	
0C3E	0C44	
0C46	0C48	
0C4A	0C4D	
0C55	0C56	
0C60	0C61	
0C66	0C6F	
0C82	0C83	
0C85	0C8C	
0C8E	0C90	
0C92	0CA8	
0CAA	0CB3	
0CB5	0CB9	
0CBC	0CC4	
0CC6	0CC8	
0CCA	0CCD	
0CD5	0CD6	
0CDE		
0CE0	0CE1	
0CE6	0CEF	
0D02	0D03	
0D05	0D0C	
0D0E	0D10	
0D12	0D28	
0D2A	0D39	
0D3E	0D43	
0D46	0D48	
0D4A	0D4D	
0D57		
0D60	0D61	
0D66	0D6F	
0D82	0D83	
0D85	0D96	
0D9A	0DB1	
0DB3	0DBB	
0DBD		
0DC0	0DC6	
0DCA		
0DCF	0DD4	
0DD6		
0DD8	0DDF	
0DF2	0DF4	
0E01	0E3A	
0E3F	0E5B	
0E81	0E82	
0E84		
0E87	0E88	
0E8A		
0E8D		
0E94	0E97	
0E99	0E9F	
0EA1	0EA3	
0EA5		
0EA7		
0EAA	0EAB	
0EAD	0EB9	
0EBB	0EBD	
0EC0	0EC4	
0EC6		
0EC8	0ECD	
0ED0	0ED9	
0EDC	0EDD	
0F00	0F47	
0F49	0F6A	
0F71	0F8B	
0F90	0F97	
0F99	0FBC	
0FBE	0FCC	
0FCF	0FD1	
1000	1021	
1023	1027	
1029	102A	
102C	1032	
1036	1039	
1040	1059	
10A0	10C5	
10D0	10FC	
1100	1159	
115F	11A2	
11A8	11F9	
1200	1248	
124A	124D	
1250	1256	
1258		
125A	125D	
1260	1288	
128A	128D	
1290	12B0	
12B2	12B5	
12B8	12BE	
12C0		
12C2	12C5	
12C8	12D6	
12D8	1310	
1312	1315	
1318	135A	
135F	137C	
1380	1399	
13A0	13F4	
1401	1676	
1681	169C	
16A0	16F0	
1700	170C	
170E	1714	
1720	1736	
1740	1753	
1760	176C	
176E	1770	
1772	1773	
1780	17DD	
17E0	17E9	
17F0	17F9	
1800	180D	
1810	1819	
1820	1877	
1880	18A9	
1900	191C	
1920	192B	
1930	193B	
1940		
1944	196D	
1970	1974	
1980	19A9	
19B0	19C9	
19D0	19D9	
19DE	1A1B	
1A1E	1A1F	
1D00	1DC3	
1E00	1E9B	
1EA0	1EF9	
1F00	1F15	
1F18	1F1D	
1F20	1F45	
1F48	1F4D	
1F50	1F57	
1F59		
1F5B		
1F5D		
1F5F	1F7D	
1F80	1FB4	
1FB6	1FC4	
1FC6	1FD3	
1FD6	1FDB	
1FDD	1FEF	
1FF2	1FF4	
1FF6	1FFE	
200B	2027	
202A	202E	
2030	205E	
2060	2063	
206A	2071	
2074	208E	
2090	2094	
20A0	20B5	
20D0	20EB	
2100	214C	
2153	2183	
2190	23DB	
2400	2426	
2440	244A	
2460	269C	
26A0	26B1	
2701	2704	
2706	2709	
270C	2727	
2729	274B	
274D		
274F	2752	
2756		
2758	275E	
2761	2794	
2798	27AF	
27B1	27BE	
27C0	27C6	
27D0	27EB	
27F0	2B13	
2C00	2C2E	
2C30	2C5E	
2C80	2CEA	
2CF9	2D25	
2D30	2D65	
2D6F		
2D80	2D96	
2DA0	2DA6	
2DA8	2DAE	
2DB0	2DB6	
2DB8	2DBE	
2DC0	2DC6	
2DC8	2DCE	
2DD0	2DD6	
2DD8	2DDE	
2E00	2E17	
2E1C	2E1D	
2E80	2E99	
2E9B	2EF3	
2F00	2FD5	
2FF0	2FFB	
3001	303F	
3041	3096	
3099	30FF	
3105	312C	
3131	318E	
3190	31B7	
31C0	31CF	
31F0	321E	
3220	3243	
3250	32FE	
3300	4DB5	
4DC0	9FBB	
A000	A48C	
A490	A4C6	
A700	A716	
A800	A82B	
AC00	D7A3	
E000	FA2D	
FA30	FA6A	
FA70	FAD9	
FB00	FB06	
FB13	FB17	
FB1D	FB36	
FB38	FB3C	
FB3E		
FB40	FB41	
FB43	FB44	
FB46	FBB1	
FBD3	FD3F	
FD50	FD8F	
FD92	FDC7	
FDF0	FDFD	
FE00	FE19	
FE20	FE23	
FE30	FE52	
FE54	FE66	
FE68	FE6B	
FE70	FE74	
FE76	FEFC	
FEFF		
FF01	FFBE	
FFC2	FFC7	
FFCA	FFCF	
FFD2	FFD7	
FFDA	FFDC	
FFE0	FFE6	
FFE8	FFEE	
FFF9	FFFD	
10000	1000B	
1000D	10026	
10028	1003A	
1003C	1003D	
1003F	1004D	
10050	1005D	
10080	100FA	
10100	10102	
10107	10133	
10137	1018A	
10300	1031E	
10320	10323	
10330	1034A	
10380	1039D	
1039F	103C3	
103C8	103D5	
10400	1049D	
104A0	104A9	
10800	10805	
10808		
1080A	10835	
10837	10838	
1083C		
1083F		
10A00	10A03	
10A05	10A06	
10A0C	10A13	
10A15	10A17	
10A19	10A33	
10A38	10A3A	
10A3F	10A47	
10A50	10A58	
1D000	1D0F5	
1D100	1D126	
1D12A	1D1DD	
1D200	1D245	
1D300	1D356	
1D400	1D454	
1D456	1D49C	
1D49E	1D49F	
1D4A2		
1D4A5	1D4A6	
1D4A9	1D4AC	
1D4AE	1D4B9	
1D4BB		
1D4BD	1D4C3	
1D4C5	1D505	
1D507	1D50A	
1D50D	1D514	
1D516	1D51C	
1D51E	1D539	
1D53B	1D53E	
1D540	1D544	
1D546		
1D54A	1D550	
1D552	1D6A5	
1D6A8	1D7C9	
1D7CE	1D7FF	
20000	2A6D6	
2F800	2FA1D	
E0001		
E0020	E007F	
E0100	E01EF	
F0000	FFFFD	
100000	10FFFD	
END
};
1;

package unicore::lib::selfloader::Grapheme;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{GraphemeLink} (and fuzzy permutations)
# 
# Meaning: Extended property 'Grapheme_Link'
#
return <<'END';
034F		Grapheme_Link
094D		Grapheme_Link
09CD		Grapheme_Link
0A4D		Grapheme_Link
0ACD		Grapheme_Link
0B4D		Grapheme_Link
0BCD		Grapheme_Link
0C4D		Grapheme_Link
0CCD		Grapheme_Link
0D4D		Grapheme_Link
0DCA		Grapheme_Link
0E3A		Grapheme_Link
1039		Grapheme_Link
17D2		Grapheme_Link
A806		Grapheme_Link
10A3F		Grapheme_Link
END
};
1;

package unicore::lib::selfloader::Grek;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Greek} (and fuzzy permutations)
# 
# Meaning: Script 'Greek'
#
return <<'END';
0374	0375	Greek
037A		Greek
0384	0386	Greek
0388	038A	Greek
038C		Greek
038E	03A1	Greek
03A3	03CE	Greek
03D0	03E1	Greek
03F0	03FF	Greek
1D26	1D2A	Greek
1D5D	1D61	Greek
1D66	1D6A	Greek
1F00	1F15	Greek
1F18	1F1D	Greek
1F20	1F45	Greek
1F48	1F4D	Greek
1F50	1F57	Greek
1F59		Greek
1F5B		Greek
1F5D		Greek
1F5F	1F7D	Greek
1F80	1FB4	Greek
1FB6	1FC4	Greek
1FC6	1FD3	Greek
1FD6	1FDB	Greek
1FDD	1FEF	Greek
1FF2	1FF4	Greek
1FF6	1FFE	Greek
2126		Greek
10140	1018A	Greek
1D200	1D245	Greek
END
};
1;

package unicore::lib::selfloader::GrLink;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Grapheme_Link'
#
return <<'END';
034F		Grapheme_Link
094D		Grapheme_Link
09CD		Grapheme_Link
0A4D		Grapheme_Link
0ACD		Grapheme_Link
0B4D		Grapheme_Link
0BCD		Grapheme_Link
0C4D		Grapheme_Link
0CCD		Grapheme_Link
0D4D		Grapheme_Link
0DCA		Grapheme_Link
0E3A		Grapheme_Link
1039		Grapheme_Link
17D2		Grapheme_Link
A806		Grapheme_Link
10A3F		Grapheme_Link
END
};
1;

package unicore::lib::selfloader::Gujr;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Gujarati} (and fuzzy permutations)
# 
# Meaning: Script 'Gujarati'
#
return <<'END';
0A81	0A83	Gujarati
0A85	0A8D	Gujarati
0A8F	0A91	Gujarati
0A93	0AA8	Gujarati
0AAA	0AB0	Gujarati
0AB2	0AB3	Gujarati
0AB5	0AB9	Gujarati
0ABC	0AC5	Gujarati
0AC7	0AC9	Gujarati
0ACB	0ACD	Gujarati
0AD0		Gujarati
0AE0	0AE3	Gujarati
0AE6	0AEF	Gujarati
0AF1		Gujarati
END
};
1;

package unicore::lib::selfloader::Guru;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Gurmukhi} (and fuzzy permutations)
# 
# Meaning: Script 'Gurmukhi'
#
return <<'END';
0A01	0A03	Gurmukhi
0A05	0A0A	Gurmukhi
0A0F	0A10	Gurmukhi
0A13	0A28	Gurmukhi
0A2A	0A30	Gurmukhi
0A32	0A33	Gurmukhi
0A35	0A36	Gurmukhi
0A38	0A39	Gurmukhi
0A3C		Gurmukhi
0A3E	0A42	Gurmukhi
0A47	0A48	Gurmukhi
0A4B	0A4D	Gurmukhi
0A59	0A5C	Gurmukhi
0A5E		Gurmukhi
0A66	0A74	Gurmukhi
END
};
1;

package unicore::lib::selfloader::Hang;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Hangul} (and fuzzy permutations)
# 
# Meaning: Script 'Hangul'
#
return <<'END';
1100	1159	Hangul
115F	11A2	Hangul
11A8	11F9	Hangul
3131	318E	Hangul
3200	321E	Hangul
3260	327D	Hangul
AC00	D7A3	Hangul
FFA0	FFBE	Hangul
FFC2	FFC7	Hangul
FFCA	FFCF	Hangul
FFD2	FFD7	Hangul
FFDA	FFDC	Hangul
END
};
1;

package unicore::lib::selfloader::Hani;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Han} (and fuzzy permutations)
# 
# Meaning: Script 'Han'
#
return <<'END';
2E80	2E99	Han
2E9B	2EF3	Han
2F00	2FD5	Han
3005		Han
3007		Han
3021	3029	Han
3038	303B	Han
3400	4DB5	Han
4E00	9FBB	Han
F900	FA2D	Han
FA30	FA6A	Han
FA70	FAD9	Han
20000	2A6D6	Han
2F800	2FA1D	Han
END
};
1;

package unicore::lib::selfloader::Hano;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Hanunoo} (and fuzzy permutations)
# 
# Meaning: Script 'Hanunoo'
#
return <<'END';
1720	1734	Hanunoo
END
};
1;

package unicore::lib::selfloader::Hebr;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Hebrew} (and fuzzy permutations)
# 
# Meaning: Script 'Hebrew'
#
return <<'END';
0591	05B9	Hebrew
05BB	05C7	Hebrew
05D0	05EA	Hebrew
05F0	05F4	Hebrew
FB1D	FB36	Hebrew
FB38	FB3C	Hebrew
FB3E		Hebrew
FB40	FB41	Hebrew
FB43	FB44	Hebrew
FB46	FB4F	Hebrew
END
};
1;

package unicore::lib::selfloader::Hex;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Hex_Digit'
#
return <<'END';
0030	0039	Hex_Digit
0041	0046	Hex_Digit
0061	0066	Hex_Digit
FF10	FF19	Hex_Digit
FF21	FF26	Hex_Digit
FF41	FF46	Hex_Digit
END
};
1;

package unicore::lib::selfloader::HexDigit;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{HexDigit} (and fuzzy permutations)
# 
# Meaning: Extended property 'Hex_Digit'
#
return <<'END';
0030	0039	Hex_Digit
0041	0046	Hex_Digit
0061	0066	Hex_Digit
FF10	FF19	Hex_Digit
FF21	FF26	Hex_Digit
FF41	FF46	Hex_Digit
END
};
1;

package unicore::lib::selfloader::Hira;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Hiragana} (and fuzzy permutations)
# 
# Meaning: Script 'Hiragana'
#
return <<'END';
3041	3096	Hiragana
309D	309F	Hiragana
END
};
1;

package unicore::lib::selfloader::Hyphen;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Hyphen'
#
return <<'END';
002D		Hyphen
00AD		Hyphen
058A		Hyphen
1806		Hyphen
2010	2011	Hyphen
2E17		Hyphen
30FB		Hyphen
FE63		Hyphen
FF0D		Hyphen
FF65		Hyphen
END
};
1;

package unicore::lib::selfloader::Hyphen2;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Hyphen} (and fuzzy permutations)
# 
# Meaning: Extended property 'Hyphen'
#
return <<'END';
002D		Hyphen
00AD		Hyphen
058A		Hyphen
1806		Hyphen
2010	2011	Hyphen
2E17		Hyphen
30FB		Hyphen
FE63		Hyphen
FF0D		Hyphen
FF65		Hyphen
END
};
1;

package unicore::lib::selfloader::IdContin;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{IdContinue} (and fuzzy permutations)
# 
# Meaning: [\p{ID_Start}\p{Mn}\p{Mc}\p{Nd}\p{Pc}]
#
return <<'END';
0030	0039	
0041	005A	
005F		
0061	007A	
00AA		
00B5		
00BA		
00C0	00D6	
00D8	00F6	
00F8	0241	
0250	02C1	
02C6	02D1	
02E0	02E4	
02EE		
0300	036F	
037A		
0386		
0388	038A	
038C		
038E	03A1	
03A3	03CE	
03D0	03F5	
03F7	0481	
0483	0486	
048A	04CE	
04D0	04F9	
0500	050F	
0531	0556	
0559		
0561	0587	
0591	05B9	
05BB	05BD	
05BF		
05C1	05C2	
05C4	05C5	
05C7		
05D0	05EA	
05F0	05F2	
0610	0615	
0621	063A	
0640	065E	
0660	0669	
066E	06D3	
06D5	06DC	
06DF	06E8	
06EA	06FC	
06FF		
0710	074A	
074D	076D	
0780	07B1	
0901	0939	
093C	094D	
0950	0954	
0958	0963	
0966	096F	
097D		
0981	0983	
0985	098C	
098F	0990	
0993	09A8	
09AA	09B0	
09B2		
09B6	09B9	
09BC	09C4	
09C7	09C8	
09CB	09CE	
09D7		
09DC	09DD	
09DF	09E3	
09E6	09F1	
0A01	0A03	
0A05	0A0A	
0A0F	0A10	
0A13	0A28	
0A2A	0A30	
0A32	0A33	
0A35	0A36	
0A38	0A39	
0A3C		
0A3E	0A42	
0A47	0A48	
0A4B	0A4D	
0A59	0A5C	
0A5E		
0A66	0A74	
0A81	0A83	
0A85	0A8D	
0A8F	0A91	
0A93	0AA8	
0AAA	0AB0	
0AB2	0AB3	
0AB5	0AB9	
0ABC	0AC5	
0AC7	0AC9	
0ACB	0ACD	
0AD0		
0AE0	0AE3	
0AE6	0AEF	
0B01	0B03	
0B05	0B0C	
0B0F	0B10	
0B13	0B28	
0B2A	0B30	
0B32	0B33	
0B35	0B39	
0B3C	0B43	
0B47	0B48	
0B4B	0B4D	
0B56	0B57	
0B5C	0B5D	
0B5F	0B61	
0B66	0B6F	
0B71		
0B82	0B83	
0B85	0B8A	
0B8E	0B90	
0B92	0B95	
0B99	0B9A	
0B9C		
0B9E	0B9F	
0BA3	0BA4	
0BA8	0BAA	
0BAE	0BB9	
0BBE	0BC2	
0BC6	0BC8	
0BCA	0BCD	
0BD7		
0BE6	0BEF	
0C01	0C03	
0C05	0C0C	
0C0E	0C10	
0C12	0C28	
0C2A	0C33	
0C35	0C39	
0C3E	0C44	
0C46	0C48	
0C4A	0C4D	
0C55	0C56	
0C60	0C61	
0C66	0C6F	
0C82	0C83	
0C85	0C8C	
0C8E	0C90	
0C92	0CA8	
0CAA	0CB3	
0CB5	0CB9	
0CBC	0CC4	
0CC6	0CC8	
0CCA	0CCD	
0CD5	0CD6	
0CDE		
0CE0	0CE1	
0CE6	0CEF	
0D02	0D03	
0D05	0D0C	
0D0E	0D10	
0D12	0D28	
0D2A	0D39	
0D3E	0D43	
0D46	0D48	
0D4A	0D4D	
0D57		
0D60	0D61	
0D66	0D6F	
0D82	0D83	
0D85	0D96	
0D9A	0DB1	
0DB3	0DBB	
0DBD		
0DC0	0DC6	
0DCA		
0DCF	0DD4	
0DD6		
0DD8	0DDF	
0DF2	0DF3	
0E01	0E3A	
0E40	0E4E	
0E50	0E59	
0E81	0E82	
0E84		
0E87	0E88	
0E8A		
0E8D		
0E94	0E97	
0E99	0E9F	
0EA1	0EA3	
0EA5		
0EA7		
0EAA	0EAB	
0EAD	0EB9	
0EBB	0EBD	
0EC0	0EC4	
0EC6		
0EC8	0ECD	
0ED0	0ED9	
0EDC	0EDD	
0F00		
0F18	0F19	
0F20	0F29	
0F35		
0F37		
0F39		
0F3E	0F47	
0F49	0F6A	
0F71	0F84	
0F86	0F8B	
0F90	0F97	
0F99	0FBC	
0FC6		
1000	1021	
1023	1027	
1029	102A	
102C	1032	
1036	1039	
1040	1049	
1050	1059	
10A0	10C5	
10D0	10FA	
10FC		
1100	1159	
115F	11A2	
11A8	11F9	
1200	1248	
124A	124D	
1250	1256	
1258		
125A	125D	
1260	1288	
128A	128D	
1290	12B0	
12B2	12B5	
12B8	12BE	
12C0		
12C2	12C5	
12C8	12D6	
12D8	1310	
1312	1315	
1318	135A	
135F		
1380	138F	
13A0	13F4	
1401	166C	
166F	1676	
1681	169A	
16A0	16EA	
16EE	16F0	
1700	170C	
170E	1714	
1720	1734	
1740	1753	
1760	176C	
176E	1770	
1772	1773	
1780	17B3	
17B6	17D3	
17D7		
17DC	17DD	
17E0	17E9	
180B	180D	
1810	1819	
1820	1877	
1880	18A9	
1900	191C	
1920	192B	
1930	193B	
1946	196D	
1970	1974	
1980	19A9	
19B0	19C9	
19D0	19D9	
1A00	1A1B	
1D00	1DC3	
1E00	1E9B	
1EA0	1EF9	
1F00	1F15	
1F18	1F1D	
1F20	1F45	
1F48	1F4D	
1F50	1F57	
1F59		
1F5B		
1F5D		
1F5F	1F7D	
1F80	1FB4	
1FB6	1FBC	
1FBE		
1FC2	1FC4	
1FC6	1FCC	
1FD0	1FD3	
1FD6	1FDB	
1FE0	1FEC	
1FF2	1FF4	
1FF6	1FFC	
203F	2040	
2054		
2071		
207F		
2090	2094	
20D0	20DC	
20E1		
20E5	20EB	
2102		
2107		
210A	2113	
2115		
2119	211D	
2124		
2126		
2128		
212A	212D	
212F	2131	
2133	2139	
213C	213F	
2145	2149	
2160	2183	
2C00	2C2E	
2C30	2C5E	
2C80	2CE4	
2D00	2D25	
2D30	2D65	
2D6F		
2D80	2D96	
2DA0	2DA6	
2DA8	2DAE	
2DB0	2DB6	
2DB8	2DBE	
2DC0	2DC6	
2DC8	2DCE	
2DD0	2DD6	
2DD8	2DDE	
3005	3007	
3021	302F	
3031	3035	
3038	303C	
3041	3096	
3099	309A	
309D	309F	
30A1	30FA	
30FC	30FF	
3105	312C	
3131	318E	
31A0	31B7	
31F0	31FF	
3400	4DB5	
4E00	9FBB	
A000	A48C	
A800	A827	
AC00	D7A3	
F900	FA2D	
FA30	FA6A	
FA70	FAD9	
FB00	FB06	
FB13	FB17	
FB1D	FB28	
FB2A	FB36	
FB38	FB3C	
FB3E		
FB40	FB41	
FB43	FB44	
FB46	FBB1	
FBD3	FD3D	
FD50	FD8F	
FD92	FDC7	
FDF0	FDFB	
FE00	FE0F	
FE20	FE23	
FE33	FE34	
FE4D	FE4F	
FE70	FE74	
FE76	FEFC	
FF10	FF19	
FF21	FF3A	
FF3F		
FF41	FF5A	
FF66	FFBE	
FFC2	FFC7	
FFCA	FFCF	
FFD2	FFD7	
FFDA	FFDC	
10000	1000B	
1000D	10026	
10028	1003A	
1003C	1003D	
1003F	1004D	
10050	1005D	
10080	100FA	
10140	10174	
10300	1031E	
10330	1034A	
10380	1039D	
103A0	103C3	
103C8	103CF	
103D1	103D5	
10400	1049D	
104A0	104A9	
10800	10805	
10808		
1080A	10835	
10837	10838	
1083C		
1083F		
10A00	10A03	
10A05	10A06	
10A0C	10A13	
10A15	10A17	
10A19	10A33	
10A38	10A3A	
10A3F		
1D165	1D169	
1D16D	1D172	
1D17B	1D182	
1D185	1D18B	
1D1AA	1D1AD	
1D242	1D244	
1D400	1D454	
1D456	1D49C	
1D49E	1D49F	
1D4A2		
1D4A5	1D4A6	
1D4A9	1D4AC	
1D4AE	1D4B9	
1D4BB		
1D4BD	1D4C3	
1D4C5	1D505	
1D507	1D50A	
1D50D	1D514	
1D516	1D51C	
1D51E	1D539	
1D53B	1D53E	
1D540	1D544	
1D546		
1D54A	1D550	
1D552	1D6A5	
1D6A8	1D6C0	
1D6C2	1D6DA	
1D6DC	1D6FA	
1D6FC	1D714	
1D716	1D734	
1D736	1D74E	
1D750	1D76E	
1D770	1D788	
1D78A	1D7A8	
1D7AA	1D7C2	
1D7C4	1D7C9	
1D7CE	1D7FF	
20000	2A6D6	
2F800	2FA1D	
E0100	E01EF	
END
};
1;

package unicore::lib::selfloader::Ideo;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Ideographic'
#
return <<'END';
3006	3007	Ideographic
3021	3029	Ideographic
3038	303A	Ideographic
3400	4DB5	Ideographic
4E00	9FBB	Ideographic
F900	FA2D	Ideographic
FA70	FAD9	Ideographic
20000	2A6D6	Ideographic
2F800	2FA1D	Ideographic
END
};
1;

package unicore::lib::selfloader::Ideograp;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Ideographic} (and fuzzy permutations)
# 
# Meaning: Extended property 'Ideographic'
#
return <<'END';
3006	3007	Ideographic
3021	3029	Ideographic
3038	303A	Ideographic
3400	4DB5	Ideographic
4E00	9FBB	Ideographic
F900	FA2D	Ideographic
FA70	FAD9	Ideographic
20000	2A6D6	Ideographic
2F800	2FA1D	Ideographic
END
};
1;

package unicore::lib::selfloader::IDSB;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'IDS_Binary_Operator'
#
return <<'END';
2FF0	2FF1	IDS_Binary_Operator
2FF4	2FFB	IDS_Binary_Operator
END
};
1;

package unicore::lib::selfloader::IdsBinar;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{IdsBinaryOperator} (and fuzzy permutations)
# 
# Meaning: Extended property 'IDS_Binary_Operator'
#
return <<'END';
2FF0	2FF1	IDS_Binary_Operator
2FF4	2FFB	IDS_Binary_Operator
END
};
1;

package unicore::lib::selfloader::IDST;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'IDS_Trinary_Operator'
#
return <<'END';
2FF2	2FF3	IDS_Trinary_Operator
END
};
1;

package unicore::lib::selfloader::IdStart;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{IdStart} (and fuzzy permutations)
# 
# Meaning: [\p{Ll}\p{Lu}\p{Lt}\p{Lm}\p{Lo}\p{Nl}]
#
return <<'END';
0041	005A	
0061	007A	
00AA		
00B5		
00BA		
00C0	00D6	
00D8	00F6	
00F8	0241	
0250	02C1	
02C6	02D1	
02E0	02E4	
02EE		
037A		
0386		
0388	038A	
038C		
038E	03A1	
03A3	03CE	
03D0	03F5	
03F7	0481	
048A	04CE	
04D0	04F9	
0500	050F	
0531	0556	
0559		
0561	0587	
05D0	05EA	
05F0	05F2	
0621	063A	
0640	064A	
066E	066F	
0671	06D3	
06D5		
06E5	06E6	
06EE	06EF	
06FA	06FC	
06FF		
0710		
0712	072F	
074D	076D	
0780	07A5	
07B1		
0904	0939	
093D		
0950		
0958	0961	
097D		
0985	098C	
098F	0990	
0993	09A8	
09AA	09B0	
09B2		
09B6	09B9	
09BD		
09CE		
09DC	09DD	
09DF	09E1	
09F0	09F1	
0A05	0A0A	
0A0F	0A10	
0A13	0A28	
0A2A	0A30	
0A32	0A33	
0A35	0A36	
0A38	0A39	
0A59	0A5C	
0A5E		
0A72	0A74	
0A85	0A8D	
0A8F	0A91	
0A93	0AA8	
0AAA	0AB0	
0AB2	0AB3	
0AB5	0AB9	
0ABD		
0AD0		
0AE0	0AE1	
0B05	0B0C	
0B0F	0B10	
0B13	0B28	
0B2A	0B30	
0B32	0B33	
0B35	0B39	
0B3D		
0B5C	0B5D	
0B5F	0B61	
0B71		
0B83		
0B85	0B8A	
0B8E	0B90	
0B92	0B95	
0B99	0B9A	
0B9C		
0B9E	0B9F	
0BA3	0BA4	
0BA8	0BAA	
0BAE	0BB9	
0C05	0C0C	
0C0E	0C10	
0C12	0C28	
0C2A	0C33	
0C35	0C39	
0C60	0C61	
0C85	0C8C	
0C8E	0C90	
0C92	0CA8	
0CAA	0CB3	
0CB5	0CB9	
0CBD		
0CDE		
0CE0	0CE1	
0D05	0D0C	
0D0E	0D10	
0D12	0D28	
0D2A	0D39	
0D60	0D61	
0D85	0D96	
0D9A	0DB1	
0DB3	0DBB	
0DBD		
0DC0	0DC6	
0E01	0E30	
0E32	0E33	
0E40	0E46	
0E81	0E82	
0E84		
0E87	0E88	
0E8A		
0E8D		
0E94	0E97	
0E99	0E9F	
0EA1	0EA3	
0EA5		
0EA7		
0EAA	0EAB	
0EAD	0EB0	
0EB2	0EB3	
0EBD		
0EC0	0EC4	
0EC6		
0EDC	0EDD	
0F00		
0F40	0F47	
0F49	0F6A	
0F88	0F8B	
1000	1021	
1023	1027	
1029	102A	
1050	1055	
10A0	10C5	
10D0	10FA	
10FC		
1100	1159	
115F	11A2	
11A8	11F9	
1200	1248	
124A	124D	
1250	1256	
1258		
125A	125D	
1260	1288	
128A	128D	
1290	12B0	
12B2	12B5	
12B8	12BE	
12C0		
12C2	12C5	
12C8	12D6	
12D8	1310	
1312	1315	
1318	135A	
1380	138F	
13A0	13F4	
1401	166C	
166F	1676	
1681	169A	
16A0	16EA	
16EE	16F0	
1700	170C	
170E	1711	
1720	1731	
1740	1751	
1760	176C	
176E	1770	
1780	17B3	
17D7		
17DC		
1820	1877	
1880	18A8	
1900	191C	
1950	196D	
1970	1974	
1980	19A9	
19C1	19C7	
1A00	1A16	
1D00	1DBF	
1E00	1E9B	
1EA0	1EF9	
1F00	1F15	
1F18	1F1D	
1F20	1F45	
1F48	1F4D	
1F50	1F57	
1F59		
1F5B		
1F5D		
1F5F	1F7D	
1F80	1FB4	
1FB6	1FBC	
1FBE		
1FC2	1FC4	
1FC6	1FCC	
1FD0	1FD3	
1FD6	1FDB	
1FE0	1FEC	
1FF2	1FF4	
1FF6	1FFC	
2071		
207F		
2090	2094	
2102		
2107		
210A	2113	
2115		
2119	211D	
2124		
2126		
2128		
212A	212D	
212F	2131	
2133	2139	
213C	213F	
2145	2149	
2160	2183	
2C00	2C2E	
2C30	2C5E	
2C80	2CE4	
2D00	2D25	
2D30	2D65	
2D6F		
2D80	2D96	
2DA0	2DA6	
2DA8	2DAE	
2DB0	2DB6	
2DB8	2DBE	
2DC0	2DC6	
2DC8	2DCE	
2DD0	2DD6	
2DD8	2DDE	
3005	3007	
3021	3029	
3031	3035	
3038	303C	
3041	3096	
309D	309F	
30A1	30FA	
30FC	30FF	
3105	312C	
3131	318E	
31A0	31B7	
31F0	31FF	
3400	4DB5	
4E00	9FBB	
A000	A48C	
A800	A801	
A803	A805	
A807	A80A	
A80C	A822	
AC00	D7A3	
F900	FA2D	
FA30	FA6A	
FA70	FAD9	
FB00	FB06	
FB13	FB17	
FB1D		
FB1F	FB28	
FB2A	FB36	
FB38	FB3C	
FB3E		
FB40	FB41	
FB43	FB44	
FB46	FBB1	
FBD3	FD3D	
FD50	FD8F	
FD92	FDC7	
FDF0	FDFB	
FE70	FE74	
FE76	FEFC	
FF21	FF3A	
FF41	FF5A	
FF66	FFBE	
FFC2	FFC7	
FFCA	FFCF	
FFD2	FFD7	
FFDA	FFDC	
10000	1000B	
1000D	10026	
10028	1003A	
1003C	1003D	
1003F	1004D	
10050	1005D	
10080	100FA	
10140	10174	
10300	1031E	
10330	1034A	
10380	1039D	
103A0	103C3	
103C8	103CF	
103D1	103D5	
10400	1049D	
10800	10805	
10808		
1080A	10835	
10837	10838	
1083C		
1083F		
10A00		
10A10	10A13	
10A15	10A17	
10A19	10A33	
1D400	1D454	
1D456	1D49C	
1D49E	1D49F	
1D4A2		
1D4A5	1D4A6	
1D4A9	1D4AC	
1D4AE	1D4B9	
1D4BB		
1D4BD	1D4C3	
1D4C5	1D505	
1D507	1D50A	
1D50D	1D514	
1D516	1D51C	
1D51E	1D539	
1D53B	1D53E	
1D540	1D544	
1D546		
1D54A	1D550	
1D552	1D6A5	
1D6A8	1D6C0	
1D6C2	1D6DA	
1D6DC	1D6FA	
1D6FC	1D714	
1D716	1D734	
1D736	1D74E	
1D750	1D76E	
1D770	1D788	
1D78A	1D7A8	
1D7AA	1D7C2	
1D7C4	1D7C9	
20000	2A6D6	
2F800	2FA1D	
END
};
1;

package unicore::lib::selfloader::IdsTrina;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{IdsTrinaryOperator} (and fuzzy permutations)
# 
# Meaning: Extended property 'IDS_Trinary_Operator'
#
return <<'END';
2FF2	2FF3	IDS_Trinary_Operator
END
};
1;

package unicore::lib::selfloader::InAegean;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InAegeanNumbers} (and fuzzy permutations)
# 
# Meaning: Block 'Aegean Numbers'
#
return <<'END';
10100	1013F	Aegean Numbers
END
};
1;

package unicore::lib::selfloader::InAlphab;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InAlphabeticPresentationForms} (and fuzzy permutations)
# 
# Meaning: Block 'Alphabetic Presentation Forms'
#
return <<'END';
FB00	FB4F	Alphabetic Presentation Forms
END
};
1;

package unicore::lib::selfloader::InAncie2;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InAncientGreekMusicalNotation} (and fuzzy permutations)
# 
# Meaning: Block 'Ancient Greek Musical Notation'
#
return <<'END';
1D200	1D24F	Ancient Greek Musical Notation
END
};
1;

package unicore::lib::selfloader::InAncien;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InAncientGreekNumbers} (and fuzzy permutations)
# 
# Meaning: Block 'Ancient Greek Numbers'
#
return <<'END';
10140	1018F	Ancient Greek Numbers
END
};
1;

package unicore::lib::selfloader::InArabi2;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InArabicSupplement} (and fuzzy permutations)
# 
# Meaning: Block 'Arabic Supplement'
#
return <<'END';
0750	077F	Arabic Supplement
END
};
1;

package unicore::lib::selfloader::InArabi3;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InArabicPresentationFormsB} (and fuzzy permutations)
# 
# Meaning: Block 'Arabic Presentation Forms-B'
#
return <<'END';
FE70	FEFF	Arabic Presentation Forms-B
END
};
1;

package unicore::lib::selfloader::InArabi4;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InArabicPresentationFormsA} (and fuzzy permutations)
# 
# Meaning: Block 'Arabic Presentation Forms-A'
#
return <<'END';
FB50	FDFF	Arabic Presentation Forms-A
END
};
1;

package unicore::lib::selfloader::InArabic;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InArabic} (and fuzzy permutations)
# 
# Meaning: Block 'Arabic'
#
return <<'END';
0600	06FF	Arabic
END
};
1;

package unicore::lib::selfloader::InArmeni;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InArmenian} (and fuzzy permutations)
# 
# Meaning: Block 'Armenian'
#
return <<'END';
0530	058F	Armenian
END
};
1;

package unicore::lib::selfloader::InArrows;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InArrows} (and fuzzy permutations)
# 
# Meaning: Block 'Arrows'
#
return <<'END';
2190	21FF	Arrows
END
};
1;

package unicore::lib::selfloader::InBasicL;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InBasicLatin} (and fuzzy permutations)
# 
# Meaning: Block 'Basic Latin'
#
return <<'END';
0000	007F	Basic Latin
END
};
1;

package unicore::lib::selfloader::InBengal;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InBengali} (and fuzzy permutations)
# 
# Meaning: Block 'Bengali'
#
return <<'END';
0980	09FF	Bengali
END
};
1;

package unicore::lib::selfloader::InBlockE;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InBlockElements} (and fuzzy permutations)
# 
# Meaning: Block 'Block Elements'
#
return <<'END';
2580	259F	Block Elements
END
};
1;

package unicore::lib::selfloader::InBopom2;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InBopomofoExtended} (and fuzzy permutations)
# 
# Meaning: Block 'Bopomofo Extended'
#
return <<'END';
31A0	31BF	Bopomofo Extended
END
};
1;

package unicore::lib::selfloader::InBopomo;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InBopomofo} (and fuzzy permutations)
# 
# Meaning: Block 'Bopomofo'
#
return <<'END';
3100	312F	Bopomofo
END
};
1;

package unicore::lib::selfloader::InBoxDra;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InBoxDrawing} (and fuzzy permutations)
# 
# Meaning: Block 'Box Drawing'
#
return <<'END';
2500	257F	Box Drawing
END
};
1;

package unicore::lib::selfloader::InBraill;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InBraillePatterns} (and fuzzy permutations)
# 
# Meaning: Block 'Braille Patterns'
#
return <<'END';
2800	28FF	Braille Patterns
END
};
1;

package unicore::lib::selfloader::InBugine;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InBuginese} (and fuzzy permutations)
# 
# Meaning: Block 'Buginese'
#
return <<'END';
1A00	1A1F	Buginese
END
};
1;

package unicore::lib::selfloader::InBuhid;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InBuhid} (and fuzzy permutations)
# 
# Meaning: Block 'Buhid'
#
return <<'END';
1740	175F	Buhid
END
};
1;

package unicore::lib::selfloader::InByzant;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InByzantineMusicalSymbols} (and fuzzy permutations)
# 
# Meaning: Block 'Byzantine Musical Symbols'
#
return <<'END';
1D000	1D0FF	Byzantine Musical Symbols
END
};
1;

package unicore::lib::selfloader::InCherok;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InCherokee} (and fuzzy permutations)
# 
# Meaning: Block 'Cherokee'
#
return <<'END';
13A0	13FF	Cherokee
END
};
1;

package unicore::lib::selfloader::InCjkCo2;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InCjkCompatibilityForms} (and fuzzy permutations)
# 
# Meaning: Block 'CJK Compatibility Forms'
#
return <<'END';
FE30	FE4F	CJK Compatibility Forms
END
};
1;

package unicore::lib::selfloader::InCjkCo3;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InCjkCompatibilityIdeographs} (and fuzzy permutations)
# 
# Meaning: Block 'CJK Compatibility Ideographs'
#
return <<'END';
F900	FAFF	CJK Compatibility Ideographs
END
};
1;

package unicore::lib::selfloader::InCjkCo4;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InCjkCompatibilityIdeographsSupplement} (and fuzzy permutations)
# 
# Meaning: Block 'CJK Compatibility Ideographs Supplement'
#
return <<'END';
2F800	2FA1F	CJK Compatibility Ideographs Supplement
END
};
1;

package unicore::lib::selfloader::InCjkCom;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InCjkCompatibility} (and fuzzy permutations)
# 
# Meaning: Block 'CJK Compatibility'
#
return <<'END';
3300	33FF	CJK Compatibility
END
};
1;

package unicore::lib::selfloader::InCjkRad;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InCjkRadicalsSupplement} (and fuzzy permutations)
# 
# Meaning: Block 'CJK Radicals Supplement'
#
return <<'END';
2E80	2EFF	CJK Radicals Supplement
END
};
1;

package unicore::lib::selfloader::InCjkStr;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InCjkStrokes} (and fuzzy permutations)
# 
# Meaning: Block 'CJK Strokes'
#
return <<'END';
31C0	31EF	CJK Strokes
END
};
1;

package unicore::lib::selfloader::InCjkSym;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InCjkSymbolsAndPunctuation} (and fuzzy permutations)
# 
# Meaning: Block 'CJK Symbols and Punctuation'
#
return <<'END';
3000	303F	CJK Symbols and Punctuation
END
};
1;

package unicore::lib::selfloader::InCjkUn2;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InCjkUnifiedIdeographsExtensionB} (and fuzzy permutations)
# 
# Meaning: Block 'CJK Unified Ideographs Extension B'
#
return <<'END';
20000	2A6DF	CJK Unified Ideographs Extension B
END
};
1;

package unicore::lib::selfloader::InCjkUn3;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InCjkUnifiedIdeographsExtensionA} (and fuzzy permutations)
# 
# Meaning: Block 'CJK Unified Ideographs Extension A'
#
return <<'END';
3400	4DBF	CJK Unified Ideographs Extension A
END
};
1;

package unicore::lib::selfloader::InCjkUni;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InCjkUnifiedIdeographs} (and fuzzy permutations)
# 
# Meaning: Block 'CJK Unified Ideographs'
#
return <<'END';
4E00	9FFF	CJK Unified Ideographs
END
};
1;

package unicore::lib::selfloader::InCombi2;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InCombiningDiacriticalMarks} (and fuzzy permutations)
# 
# Meaning: Block 'Combining Diacritical Marks'
#
return <<'END';
0300	036F	Combining Diacritical Marks
END
};
1;

package unicore::lib::selfloader::InCombi3;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InCombiningDiacriticalMarksSupplement} (and fuzzy permutations)
# 
# Meaning: Block 'Combining Diacritical Marks Supplement'
#
return <<'END';
1DC0	1DFF	Combining Diacritical Marks Supplement
END
};
1;

package unicore::lib::selfloader::InCombi4;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InCombiningDiacriticalMarksForSymbols} (and fuzzy permutations)
# 
# Meaning: Block 'Combining Diacritical Marks for Symbols'
#
return <<'END';
20D0	20FF	Combining Diacritical Marks for Symbols
END
};
1;

package unicore::lib::selfloader::InCombin;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InCombiningHalfMarks} (and fuzzy permutations)
# 
# Meaning: Block 'Combining Half Marks'
#
return <<'END';
FE20	FE2F	Combining Half Marks
END
};
1;

package unicore::lib::selfloader::InContro;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InControlPictures} (and fuzzy permutations)
# 
# Meaning: Block 'Control Pictures'
#
return <<'END';
2400	243F	Control Pictures
END
};
1;

package unicore::lib::selfloader::InCoptic;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InCoptic} (and fuzzy permutations)
# 
# Meaning: Block 'Coptic'
#
return <<'END';
2C80	2CFF	Coptic
END
};
1;

package unicore::lib::selfloader::InCurren;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InCurrencySymbols} (and fuzzy permutations)
# 
# Meaning: Block 'Currency Symbols'
#
return <<'END';
20A0	20CF	Currency Symbols
END
};
1;

package unicore::lib::selfloader::InCyprio;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InCypriotSyllabary} (and fuzzy permutations)
# 
# Meaning: Block 'Cypriot Syllabary'
#
return <<'END';
10800	1083F	Cypriot Syllabary
END
};
1;

package unicore::lib::selfloader::InCyril2;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InCyrillicSupplement} (and fuzzy permutations)
# 
# Meaning: Block 'Cyrillic Supplement'
#
return <<'END';
0500	052F	Cyrillic Supplement
END
};
1;

package unicore::lib::selfloader::InCyrill;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InCyrillic} (and fuzzy permutations)
# 
# Meaning: Block 'Cyrillic'
#
return <<'END';
0400	04FF	Cyrillic
END
};
1;

package unicore::lib::selfloader::InDesere;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InDeseret} (and fuzzy permutations)
# 
# Meaning: Block 'Deseret'
#
return <<'END';
10400	1044F	Deseret
END
};
1;

package unicore::lib::selfloader::InDevana;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InDevanagari} (and fuzzy permutations)
# 
# Meaning: Block 'Devanagari'
#
return <<'END';
0900	097F	Devanagari
END
};
1;

package unicore::lib::selfloader::InDingba;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InDingbats} (and fuzzy permutations)
# 
# Meaning: Block 'Dingbats'
#
return <<'END';
2700	27BF	Dingbats
END
};
1;

package unicore::lib::selfloader::InEnclo2;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InEnclosedCjkLettersAndMonths} (and fuzzy permutations)
# 
# Meaning: Block 'Enclosed CJK Letters and Months'
#
return <<'END';
3200	32FF	Enclosed CJK Letters and Months
END
};
1;

package unicore::lib::selfloader::InEnclos;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InEnclosedAlphanumerics} (and fuzzy permutations)
# 
# Meaning: Block 'Enclosed Alphanumerics'
#
return <<'END';
2460	24FF	Enclosed Alphanumerics
END
};
1;

package unicore::lib::selfloader::InEthio2;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InEthiopicExtended} (and fuzzy permutations)
# 
# Meaning: Block 'Ethiopic Extended'
#
return <<'END';
2D80	2DDF	Ethiopic Extended
END
};
1;

package unicore::lib::selfloader::InEthio3;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InEthiopicSupplement} (and fuzzy permutations)
# 
# Meaning: Block 'Ethiopic Supplement'
#
return <<'END';
1380	139F	Ethiopic Supplement
END
};
1;

package unicore::lib::selfloader::InEthiop;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InEthiopic} (and fuzzy permutations)
# 
# Meaning: Block 'Ethiopic'
#
return <<'END';
1200	137F	Ethiopic
END
};
1;

package unicore::lib::selfloader::InGenera;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InGeneralPunctuation} (and fuzzy permutations)
# 
# Meaning: Block 'General Punctuation'
#
return <<'END';
2000	206F	General Punctuation
END
};
1;

package unicore::lib::selfloader::InGeomet;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InGeometricShapes} (and fuzzy permutations)
# 
# Meaning: Block 'Geometric Shapes'
#
return <<'END';
25A0	25FF	Geometric Shapes
END
};
1;

package unicore::lib::selfloader::InGeorg2;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InGeorgianSupplement} (and fuzzy permutations)
# 
# Meaning: Block 'Georgian Supplement'
#
return <<'END';
2D00	2D2F	Georgian Supplement
END
};
1;

package unicore::lib::selfloader::InGeorgi;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InGeorgian} (and fuzzy permutations)
# 
# Meaning: Block 'Georgian'
#
return <<'END';
10A0	10FF	Georgian
END
};
1;

package unicore::lib::selfloader::InGlagol;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InGlagolitic} (and fuzzy permutations)
# 
# Meaning: Block 'Glagolitic'
#
return <<'END';
2C00	2C5F	Glagolitic
END
};
1;

package unicore::lib::selfloader::InGothic;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InGothic} (and fuzzy permutations)
# 
# Meaning: Block 'Gothic'
#
return <<'END';
10330	1034F	Gothic
END
};
1;

package unicore::lib::selfloader::InGreekA;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InGreekAndCoptic} (and fuzzy permutations)
# 
# Meaning: Block 'Greek and Coptic'
#
return <<'END';
0370	03FF	Greek and Coptic
END
};
1;

package unicore::lib::selfloader::InGreekE;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InGreekExtended} (and fuzzy permutations)
# 
# Meaning: Block 'Greek Extended'
#
return <<'END';
1F00	1FFF	Greek Extended
END
};
1;

package unicore::lib::selfloader::InGujara;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InGujarati} (and fuzzy permutations)
# 
# Meaning: Block 'Gujarati'
#
return <<'END';
0A80	0AFF	Gujarati
END
};
1;

package unicore::lib::selfloader::InGurmuk;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InGurmukhi} (and fuzzy permutations)
# 
# Meaning: Block 'Gurmukhi'
#
return <<'END';
0A00	0A7F	Gurmukhi
END
};
1;

package unicore::lib::selfloader::InHalfwi;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InHalfwidthAndFullwidthForms} (and fuzzy permutations)
# 
# Meaning: Block 'Halfwidth and Fullwidth Forms'
#
return <<'END';
FF00	FFEF	Halfwidth and Fullwidth Forms
END
};
1;

package unicore::lib::selfloader::InHangu2;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InHangulSyllables} (and fuzzy permutations)
# 
# Meaning: Block 'Hangul Syllables'
#
return <<'END';
AC00	D7AF	Hangul Syllables
END
};
1;

package unicore::lib::selfloader::InHangu3;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InHangulCompatibilityJamo} (and fuzzy permutations)
# 
# Meaning: Block 'Hangul Compatibility Jamo'
#
return <<'END';
3130	318F	Hangul Compatibility Jamo
END
};
1;

package unicore::lib::selfloader::InHangul;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InHangulJamo} (and fuzzy permutations)
# 
# Meaning: Block 'Hangul Jamo'
#
return <<'END';
1100	11FF	Hangul Jamo
END
};
1;

package unicore::lib::selfloader::InHanuno;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InHanunoo} (and fuzzy permutations)
# 
# Meaning: Block 'Hanunoo'
#
return <<'END';
1720	173F	Hanunoo
END
};
1;

package unicore::lib::selfloader::InHebrew;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InHebrew} (and fuzzy permutations)
# 
# Meaning: Block 'Hebrew'
#
return <<'END';
0590	05FF	Hebrew
END
};
1;

package unicore::lib::selfloader::InHighPr;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InHighPrivateUseSurrogates} (and fuzzy permutations)
# 
# Meaning: Block 'High Private Use Surrogates'
#
return <<'END';
DB80	DBFF	High Private Use Surrogates
END
};
1;

package unicore::lib::selfloader::InHighSu;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InHighSurrogates} (and fuzzy permutations)
# 
# Meaning: Block 'High Surrogates'
#
return <<'END';
D800	DB7F	High Surrogates
END
};
1;

package unicore::lib::selfloader::InHiraga;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InHiragana} (and fuzzy permutations)
# 
# Meaning: Block 'Hiragana'
#
return <<'END';
3040	309F	Hiragana
END
};
1;

package unicore::lib::selfloader::InIdeogr;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InIdeographicDescriptionCharacters} (and fuzzy permutations)
# 
# Meaning: Block 'Ideographic Description Characters'
#
return <<'END';
2FF0	2FFF	Ideographic Description Characters
END
};
1;

package unicore::lib::selfloader::InIpaExt;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InIpaExtensions} (and fuzzy permutations)
# 
# Meaning: Block 'IPA Extensions'
#
return <<'END';
0250	02AF	IPA Extensions
END
};
1;

package unicore::lib::selfloader::InKanbun;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InKanbun} (and fuzzy permutations)
# 
# Meaning: Block 'Kanbun'
#
return <<'END';
3190	319F	Kanbun
END
};
1;

package unicore::lib::selfloader::InKangxi;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InKangxiRadicals} (and fuzzy permutations)
# 
# Meaning: Block 'Kangxi Radicals'
#
return <<'END';
2F00	2FDF	Kangxi Radicals
END
};
1;

package unicore::lib::selfloader::InKannad;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InKannada} (and fuzzy permutations)
# 
# Meaning: Block 'Kannada'
#
return <<'END';
0C80	0CFF	Kannada
END
};
1;

package unicore::lib::selfloader::InKatak2;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InKatakanaPhoneticExtensions} (and fuzzy permutations)
# 
# Meaning: Block 'Katakana Phonetic Extensions'
#
return <<'END';
31F0	31FF	Katakana Phonetic Extensions
END
};
1;

package unicore::lib::selfloader::InKataka;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InKatakana} (and fuzzy permutations)
# 
# Meaning: Block 'Katakana'
#
return <<'END';
30A0	30FF	Katakana
END
};
1;

package unicore::lib::selfloader::InKharos;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InKharoshthi} (and fuzzy permutations)
# 
# Meaning: Block 'Kharoshthi'
#
return <<'END';
10A00	10A5F	Kharoshthi
END
};
1;

package unicore::lib::selfloader::InKhmer;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InKhmer} (and fuzzy permutations)
# 
# Meaning: Block 'Khmer'
#
return <<'END';
1780	17FF	Khmer
END
};
1;

package unicore::lib::selfloader::InKhmerS;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InKhmerSymbols} (and fuzzy permutations)
# 
# Meaning: Block 'Khmer Symbols'
#
return <<'END';
19E0	19FF	Khmer Symbols
END
};
1;

package unicore::lib::selfloader::InLao;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InLao} (and fuzzy permutations)
# 
# Meaning: Block 'Lao'
#
return <<'END';
0E80	0EFF	Lao
END
};
1;

package unicore::lib::selfloader::InLatin1;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InLatin1Supplement} (and fuzzy permutations)
# 
# Meaning: Block 'Latin-1 Supplement'
#
return <<'END';
0080	00FF	Latin-1 Supplement
END
};
1;

package unicore::lib::selfloader::InLatin2;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InLatinExtendedA} (and fuzzy permutations)
# 
# Meaning: Block 'Latin Extended-A'
#
return <<'END';
0100	017F	Latin Extended-A
END
};
1;

package unicore::lib::selfloader::InLatin3;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InLatinExtendedAdditional} (and fuzzy permutations)
# 
# Meaning: Block 'Latin Extended Additional'
#
return <<'END';
1E00	1EFF	Latin Extended Additional
END
};
1;

package unicore::lib::selfloader::InLatinE;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InLatinExtendedB} (and fuzzy permutations)
# 
# Meaning: Block 'Latin Extended-B'
#
return <<'END';
0180	024F	Latin Extended-B
END
};
1;

package unicore::lib::selfloader::InLetter;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InLetterlikeSymbols} (and fuzzy permutations)
# 
# Meaning: Block 'Letterlike Symbols'
#
return <<'END';
2100	214F	Letterlike Symbols
END
};
1;

package unicore::lib::selfloader::InLimbu;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InLimbu} (and fuzzy permutations)
# 
# Meaning: Block 'Limbu'
#
return <<'END';
1900	194F	Limbu
END
};
1;

package unicore::lib::selfloader::InLinea2;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InLinearBSyllabary} (and fuzzy permutations)
# 
# Meaning: Block 'Linear B Syllabary'
#
return <<'END';
10000	1007F	Linear B Syllabary
END
};
1;

package unicore::lib::selfloader::InLinear;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InLinearBIdeograms} (and fuzzy permutations)
# 
# Meaning: Block 'Linear B Ideograms'
#
return <<'END';
10080	100FF	Linear B Ideograms
END
};
1;

package unicore::lib::selfloader::InLowSur;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InLowSurrogates} (and fuzzy permutations)
# 
# Meaning: Block 'Low Surrogates'
#
return <<'END';
DC00	DFFF	Low Surrogates
END
};
1;

package unicore::lib::selfloader::InMalaya;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InMalayalam} (and fuzzy permutations)
# 
# Meaning: Block 'Malayalam'
#
return <<'END';
0D00	0D7F	Malayalam
END
};
1;

package unicore::lib::selfloader::InMathe2;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InMathematicalAlphanumericSymbols} (and fuzzy permutations)
# 
# Meaning: Block 'Mathematical Alphanumeric Symbols'
#
return <<'END';
1D400	1D7FF	Mathematical Alphanumeric Symbols
END
};
1;

package unicore::lib::selfloader::InMathem;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InMathematicalOperators} (and fuzzy permutations)
# 
# Meaning: Block 'Mathematical Operators'
#
return <<'END';
2200	22FF	Mathematical Operators
END
};
1;

package unicore::lib::selfloader::InMisce2;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InMiscellaneousTechnical} (and fuzzy permutations)
# 
# Meaning: Block 'Miscellaneous Technical'
#
return <<'END';
2300	23FF	Miscellaneous Technical
END
};
1;

package unicore::lib::selfloader::InMisce3;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InMiscellaneousSymbolsAndArrows} (and fuzzy permutations)
# 
# Meaning: Block 'Miscellaneous Symbols and Arrows'
#
return <<'END';
2B00	2BFF	Miscellaneous Symbols and Arrows
END
};
1;

package unicore::lib::selfloader::InMisce4;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InMiscellaneousMathematicalSymbolsA} (and fuzzy permutations)
# 
# Meaning: Block 'Miscellaneous Mathematical Symbols-A'
#
return <<'END';
27C0	27EF	Miscellaneous Mathematical Symbols-A
END
};
1;

package unicore::lib::selfloader::InMisce5;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InMiscellaneousMathematicalSymbolsB} (and fuzzy permutations)
# 
# Meaning: Block 'Miscellaneous Mathematical Symbols-B'
#
return <<'END';
2980	29FF	Miscellaneous Mathematical Symbols-B
END
};
1;

package unicore::lib::selfloader::InMiscel;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InMiscellaneousSymbols} (and fuzzy permutations)
# 
# Meaning: Block 'Miscellaneous Symbols'
#
return <<'END';
2600	26FF	Miscellaneous Symbols
END
};
1;

package unicore::lib::selfloader::InModifi;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InModifierToneLetters} (and fuzzy permutations)
# 
# Meaning: Block 'Modifier Tone Letters'
#
return <<'END';
A700	A71F	Modifier Tone Letters
END
};
1;

package unicore::lib::selfloader::InMongol;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InMongolian} (and fuzzy permutations)
# 
# Meaning: Block 'Mongolian'
#
return <<'END';
1800	18AF	Mongolian
END
};
1;

package unicore::lib::selfloader::InMusica;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InMusicalSymbols} (and fuzzy permutations)
# 
# Meaning: Block 'Musical Symbols'
#
return <<'END';
1D100	1D1FF	Musical Symbols
END
};
1;

package unicore::lib::selfloader::InMyanma;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InMyanmar} (and fuzzy permutations)
# 
# Meaning: Block 'Myanmar'
#
return <<'END';
1000	109F	Myanmar
END
};
1;

package unicore::lib::selfloader::InNewTai;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InNewTaiLue} (and fuzzy permutations)
# 
# Meaning: Block 'New Tai Lue'
#
return <<'END';
1980	19DF	New Tai Lue
END
};
1;

package unicore::lib::selfloader::InNumber;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InNumberForms} (and fuzzy permutations)
# 
# Meaning: Block 'Number Forms'
#
return <<'END';
2150	218F	Number Forms
END
};
1;

package unicore::lib::selfloader::InOgham;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InOgham} (and fuzzy permutations)
# 
# Meaning: Block 'Ogham'
#
return <<'END';
1680	169F	Ogham
END
};
1;

package unicore::lib::selfloader::InOldIta;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InOldItalic} (and fuzzy permutations)
# 
# Meaning: Block 'Old Italic'
#
return <<'END';
10300	1032F	Old Italic
END
};
1;

package unicore::lib::selfloader::InOldPer;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InOldPersian} (and fuzzy permutations)
# 
# Meaning: Block 'Old Persian'
#
return <<'END';
103A0	103DF	Old Persian
END
};
1;

package unicore::lib::selfloader::InOptica;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InOpticalCharacterRecognition} (and fuzzy permutations)
# 
# Meaning: Block 'Optical Character Recognition'
#
return <<'END';
2440	245F	Optical Character Recognition
END
};
1;

package unicore::lib::selfloader::InOriya;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InOriya} (and fuzzy permutations)
# 
# Meaning: Block 'Oriya'
#
return <<'END';
0B00	0B7F	Oriya
END
};
1;

package unicore::lib::selfloader::InOsmany;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InOsmanya} (and fuzzy permutations)
# 
# Meaning: Block 'Osmanya'
#
return <<'END';
10480	104AF	Osmanya
END
};
1;

package unicore::lib::selfloader::InPhone2;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InPhoneticExtensionsSupplement} (and fuzzy permutations)
# 
# Meaning: Block 'Phonetic Extensions Supplement'
#
return <<'END';
1D80	1DBF	Phonetic Extensions Supplement
END
};
1;

package unicore::lib::selfloader::InPhonet;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InPhoneticExtensions} (and fuzzy permutations)
# 
# Meaning: Block 'Phonetic Extensions'
#
return <<'END';
1D00	1D7F	Phonetic Extensions
END
};
1;

package unicore::lib::selfloader::InPrivat;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InPrivateUseArea} (and fuzzy permutations)
# 
# Meaning: Block 'Private Use Area'
#
return <<'END';
E000	F8FF	Private Use Area
END
};
1;

package unicore::lib::selfloader::InRunic;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InRunic} (and fuzzy permutations)
# 
# Meaning: Block 'Runic'
#
return <<'END';
16A0	16FF	Runic
END
};
1;

package unicore::lib::selfloader::InShavia;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InShavian} (and fuzzy permutations)
# 
# Meaning: Block 'Shavian'
#
return <<'END';
10450	1047F	Shavian
END
};
1;

package unicore::lib::selfloader::InSinhal;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InSinhala} (and fuzzy permutations)
# 
# Meaning: Block 'Sinhala'
#
return <<'END';
0D80	0DFF	Sinhala
END
};
1;

package unicore::lib::selfloader::InSmallF;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InSmallFormVariants} (and fuzzy permutations)
# 
# Meaning: Block 'Small Form Variants'
#
return <<'END';
FE50	FE6F	Small Form Variants
END
};
1;

package unicore::lib::selfloader::InSpacin;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InSpacingModifierLetters} (and fuzzy permutations)
# 
# Meaning: Block 'Spacing Modifier Letters'
#
return <<'END';
02B0	02FF	Spacing Modifier Letters
END
};
1;

package unicore::lib::selfloader::InSpecia;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InSpecials} (and fuzzy permutations)
# 
# Meaning: Block 'Specials'
#
return <<'END';
FFF0	FFFF	Specials
END
};
1;

package unicore::lib::selfloader::InSupers;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InSuperscriptsAndSubscripts} (and fuzzy permutations)
# 
# Meaning: Block 'Superscripts and Subscripts'
#
return <<'END';
2070	209F	Superscripts and Subscripts
END
};
1;

package unicore::lib::selfloader::InSuppl2;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InSupplementalArrowsA} (and fuzzy permutations)
# 
# Meaning: Block 'Supplemental Arrows-A'
#
return <<'END';
27F0	27FF	Supplemental Arrows-A
END
};
1;

package unicore::lib::selfloader::InSuppl3;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InSupplementalPunctuation} (and fuzzy permutations)
# 
# Meaning: Block 'Supplemental Punctuation'
#
return <<'END';
2E00	2E7F	Supplemental Punctuation
END
};
1;

package unicore::lib::selfloader::InSuppl4;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InSupplementaryPrivateUseAreaA} (and fuzzy permutations)
# 
# Meaning: Block 'Supplementary Private Use Area-A'
#
return <<'END';
F0000	FFFFF	Supplementary Private Use Area-A
END
};
1;

package unicore::lib::selfloader::InSuppl5;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InSupplementaryPrivateUseAreaB} (and fuzzy permutations)
# 
# Meaning: Block 'Supplementary Private Use Area-B'
#
return <<'END';
100000	10FFFF	Supplementary Private Use Area-B
END
};
1;

package unicore::lib::selfloader::InSuppl6;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InSupplementalMathematicalOperators} (and fuzzy permutations)
# 
# Meaning: Block 'Supplemental Mathematical Operators'
#
return <<'END';
2A00	2AFF	Supplemental Mathematical Operators
END
};
1;

package unicore::lib::selfloader::InSupple;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InSupplementalArrowsB} (and fuzzy permutations)
# 
# Meaning: Block 'Supplemental Arrows-B'
#
return <<'END';
2900	297F	Supplemental Arrows-B
END
};
1;

package unicore::lib::selfloader::InSyloti;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InSylotiNagri} (and fuzzy permutations)
# 
# Meaning: Block 'Syloti Nagri'
#
return <<'END';
A800	A82F	Syloti Nagri
END
};
1;

package unicore::lib::selfloader::InSyriac;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InSyriac} (and fuzzy permutations)
# 
# Meaning: Block 'Syriac'
#
return <<'END';
0700	074F	Syriac
END
};
1;

package unicore::lib::selfloader::InTagalo;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InTagalog} (and fuzzy permutations)
# 
# Meaning: Block 'Tagalog'
#
return <<'END';
1700	171F	Tagalog
END
};
1;

package unicore::lib::selfloader::InTagban;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InTagbanwa} (and fuzzy permutations)
# 
# Meaning: Block 'Tagbanwa'
#
return <<'END';
1760	177F	Tagbanwa
END
};
1;

package unicore::lib::selfloader::InTags;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InTags} (and fuzzy permutations)
# 
# Meaning: Block 'Tags'
#
return <<'END';
E0000	E007F	Tags
END
};
1;

package unicore::lib::selfloader::InTaiLe;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InTaiLe} (and fuzzy permutations)
# 
# Meaning: Block 'Tai Le'
#
return <<'END';
1950	197F	Tai Le
END
};
1;

package unicore::lib::selfloader::InTaiXua;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InTaiXuanJingSymbols} (and fuzzy permutations)
# 
# Meaning: Block 'Tai Xuan Jing Symbols'
#
return <<'END';
1D300	1D35F	Tai Xuan Jing Symbols
END
};
1;

package unicore::lib::selfloader::InTamil;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InTamil} (and fuzzy permutations)
# 
# Meaning: Block 'Tamil'
#
return <<'END';
0B80	0BFF	Tamil
END
};
1;

package unicore::lib::selfloader::InTelugu;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InTelugu} (and fuzzy permutations)
# 
# Meaning: Block 'Telugu'
#
return <<'END';
0C00	0C7F	Telugu
END
};
1;

package unicore::lib::selfloader::InThaana;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InThaana} (and fuzzy permutations)
# 
# Meaning: Block 'Thaana'
#
return <<'END';
0780	07BF	Thaana
END
};
1;

package unicore::lib::selfloader::InThai;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InThai} (and fuzzy permutations)
# 
# Meaning: Block 'Thai'
#
return <<'END';
0E00	0E7F	Thai
END
};
1;

package unicore::lib::selfloader::InTibeta;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InTibetan} (and fuzzy permutations)
# 
# Meaning: Block 'Tibetan'
#
return <<'END';
0F00	0FFF	Tibetan
END
};
1;

package unicore::lib::selfloader::InTifina;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InTifinagh} (and fuzzy permutations)
# 
# Meaning: Block 'Tifinagh'
#
return <<'END';
2D30	2D7F	Tifinagh
END
};
1;

package unicore::lib::selfloader::InUgarit;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InUgaritic} (and fuzzy permutations)
# 
# Meaning: Block 'Ugaritic'
#
return <<'END';
10380	1039F	Ugaritic
END
};
1;

package unicore::lib::selfloader::InUnifie;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InUnifiedCanadianAboriginalSyllabics} (and fuzzy permutations)
# 
# Meaning: Block 'Unified Canadian Aboriginal Syllabics'
#
return <<'END';
1400	167F	Unified Canadian Aboriginal Syllabics
END
};
1;

package unicore::lib::selfloader::InVaria2;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InVariationSelectorsSupplement} (and fuzzy permutations)
# 
# Meaning: Block 'Variation Selectors Supplement'
#
return <<'END';
E0100	E01EF	Variation Selectors Supplement
END
};
1;

package unicore::lib::selfloader::InVariat;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InVariationSelectors} (and fuzzy permutations)
# 
# Meaning: Block 'Variation Selectors'
#
return <<'END';
FE00	FE0F	Variation Selectors
END
};
1;

package unicore::lib::selfloader::InVertic;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InVerticalForms} (and fuzzy permutations)
# 
# Meaning: Block 'Vertical Forms'
#
return <<'END';
FE10	FE1F	Vertical Forms
END
};
1;

package unicore::lib::selfloader::InYijing;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InYijingHexagramSymbols} (and fuzzy permutations)
# 
# Meaning: Block 'Yijing Hexagram Symbols'
#
return <<'END';
4DC0	4DFF	Yijing Hexagram Symbols
END
};
1;

package unicore::lib::selfloader::InYiRadi;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InYiRadicals} (and fuzzy permutations)
# 
# Meaning: Block 'Yi Radicals'
#
return <<'END';
A490	A4CF	Yi Radicals
END
};
1;

package unicore::lib::selfloader::InYiSyll;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{InYiSyllables} (and fuzzy permutations)
# 
# Meaning: Block 'Yi Syllables'
#
return <<'END';
A000	A48F	Yi Syllables
END
};
1;

package unicore::lib::selfloader::JoinC;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Join_Control'
#
return <<'END';
200C	200D	Join_Control
END
};
1;

package unicore::lib::selfloader::JoinCont;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{JoinControl} (and fuzzy permutations)
# 
# Meaning: Extended property 'Join_Control'
#
return <<'END';
200C	200D	Join_Control
END
};
1;

package unicore::lib::selfloader::Kana;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Katakana} (and fuzzy permutations)
# 
# Meaning: Script 'Katakana'
#
return <<'END';
30A1	30FA	Katakana
30FD	30FF	Katakana
31F0	31FF	Katakana
FF66	FF6F	Katakana
FF71	FF9D	Katakana
END
};
1;

package unicore::lib::selfloader::Khar;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Kharoshthi} (and fuzzy permutations)
# 
# Meaning: Script 'Kharoshthi'
#
return <<'END';
10A00	10A03	Kharoshthi
10A05	10A06	Kharoshthi
10A0C	10A13	Kharoshthi
10A15	10A17	Kharoshthi
10A19	10A33	Kharoshthi
10A38	10A3A	Kharoshthi
10A3F	10A47	Kharoshthi
10A50	10A58	Kharoshthi
END
};
1;

package unicore::lib::selfloader::Khmr;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Khmer} (and fuzzy permutations)
# 
# Meaning: Script 'Khmer'
#
return <<'END';
1780	17DD	Khmer
17E0	17E9	Khmer
17F0	17F9	Khmer
19E0	19FF	Khmer
END
};
1;

package unicore::lib::selfloader::Knda;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Kannada} (and fuzzy permutations)
# 
# Meaning: Script 'Kannada'
#
return <<'END';
0C82	0C83	Kannada
0C85	0C8C	Kannada
0C8E	0C90	Kannada
0C92	0CA8	Kannada
0CAA	0CB3	Kannada
0CB5	0CB9	Kannada
0CBC	0CC4	Kannada
0CC6	0CC8	Kannada
0CCA	0CCD	Kannada
0CD5	0CD6	Kannada
0CDE		Kannada
0CE0	0CE1	Kannada
0CE6	0CEF	Kannada
END
};
1;

package unicore::lib::selfloader::L;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{L}
# 	\p{L} (and fuzzy permutations)
# 
# Meaning: Major Category 'L'
#
return <<'END';
0041	005A	
0061	007A	
00AA		
00B5		
00BA		
00C0	00D6	
00D8	00F6	
00F8	0241	
0250	02C1	
02C6	02D1	
02E0	02E4	
02EE		
037A		
0386		
0388	038A	
038C		
038E	03A1	
03A3	03CE	
03D0	03F5	
03F7	0481	
048A	04CE	
04D0	04F9	
0500	050F	
0531	0556	
0559		
0561	0587	
05D0	05EA	
05F0	05F2	
0621	063A	
0640	064A	
066E	066F	
0671	06D3	
06D5		
06E5	06E6	
06EE	06EF	
06FA	06FC	
06FF		
0710		
0712	072F	
074D	076D	
0780	07A5	
07B1		
0904	0939	
093D		
0950		
0958	0961	
097D		
0985	098C	
098F	0990	
0993	09A8	
09AA	09B0	
09B2		
09B6	09B9	
09BD		
09CE		
09DC	09DD	
09DF	09E1	
09F0	09F1	
0A05	0A0A	
0A0F	0A10	
0A13	0A28	
0A2A	0A30	
0A32	0A33	
0A35	0A36	
0A38	0A39	
0A59	0A5C	
0A5E		
0A72	0A74	
0A85	0A8D	
0A8F	0A91	
0A93	0AA8	
0AAA	0AB0	
0AB2	0AB3	
0AB5	0AB9	
0ABD		
0AD0		
0AE0	0AE1	
0B05	0B0C	
0B0F	0B10	
0B13	0B28	
0B2A	0B30	
0B32	0B33	
0B35	0B39	
0B3D		
0B5C	0B5D	
0B5F	0B61	
0B71		
0B83		
0B85	0B8A	
0B8E	0B90	
0B92	0B95	
0B99	0B9A	
0B9C		
0B9E	0B9F	
0BA3	0BA4	
0BA8	0BAA	
0BAE	0BB9	
0C05	0C0C	
0C0E	0C10	
0C12	0C28	
0C2A	0C33	
0C35	0C39	
0C60	0C61	
0C85	0C8C	
0C8E	0C90	
0C92	0CA8	
0CAA	0CB3	
0CB5	0CB9	
0CBD		
0CDE		
0CE0	0CE1	
0D05	0D0C	
0D0E	0D10	
0D12	0D28	
0D2A	0D39	
0D60	0D61	
0D85	0D96	
0D9A	0DB1	
0DB3	0DBB	
0DBD		
0DC0	0DC6	
0E01	0E30	
0E32	0E33	
0E40	0E46	
0E81	0E82	
0E84		
0E87	0E88	
0E8A		
0E8D		
0E94	0E97	
0E99	0E9F	
0EA1	0EA3	
0EA5		
0EA7		
0EAA	0EAB	
0EAD	0EB0	
0EB2	0EB3	
0EBD		
0EC0	0EC4	
0EC6		
0EDC	0EDD	
0F00		
0F40	0F47	
0F49	0F6A	
0F88	0F8B	
1000	1021	
1023	1027	
1029	102A	
1050	1055	
10A0	10C5	
10D0	10FA	
10FC		
1100	1159	
115F	11A2	
11A8	11F9	
1200	1248	
124A	124D	
1250	1256	
1258		
125A	125D	
1260	1288	
128A	128D	
1290	12B0	
12B2	12B5	
12B8	12BE	
12C0		
12C2	12C5	
12C8	12D6	
12D8	1310	
1312	1315	
1318	135A	
1380	138F	
13A0	13F4	
1401	166C	
166F	1676	
1681	169A	
16A0	16EA	
1700	170C	
170E	1711	
1720	1731	
1740	1751	
1760	176C	
176E	1770	
1780	17B3	
17D7		
17DC		
1820	1877	
1880	18A8	
1900	191C	
1950	196D	
1970	1974	
1980	19A9	
19C1	19C7	
1A00	1A16	
1D00	1DBF	
1E00	1E9B	
1EA0	1EF9	
1F00	1F15	
1F18	1F1D	
1F20	1F45	
1F48	1F4D	
1F50	1F57	
1F59		
1F5B		
1F5D		
1F5F	1F7D	
1F80	1FB4	
1FB6	1FBC	
1FBE		
1FC2	1FC4	
1FC6	1FCC	
1FD0	1FD3	
1FD6	1FDB	
1FE0	1FEC	
1FF2	1FF4	
1FF6	1FFC	
2071		
207F		
2090	2094	
2102		
2107		
210A	2113	
2115		
2119	211D	
2124		
2126		
2128		
212A	212D	
212F	2131	
2133	2139	
213C	213F	
2145	2149	
2C00	2C2E	
2C30	2C5E	
2C80	2CE4	
2D00	2D25	
2D30	2D65	
2D6F		
2D80	2D96	
2DA0	2DA6	
2DA8	2DAE	
2DB0	2DB6	
2DB8	2DBE	
2DC0	2DC6	
2DC8	2DCE	
2DD0	2DD6	
2DD8	2DDE	
3005	3006	
3031	3035	
303B	303C	
3041	3096	
309D	309F	
30A1	30FA	
30FC	30FF	
3105	312C	
3131	318E	
31A0	31B7	
31F0	31FF	
3400	4DB5	
4E00	9FBB	
A000	A48C	
A800	A801	
A803	A805	
A807	A80A	
A80C	A822	
AC00	D7A3	
F900	FA2D	
FA30	FA6A	
FA70	FAD9	
FB00	FB06	
FB13	FB17	
FB1D		
FB1F	FB28	
FB2A	FB36	
FB38	FB3C	
FB3E		
FB40	FB41	
FB43	FB44	
FB46	FBB1	
FBD3	FD3D	
FD50	FD8F	
FD92	FDC7	
FDF0	FDFB	
FE70	FE74	
FE76	FEFC	
FF21	FF3A	
FF41	FF5A	
FF66	FFBE	
FFC2	FFC7	
FFCA	FFCF	
FFD2	FFD7	
FFDA	FFDC	
10000	1000B	
1000D	10026	
10028	1003A	
1003C	1003D	
1003F	1004D	
10050	1005D	
10080	100FA	
10300	1031E	
10330	10349	
10380	1039D	
103A0	103C3	
103C8	103CF	
10400	1049D	
10800	10805	
10808		
1080A	10835	
10837	10838	
1083C		
1083F		
10A00		
10A10	10A13	
10A15	10A17	
10A19	10A33	
1D400	1D454	
1D456	1D49C	
1D49E	1D49F	
1D4A2		
1D4A5	1D4A6	
1D4A9	1D4AC	
1D4AE	1D4B9	
1D4BB		
1D4BD	1D4C3	
1D4C5	1D505	
1D507	1D50A	
1D50D	1D514	
1D516	1D51C	
1D51E	1D539	
1D53B	1D53E	
1D540	1D544	
1D546		
1D54A	1D550	
1D552	1D6A5	
1D6A8	1D6C0	
1D6C2	1D6DA	
1D6DC	1D6FA	
1D6FC	1D714	
1D716	1D734	
1D736	1D74E	
1D750	1D76E	
1D770	1D788	
1D78A	1D7A8	
1D7AA	1D7C2	
1D7C4	1D7C9	
20000	2A6D6	
2F800	2FA1D	
END
};
1;

package unicore::lib::selfloader::Laoo;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Lao} (and fuzzy permutations)
# 
# Meaning: Script 'Lao'
#
return <<'END';
0E81	0E82	Lao
0E84		Lao
0E87	0E88	Lao
0E8A		Lao
0E8D		Lao
0E94	0E97	Lao
0E99	0E9F	Lao
0EA1	0EA3	Lao
0EA5		Lao
0EA7		Lao
0EAA	0EAB	Lao
0EAD	0EB9	Lao
0EBB	0EBD	Lao
0EC0	0EC4	Lao
0EC6		Lao
0EC8	0ECD	Lao
0ED0	0ED9	Lao
0EDC	0EDD	Lao
END
};
1;

package unicore::lib::selfloader::Latn;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Latin} (and fuzzy permutations)
# 
# Meaning: Script 'Latin'
#
return <<'END';
0041	005A	Latin
0061	007A	Latin
00AA		Latin
00BA		Latin
00C0	00D6	Latin
00D8	00F6	Latin
00F8	0241	Latin
0250	02B8	Latin
02E0	02E4	Latin
1D00	1D25	Latin
1D2C	1D5C	Latin
1D62	1D65	Latin
1D6B	1D77	Latin
1D79	1DBF	Latin
1E00	1E9B	Latin
1EA0	1EF9	Latin
2071		Latin
207F		Latin
2090	2094	Latin
212A	212B	Latin
FB00	FB06	Latin
FF21	FF3A	Latin
FF41	FF5A	Latin
END
};
1;

package unicore::lib::selfloader::LC;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{LC}
# 	\p{LC} (and fuzzy permutations)
# 
# Meaning: [\p{Ll}\p{Lu}\p{Lt}]
#
return <<'END';
0041	005A	
0061	007A	
00AA		
00B5		
00BA		
00C0	00D6	
00D8	00F6	
00F8	01BA	
01BC	01BF	
01C4	0241	
0250	02AF	
0386		
0388	038A	
038C		
038E	03A1	
03A3	03CE	
03D0	03F5	
03F7	0481	
048A	04CE	
04D0	04F9	
0500	050F	
0531	0556	
0561	0587	
10A0	10C5	
1D00	1D2B	
1D62	1D77	
1D79	1D9A	
1E00	1E9B	
1EA0	1EF9	
1F00	1F15	
1F18	1F1D	
1F20	1F45	
1F48	1F4D	
1F50	1F57	
1F59		
1F5B		
1F5D		
1F5F	1F7D	
1F80	1FB4	
1FB6	1FBC	
1FBE		
1FC2	1FC4	
1FC6	1FCC	
1FD0	1FD3	
1FD6	1FDB	
1FE0	1FEC	
1FF2	1FF4	
1FF6	1FFC	
2071		
207F		
2102		
2107		
210A	2113	
2115		
2119	211D	
2124		
2126		
2128		
212A	212D	
212F	2131	
2133	2134	
2139		
213C	213F	
2145	2149	
2C00	2C2E	
2C30	2C5E	
2C80	2CE4	
2D00	2D25	
FB00	FB06	
FB13	FB17	
FF21	FF3A	
FF41	FF5A	
10400	1044F	
1D400	1D454	
1D456	1D49C	
1D49E	1D49F	
1D4A2		
1D4A5	1D4A6	
1D4A9	1D4AC	
1D4AE	1D4B9	
1D4BB		
1D4BD	1D4C3	
1D4C5	1D505	
1D507	1D50A	
1D50D	1D514	
1D516	1D51C	
1D51E	1D539	
1D53B	1D53E	
1D540	1D544	
1D546		
1D54A	1D550	
1D552	1D6A5	
1D6A8	1D6C0	
1D6C2	1D6DA	
1D6DC	1D6FA	
1D6FC	1D714	
1D716	1D734	
1D736	1D74E	
1D750	1D76E	
1D770	1D788	
1D78A	1D7A8	
1D7AA	1D7C2	
1D7C4	1D7C9	
END
};
1;

package unicore::lib::selfloader::Limb;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Limbu} (and fuzzy permutations)
# 
# Meaning: Script 'Limbu'
#
return <<'END';
1900	191C	Limbu
1920	192B	Limbu
1930	193B	Limbu
1940		Limbu
1944	194F	Limbu
END
};
1;

package unicore::lib::selfloader::LinearB;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{LinearB} (and fuzzy permutations)
# 
# Meaning: Script 'Linear_B'
#
return <<'END';
10000	1000B	Linear_B
1000D	10026	Linear_B
10028	1003A	Linear_B
1003C	1003D	Linear_B
1003F	1004D	Linear_B
10050	1005D	Linear_B
10080	100FA	Linear_B
END
};
1;

package unicore::lib::selfloader::Ll;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Ll}
# 	\p{Ll} (and fuzzy permutations)
# 
# Meaning: General Category 'Ll'
#
return <<'END';
0061	007A	
00AA		
00B5		
00BA		
00DF	00F6	
00F8	00FF	
0101		
0103		
0105		
0107		
0109		
010B		
010D		
010F		
0111		
0113		
0115		
0117		
0119		
011B		
011D		
011F		
0121		
0123		
0125		
0127		
0129		
012B		
012D		
012F		
0131		
0133		
0135		
0137	0138	
013A		
013C		
013E		
0140		
0142		
0144		
0146		
0148	0149	
014B		
014D		
014F		
0151		
0153		
0155		
0157		
0159		
015B		
015D		
015F		
0161		
0163		
0165		
0167		
0169		
016B		
016D		
016F		
0171		
0173		
0175		
0177		
017A		
017C		
017E	0180	
0183		
0185		
0188		
018C	018D	
0192		
0195		
0199	019B	
019E		
01A1		
01A3		
01A5		
01A8		
01AA	01AB	
01AD		
01B0		
01B4		
01B6		
01B9	01BA	
01BD	01BF	
01C6		
01C9		
01CC		
01CE		
01D0		
01D2		
01D4		
01D6		
01D8		
01DA		
01DC	01DD	
01DF		
01E1		
01E3		
01E5		
01E7		
01E9		
01EB		
01ED		
01EF	01F0	
01F3		
01F5		
01F9		
01FB		
01FD		
01FF		
0201		
0203		
0205		
0207		
0209		
020B		
020D		
020F		
0211		
0213		
0215		
0217		
0219		
021B		
021D		
021F		
0221		
0223		
0225		
0227		
0229		
022B		
022D		
022F		
0231		
0233	0239	
023C		
023F	0240	
0250	02AF	
0390		
03AC	03CE	
03D0	03D1	
03D5	03D7	
03D9		
03DB		
03DD		
03DF		
03E1		
03E3		
03E5		
03E7		
03E9		
03EB		
03ED		
03EF	03F3	
03F5		
03F8		
03FB	03FC	
0430	045F	
0461		
0463		
0465		
0467		
0469		
046B		
046D		
046F		
0471		
0473		
0475		
0477		
0479		
047B		
047D		
047F		
0481		
048B		
048D		
048F		
0491		
0493		
0495		
0497		
0499		
049B		
049D		
049F		
04A1		
04A3		
04A5		
04A7		
04A9		
04AB		
04AD		
04AF		
04B1		
04B3		
04B5		
04B7		
04B9		
04BB		
04BD		
04BF		
04C2		
04C4		
04C6		
04C8		
04CA		
04CC		
04CE		
04D1		
04D3		
04D5		
04D7		
04D9		
04DB		
04DD		
04DF		
04E1		
04E3		
04E5		
04E7		
04E9		
04EB		
04ED		
04EF		
04F1		
04F3		
04F5		
04F7		
04F9		
0501		
0503		
0505		
0507		
0509		
050B		
050D		
050F		
0561	0587	
1D00	1D2B	
1D62	1D77	
1D79	1D9A	
1E01		
1E03		
1E05		
1E07		
1E09		
1E0B		
1E0D		
1E0F		
1E11		
1E13		
1E15		
1E17		
1E19		
1E1B		
1E1D		
1E1F		
1E21		
1E23		
1E25		
1E27		
1E29		
1E2B		
1E2D		
1E2F		
1E31		
1E33		
1E35		
1E37		
1E39		
1E3B		
1E3D		
1E3F		
1E41		
1E43		
1E45		
1E47		
1E49		
1E4B		
1E4D		
1E4F		
1E51		
1E53		
1E55		
1E57		
1E59		
1E5B		
1E5D		
1E5F		
1E61		
1E63		
1E65		
1E67		
1E69		
1E6B		
1E6D		
1E6F		
1E71		
1E73		
1E75		
1E77		
1E79		
1E7B		
1E7D		
1E7F		
1E81		
1E83		
1E85		
1E87		
1E89		
1E8B		
1E8D		
1E8F		
1E91		
1E93		
1E95	1E9B	
1EA1		
1EA3		
1EA5		
1EA7		
1EA9		
1EAB		
1EAD		
1EAF		
1EB1		
1EB3		
1EB5		
1EB7		
1EB9		
1EBB		
1EBD		
1EBF		
1EC1		
1EC3		
1EC5		
1EC7		
1EC9		
1ECB		
1ECD		
1ECF		
1ED1		
1ED3		
1ED5		
1ED7		
1ED9		
1EDB		
1EDD		
1EDF		
1EE1		
1EE3		
1EE5		
1EE7		
1EE9		
1EEB		
1EED		
1EEF		
1EF1		
1EF3		
1EF5		
1EF7		
1EF9		
1F00	1F07	
1F10	1F15	
1F20	1F27	
1F30	1F37	
1F40	1F45	
1F50	1F57	
1F60	1F67	
1F70	1F7D	
1F80	1F87	
1F90	1F97	
1FA0	1FA7	
1FB0	1FB4	
1FB6	1FB7	
1FBE		
1FC2	1FC4	
1FC6	1FC7	
1FD0	1FD3	
1FD6	1FD7	
1FE0	1FE7	
1FF2	1FF4	
1FF6	1FF7	
2071		
207F		
210A		
210E	210F	
2113		
212F		
2134		
2139		
213C	213D	
2146	2149	
2C30	2C5E	
2C81		
2C83		
2C85		
2C87		
2C89		
2C8B		
2C8D		
2C8F		
2C91		
2C93		
2C95		
2C97		
2C99		
2C9B		
2C9D		
2C9F		
2CA1		
2CA3		
2CA5		
2CA7		
2CA9		
2CAB		
2CAD		
2CAF		
2CB1		
2CB3		
2CB5		
2CB7		
2CB9		
2CBB		
2CBD		
2CBF		
2CC1		
2CC3		
2CC5		
2CC7		
2CC9		
2CCB		
2CCD		
2CCF		
2CD1		
2CD3		
2CD5		
2CD7		
2CD9		
2CDB		
2CDD		
2CDF		
2CE1		
2CE3	2CE4	
2D00	2D25	
FB00	FB06	
FB13	FB17	
FF41	FF5A	
10428	1044F	
1D41A	1D433	
1D44E	1D454	
1D456	1D467	
1D482	1D49B	
1D4B6	1D4B9	
1D4BB		
1D4BD	1D4C3	
1D4C5	1D4CF	
1D4EA	1D503	
1D51E	1D537	
1D552	1D56B	
1D586	1D59F	
1D5BA	1D5D3	
1D5EE	1D607	
1D622	1D63B	
1D656	1D66F	
1D68A	1D6A5	
1D6C2	1D6DA	
1D6DC	1D6E1	
1D6FC	1D714	
1D716	1D71B	
1D736	1D74E	
1D750	1D755	
1D770	1D788	
1D78A	1D78F	
1D7AA	1D7C2	
1D7C4	1D7C9	
END
};
1;

package unicore::lib::selfloader::Lm;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Lm}
# 	\p{Lm} (and fuzzy permutations)
# 
# Meaning: General Category 'Lm'
#
return <<'END';
02B0	02C1	
02C6	02D1	
02E0	02E4	
02EE		
037A		
0559		
0640		
06E5	06E6	
0E46		
0EC6		
10FC		
17D7		
1843		
1D2C	1D61	
1D78		
1D9B	1DBF	
2090	2094	
2D6F		
3005		
3031	3035	
303B		
309D	309E	
30FC	30FE	
A015		
FF70		
FF9E	FF9F	
END
};
1;

package unicore::lib::selfloader::Lo;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Lo}
# 	\p{Lo} (and fuzzy permutations)
# 
# Meaning: General Category 'Lo'
#
return <<'END';
01BB		
01C0	01C3	
05D0	05EA	
05F0	05F2	
0621	063A	
0641	064A	
066E	066F	
0671	06D3	
06D5		
06EE	06EF	
06FA	06FC	
06FF		
0710		
0712	072F	
074D	076D	
0780	07A5	
07B1		
0904	0939	
093D		
0950		
0958	0961	
097D		
0985	098C	
098F	0990	
0993	09A8	
09AA	09B0	
09B2		
09B6	09B9	
09BD		
09CE		
09DC	09DD	
09DF	09E1	
09F0	09F1	
0A05	0A0A	
0A0F	0A10	
0A13	0A28	
0A2A	0A30	
0A32	0A33	
0A35	0A36	
0A38	0A39	
0A59	0A5C	
0A5E		
0A72	0A74	
0A85	0A8D	
0A8F	0A91	
0A93	0AA8	
0AAA	0AB0	
0AB2	0AB3	
0AB5	0AB9	
0ABD		
0AD0		
0AE0	0AE1	
0B05	0B0C	
0B0F	0B10	
0B13	0B28	
0B2A	0B30	
0B32	0B33	
0B35	0B39	
0B3D		
0B5C	0B5D	
0B5F	0B61	
0B71		
0B83		
0B85	0B8A	
0B8E	0B90	
0B92	0B95	
0B99	0B9A	
0B9C		
0B9E	0B9F	
0BA3	0BA4	
0BA8	0BAA	
0BAE	0BB9	
0C05	0C0C	
0C0E	0C10	
0C12	0C28	
0C2A	0C33	
0C35	0C39	
0C60	0C61	
0C85	0C8C	
0C8E	0C90	
0C92	0CA8	
0CAA	0CB3	
0CB5	0CB9	
0CBD		
0CDE		
0CE0	0CE1	
0D05	0D0C	
0D0E	0D10	
0D12	0D28	
0D2A	0D39	
0D60	0D61	
0D85	0D96	
0D9A	0DB1	
0DB3	0DBB	
0DBD		
0DC0	0DC6	
0E01	0E30	
0E32	0E33	
0E40	0E45	
0E81	0E82	
0E84		
0E87	0E88	
0E8A		
0E8D		
0E94	0E97	
0E99	0E9F	
0EA1	0EA3	
0EA5		
0EA7		
0EAA	0EAB	
0EAD	0EB0	
0EB2	0EB3	
0EBD		
0EC0	0EC4	
0EDC	0EDD	
0F00		
0F40	0F47	
0F49	0F6A	
0F88	0F8B	
1000	1021	
1023	1027	
1029	102A	
1050	1055	
10D0	10FA	
1100	1159	
115F	11A2	
11A8	11F9	
1200	1248	
124A	124D	
1250	1256	
1258		
125A	125D	
1260	1288	
128A	128D	
1290	12B0	
12B2	12B5	
12B8	12BE	
12C0		
12C2	12C5	
12C8	12D6	
12D8	1310	
1312	1315	
1318	135A	
1380	138F	
13A0	13F4	
1401	166C	
166F	1676	
1681	169A	
16A0	16EA	
1700	170C	
170E	1711	
1720	1731	
1740	1751	
1760	176C	
176E	1770	
1780	17B3	
17DC		
1820	1842	
1844	1877	
1880	18A8	
1900	191C	
1950	196D	
1970	1974	
1980	19A9	
19C1	19C7	
1A00	1A16	
2135	2138	
2D30	2D65	
2D80	2D96	
2DA0	2DA6	
2DA8	2DAE	
2DB0	2DB6	
2DB8	2DBE	
2DC0	2DC6	
2DC8	2DCE	
2DD0	2DD6	
2DD8	2DDE	
3006		
303C		
3041	3096	
309F		
30A1	30FA	
30FF		
3105	312C	
3131	318E	
31A0	31B7	
31F0	31FF	
3400	4DB5	
4E00	9FBB	
A000	A014	
A016	A48C	
A800	A801	
A803	A805	
A807	A80A	
A80C	A822	
AC00	D7A3	
F900	FA2D	
FA30	FA6A	
FA70	FAD9	
FB1D		
FB1F	FB28	
FB2A	FB36	
FB38	FB3C	
FB3E		
FB40	FB41	
FB43	FB44	
FB46	FBB1	
FBD3	FD3D	
FD50	FD8F	
FD92	FDC7	
FDF0	FDFB	
FE70	FE74	
FE76	FEFC	
FF66	FF6F	
FF71	FF9D	
FFA0	FFBE	
FFC2	FFC7	
FFCA	FFCF	
FFD2	FFD7	
FFDA	FFDC	
10000	1000B	
1000D	10026	
10028	1003A	
1003C	1003D	
1003F	1004D	
10050	1005D	
10080	100FA	
10300	1031E	
10330	10349	
10380	1039D	
103A0	103C3	
103C8	103CF	
10450	1049D	
10800	10805	
10808		
1080A	10835	
10837	10838	
1083C		
1083F		
10A00		
10A10	10A13	
10A15	10A17	
10A19	10A33	
20000	2A6D6	
2F800	2FA1D	
END
};
1;

package unicore::lib::selfloader::LOE;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Logical_Order_Exception'
#
return <<'END';
0E40	0E44	Logical_Order_Exception
0EC0	0EC4	Logical_Order_Exception
END
};
1;

package unicore::lib::selfloader::LogicalO;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{LogicalOrderException} (and fuzzy permutations)
# 
# Meaning: Extended property 'Logical_Order_Exception'
#
return <<'END';
0E40	0E44	Logical_Order_Exception
0EC0	0EC4	Logical_Order_Exception
END
};
1;

package unicore::lib::selfloader::Lower;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Lower}
# 
# Meaning: [[:Lower:]]
#
return <<'END';
0061	007A	
00AA		
00B5		
00BA		
00DF	00F6	
00F8	00FF	
0101		
0103		
0105		
0107		
0109		
010B		
010D		
010F		
0111		
0113		
0115		
0117		
0119		
011B		
011D		
011F		
0121		
0123		
0125		
0127		
0129		
012B		
012D		
012F		
0131		
0133		
0135		
0137	0138	
013A		
013C		
013E		
0140		
0142		
0144		
0146		
0148	0149	
014B		
014D		
014F		
0151		
0153		
0155		
0157		
0159		
015B		
015D		
015F		
0161		
0163		
0165		
0167		
0169		
016B		
016D		
016F		
0171		
0173		
0175		
0177		
017A		
017C		
017E	0180	
0183		
0185		
0188		
018C	018D	
0192		
0195		
0199	019B	
019E		
01A1		
01A3		
01A5		
01A8		
01AA	01AB	
01AD		
01B0		
01B4		
01B6		
01B9	01BA	
01BD	01BF	
01C6		
01C9		
01CC		
01CE		
01D0		
01D2		
01D4		
01D6		
01D8		
01DA		
01DC	01DD	
01DF		
01E1		
01E3		
01E5		
01E7		
01E9		
01EB		
01ED		
01EF	01F0	
01F3		
01F5		
01F9		
01FB		
01FD		
01FF		
0201		
0203		
0205		
0207		
0209		
020B		
020D		
020F		
0211		
0213		
0215		
0217		
0219		
021B		
021D		
021F		
0221		
0223		
0225		
0227		
0229		
022B		
022D		
022F		
0231		
0233	0239	
023C		
023F	0240	
0250	02AF	
0390		
03AC	03CE	
03D0	03D1	
03D5	03D7	
03D9		
03DB		
03DD		
03DF		
03E1		
03E3		
03E5		
03E7		
03E9		
03EB		
03ED		
03EF	03F3	
03F5		
03F8		
03FB	03FC	
0430	045F	
0461		
0463		
0465		
0467		
0469		
046B		
046D		
046F		
0471		
0473		
0475		
0477		
0479		
047B		
047D		
047F		
0481		
048B		
048D		
048F		
0491		
0493		
0495		
0497		
0499		
049B		
049D		
049F		
04A1		
04A3		
04A5		
04A7		
04A9		
04AB		
04AD		
04AF		
04B1		
04B3		
04B5		
04B7		
04B9		
04BB		
04BD		
04BF		
04C2		
04C4		
04C6		
04C8		
04CA		
04CC		
04CE		
04D1		
04D3		
04D5		
04D7		
04D9		
04DB		
04DD		
04DF		
04E1		
04E3		
04E5		
04E7		
04E9		
04EB		
04ED		
04EF		
04F1		
04F3		
04F5		
04F7		
04F9		
0501		
0503		
0505		
0507		
0509		
050B		
050D		
050F		
0561	0587	
1D00	1D2B	
1D62	1D77	
1D79	1D9A	
1E01		
1E03		
1E05		
1E07		
1E09		
1E0B		
1E0D		
1E0F		
1E11		
1E13		
1E15		
1E17		
1E19		
1E1B		
1E1D		
1E1F		
1E21		
1E23		
1E25		
1E27		
1E29		
1E2B		
1E2D		
1E2F		
1E31		
1E33		
1E35		
1E37		
1E39		
1E3B		
1E3D		
1E3F		
1E41		
1E43		
1E45		
1E47		
1E49		
1E4B		
1E4D		
1E4F		
1E51		
1E53		
1E55		
1E57		
1E59		
1E5B		
1E5D		
1E5F		
1E61		
1E63		
1E65		
1E67		
1E69		
1E6B		
1E6D		
1E6F		
1E71		
1E73		
1E75		
1E77		
1E79		
1E7B		
1E7D		
1E7F		
1E81		
1E83		
1E85		
1E87		
1E89		
1E8B		
1E8D		
1E8F		
1E91		
1E93		
1E95	1E9B	
1EA1		
1EA3		
1EA5		
1EA7		
1EA9		
1EAB		
1EAD		
1EAF		
1EB1		
1EB3		
1EB5		
1EB7		
1EB9		
1EBB		
1EBD		
1EBF		
1EC1		
1EC3		
1EC5		
1EC7		
1EC9		
1ECB		
1ECD		
1ECF		
1ED1		
1ED3		
1ED5		
1ED7		
1ED9		
1EDB		
1EDD		
1EDF		
1EE1		
1EE3		
1EE5		
1EE7		
1EE9		
1EEB		
1EED		
1EEF		
1EF1		
1EF3		
1EF5		
1EF7		
1EF9		
1F00	1F07	
1F10	1F15	
1F20	1F27	
1F30	1F37	
1F40	1F45	
1F50	1F57	
1F60	1F67	
1F70	1F7D	
1F80	1F87	
1F90	1F97	
1FA0	1FA7	
1FB0	1FB4	
1FB6	1FB7	
1FBE		
1FC2	1FC4	
1FC6	1FC7	
1FD0	1FD3	
1FD6	1FD7	
1FE0	1FE7	
1FF2	1FF4	
1FF6	1FF7	
2071		
207F		
210A		
210E	210F	
2113		
212F		
2134		
2139		
213C	213D	
2146	2149	
2C30	2C5E	
2C81		
2C83		
2C85		
2C87		
2C89		
2C8B		
2C8D		
2C8F		
2C91		
2C93		
2C95		
2C97		
2C99		
2C9B		
2C9D		
2C9F		
2CA1		
2CA3		
2CA5		
2CA7		
2CA9		
2CAB		
2CAD		
2CAF		
2CB1		
2CB3		
2CB5		
2CB7		
2CB9		
2CBB		
2CBD		
2CBF		
2CC1		
2CC3		
2CC5		
2CC7		
2CC9		
2CCB		
2CCD		
2CCF		
2CD1		
2CD3		
2CD5		
2CD7		
2CD9		
2CDB		
2CDD		
2CDF		
2CE1		
2CE3	2CE4	
2D00	2D25	
FB00	FB06	
FB13	FB17	
FF41	FF5A	
10428	1044F	
1D41A	1D433	
1D44E	1D454	
1D456	1D467	
1D482	1D49B	
1D4B6	1D4B9	
1D4BB		
1D4BD	1D4C3	
1D4C5	1D4CF	
1D4EA	1D503	
1D51E	1D537	
1D552	1D56B	
1D586	1D59F	
1D5BA	1D5D3	
1D5EE	1D607	
1D622	1D63B	
1D656	1D66F	
1D68A	1D6A5	
1D6C2	1D6DA	
1D6DC	1D6E1	
1D6FC	1D714	
1D716	1D71B	
1D736	1D74E	
1D750	1D755	
1D770	1D788	
1D78A	1D78F	
1D7AA	1D7C2	
1D7C4	1D7C9	
END
};
1;

package unicore::lib::selfloader::Lowercas;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Lowercase} (and fuzzy permutations)
# 
# Meaning: [\p{Ll}\p{OtherLowercase}]
#
return <<'END';
0061	007A	
00AA		
00B5		
00BA		
00DF	00F6	
00F8	00FF	
0101		
0103		
0105		
0107		
0109		
010B		
010D		
010F		
0111		
0113		
0115		
0117		
0119		
011B		
011D		
011F		
0121		
0123		
0125		
0127		
0129		
012B		
012D		
012F		
0131		
0133		
0135		
0137	0138	
013A		
013C		
013E		
0140		
0142		
0144		
0146		
0148	0149	
014B		
014D		
014F		
0151		
0153		
0155		
0157		
0159		
015B		
015D		
015F		
0161		
0163		
0165		
0167		
0169		
016B		
016D		
016F		
0171		
0173		
0175		
0177		
017A		
017C		
017E	0180	
0183		
0185		
0188		
018C	018D	
0192		
0195		
0199	019B	
019E		
01A1		
01A3		
01A5		
01A8		
01AA	01AB	
01AD		
01B0		
01B4		
01B6		
01B9	01BA	
01BD	01BF	
01C6		
01C9		
01CC		
01CE		
01D0		
01D2		
01D4		
01D6		
01D8		
01DA		
01DC	01DD	
01DF		
01E1		
01E3		
01E5		
01E7		
01E9		
01EB		
01ED		
01EF	01F0	
01F3		
01F5		
01F9		
01FB		
01FD		
01FF		
0201		
0203		
0205		
0207		
0209		
020B		
020D		
020F		
0211		
0213		
0215		
0217		
0219		
021B		
021D		
021F		
0221		
0223		
0225		
0227		
0229		
022B		
022D		
022F		
0231		
0233	0239	
023C		
023F	0240	
0250	02B8	
02C0	02C1	
02E0	02E4	
0345		
037A		
0390		
03AC	03CE	
03D0	03D1	
03D5	03D7	
03D9		
03DB		
03DD		
03DF		
03E1		
03E3		
03E5		
03E7		
03E9		
03EB		
03ED		
03EF	03F3	
03F5		
03F8		
03FB	03FC	
0430	045F	
0461		
0463		
0465		
0467		
0469		
046B		
046D		
046F		
0471		
0473		
0475		
0477		
0479		
047B		
047D		
047F		
0481		
048B		
048D		
048F		
0491		
0493		
0495		
0497		
0499		
049B		
049D		
049F		
04A1		
04A3		
04A5		
04A7		
04A9		
04AB		
04AD		
04AF		
04B1		
04B3		
04B5		
04B7		
04B9		
04BB		
04BD		
04BF		
04C2		
04C4		
04C6		
04C8		
04CA		
04CC		
04CE		
04D1		
04D3		
04D5		
04D7		
04D9		
04DB		
04DD		
04DF		
04E1		
04E3		
04E5		
04E7		
04E9		
04EB		
04ED		
04EF		
04F1		
04F3		
04F5		
04F7		
04F9		
0501		
0503		
0505		
0507		
0509		
050B		
050D		
050F		
0561	0587	
1D00	1DBF	
1E01		
1E03		
1E05		
1E07		
1E09		
1E0B		
1E0D		
1E0F		
1E11		
1E13		
1E15		
1E17		
1E19		
1E1B		
1E1D		
1E1F		
1E21		
1E23		
1E25		
1E27		
1E29		
1E2B		
1E2D		
1E2F		
1E31		
1E33		
1E35		
1E37		
1E39		
1E3B		
1E3D		
1E3F		
1E41		
1E43		
1E45		
1E47		
1E49		
1E4B		
1E4D		
1E4F		
1E51		
1E53		
1E55		
1E57		
1E59		
1E5B		
1E5D		
1E5F		
1E61		
1E63		
1E65		
1E67		
1E69		
1E6B		
1E6D		
1E6F		
1E71		
1E73		
1E75		
1E77		
1E79		
1E7B		
1E7D		
1E7F		
1E81		
1E83		
1E85		
1E87		
1E89		
1E8B		
1E8D		
1E8F		
1E91		
1E93		
1E95	1E9B	
1EA1		
1EA3		
1EA5		
1EA7		
1EA9		
1EAB		
1EAD		
1EAF		
1EB1		
1EB3		
1EB5		
1EB7		
1EB9		
1EBB		
1EBD		
1EBF		
1EC1		
1EC3		
1EC5		
1EC7		
1EC9		
1ECB		
1ECD		
1ECF		
1ED1		
1ED3		
1ED5		
1ED7		
1ED9		
1EDB		
1EDD		
1EDF		
1EE1		
1EE3		
1EE5		
1EE7		
1EE9		
1EEB		
1EED		
1EEF		
1EF1		
1EF3		
1EF5		
1EF7		
1EF9		
1F00	1F07	
1F10	1F15	
1F20	1F27	
1F30	1F37	
1F40	1F45	
1F50	1F57	
1F60	1F67	
1F70	1F7D	
1F80	1F87	
1F90	1F97	
1FA0	1FA7	
1FB0	1FB4	
1FB6	1FB7	
1FBE		
1FC2	1FC4	
1FC6	1FC7	
1FD0	1FD3	
1FD6	1FD7	
1FE0	1FE7	
1FF2	1FF4	
1FF6	1FF7	
2071		
207F		
2090	2094	
210A		
210E	210F	
2113		
212F		
2134		
2139		
213C	213D	
2146	2149	
2170	217F	
24D0	24E9	
2C30	2C5E	
2C81		
2C83		
2C85		
2C87		
2C89		
2C8B		
2C8D		
2C8F		
2C91		
2C93		
2C95		
2C97		
2C99		
2C9B		
2C9D		
2C9F		
2CA1		
2CA3		
2CA5		
2CA7		
2CA9		
2CAB		
2CAD		
2CAF		
2CB1		
2CB3		
2CB5		
2CB7		
2CB9		
2CBB		
2CBD		
2CBF		
2CC1		
2CC3		
2CC5		
2CC7		
2CC9		
2CCB		
2CCD		
2CCF		
2CD1		
2CD3		
2CD5		
2CD7		
2CD9		
2CDB		
2CDD		
2CDF		
2CE1		
2CE3	2CE4	
2D00	2D25	
FB00	FB06	
FB13	FB17	
FF41	FF5A	
10428	1044F	
1D41A	1D433	
1D44E	1D454	
1D456	1D467	
1D482	1D49B	
1D4B6	1D4B9	
1D4BB		
1D4BD	1D4C3	
1D4C5	1D4CF	
1D4EA	1D503	
1D51E	1D537	
1D552	1D56B	
1D586	1D59F	
1D5BA	1D5D3	
1D5EE	1D607	
1D622	1D63B	
1D656	1D66F	
1D68A	1D6A5	
1D6C2	1D6DA	
1D6DC	1D6E1	
1D6FC	1D714	
1D716	1D71B	
1D736	1D74E	
1D750	1D755	
1D770	1D788	
1D78A	1D78F	
1D7AA	1D7C2	
1D7C4	1D7C9	
END
};
1;

package unicore::lib::selfloader::Lt;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Lt}
# 	\p{Lt} (and fuzzy permutations)
# 
# Meaning: General Category 'Lt'
#
return <<'END';
01C5		
01C8		
01CB		
01F2		
1F88	1F8F	
1F98	1F9F	
1FA8	1FAF	
1FBC		
1FCC		
1FFC		
END
};
1;

package unicore::lib::selfloader::Lu;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Lu}
# 	\p{Lu} (and fuzzy permutations)
# 
# Meaning: General Category 'Lu'
#
return <<'END';
0041	005A	
00C0	00D6	
00D8	00DE	
0100		
0102		
0104		
0106		
0108		
010A		
010C		
010E		
0110		
0112		
0114		
0116		
0118		
011A		
011C		
011E		
0120		
0122		
0124		
0126		
0128		
012A		
012C		
012E		
0130		
0132		
0134		
0136		
0139		
013B		
013D		
013F		
0141		
0143		
0145		
0147		
014A		
014C		
014E		
0150		
0152		
0154		
0156		
0158		
015A		
015C		
015E		
0160		
0162		
0164		
0166		
0168		
016A		
016C		
016E		
0170		
0172		
0174		
0176		
0178	0179	
017B		
017D		
0181	0182	
0184		
0186	0187	
0189	018B	
018E	0191	
0193	0194	
0196	0198	
019C	019D	
019F	01A0	
01A2		
01A4		
01A6	01A7	
01A9		
01AC		
01AE	01AF	
01B1	01B3	
01B5		
01B7	01B8	
01BC		
01C4		
01C7		
01CA		
01CD		
01CF		
01D1		
01D3		
01D5		
01D7		
01D9		
01DB		
01DE		
01E0		
01E2		
01E4		
01E6		
01E8		
01EA		
01EC		
01EE		
01F1		
01F4		
01F6	01F8	
01FA		
01FC		
01FE		
0200		
0202		
0204		
0206		
0208		
020A		
020C		
020E		
0210		
0212		
0214		
0216		
0218		
021A		
021C		
021E		
0220		
0222		
0224		
0226		
0228		
022A		
022C		
022E		
0230		
0232		
023A	023B	
023D	023E	
0241		
0386		
0388	038A	
038C		
038E	038F	
0391	03A1	
03A3	03AB	
03D2	03D4	
03D8		
03DA		
03DC		
03DE		
03E0		
03E2		
03E4		
03E6		
03E8		
03EA		
03EC		
03EE		
03F4		
03F7		
03F9	03FA	
03FD	042F	
0460		
0462		
0464		
0466		
0468		
046A		
046C		
046E		
0470		
0472		
0474		
0476		
0478		
047A		
047C		
047E		
0480		
048A		
048C		
048E		
0490		
0492		
0494		
0496		
0498		
049A		
049C		
049E		
04A0		
04A2		
04A4		
04A6		
04A8		
04AA		
04AC		
04AE		
04B0		
04B2		
04B4		
04B6		
04B8		
04BA		
04BC		
04BE		
04C0	04C1	
04C3		
04C5		
04C7		
04C9		
04CB		
04CD		
04D0		
04D2		
04D4		
04D6		
04D8		
04DA		
04DC		
04DE		
04E0		
04E2		
04E4		
04E6		
04E8		
04EA		
04EC		
04EE		
04F0		
04F2		
04F4		
04F6		
04F8		
0500		
0502		
0504		
0506		
0508		
050A		
050C		
050E		
0531	0556	
10A0	10C5	
1E00		
1E02		
1E04		
1E06		
1E08		
1E0A		
1E0C		
1E0E		
1E10		
1E12		
1E14		
1E16		
1E18		
1E1A		
1E1C		
1E1E		
1E20		
1E22		
1E24		
1E26		
1E28		
1E2A		
1E2C		
1E2E		
1E30		
1E32		
1E34		
1E36		
1E38		
1E3A		
1E3C		
1E3E		
1E40		
1E42		
1E44		
1E46		
1E48		
1E4A		
1E4C		
1E4E		
1E50		
1E52		
1E54		
1E56		
1E58		
1E5A		
1E5C		
1E5E		
1E60		
1E62		
1E64		
1E66		
1E68		
1E6A		
1E6C		
1E6E		
1E70		
1E72		
1E74		
1E76		
1E78		
1E7A		
1E7C		
1E7E		
1E80		
1E82		
1E84		
1E86		
1E88		
1E8A		
1E8C		
1E8E		
1E90		
1E92		
1E94		
1EA0		
1EA2		
1EA4		
1EA6		
1EA8		
1EAA		
1EAC		
1EAE		
1EB0		
1EB2		
1EB4		
1EB6		
1EB8		
1EBA		
1EBC		
1EBE		
1EC0		
1EC2		
1EC4		
1EC6		
1EC8		
1ECA		
1ECC		
1ECE		
1ED0		
1ED2		
1ED4		
1ED6		
1ED8		
1EDA		
1EDC		
1EDE		
1EE0		
1EE2		
1EE4		
1EE6		
1EE8		
1EEA		
1EEC		
1EEE		
1EF0		
1EF2		
1EF4		
1EF6		
1EF8		
1F08	1F0F	
1F18	1F1D	
1F28	1F2F	
1F38	1F3F	
1F48	1F4D	
1F59		
1F5B		
1F5D		
1F5F		
1F68	1F6F	
1FB8	1FBB	
1FC8	1FCB	
1FD8	1FDB	
1FE8	1FEC	
1FF8	1FFB	
2102		
2107		
210B	210D	
2110	2112	
2115		
2119	211D	
2124		
2126		
2128		
212A	212D	
2130	2131	
2133		
213E	213F	
2145		
2C00	2C2E	
2C80		
2C82		
2C84		
2C86		
2C88		
2C8A		
2C8C		
2C8E		
2C90		
2C92		
2C94		
2C96		
2C98		
2C9A		
2C9C		
2C9E		
2CA0		
2CA2		
2CA4		
2CA6		
2CA8		
2CAA		
2CAC		
2CAE		
2CB0		
2CB2		
2CB4		
2CB6		
2CB8		
2CBA		
2CBC		
2CBE		
2CC0		
2CC2		
2CC4		
2CC6		
2CC8		
2CCA		
2CCC		
2CCE		
2CD0		
2CD2		
2CD4		
2CD6		
2CD8		
2CDA		
2CDC		
2CDE		
2CE0		
2CE2		
FF21	FF3A	
10400	10427	
1D400	1D419	
1D434	1D44D	
1D468	1D481	
1D49C		
1D49E	1D49F	
1D4A2		
1D4A5	1D4A6	
1D4A9	1D4AC	
1D4AE	1D4B5	
1D4D0	1D4E9	
1D504	1D505	
1D507	1D50A	
1D50D	1D514	
1D516	1D51C	
1D538	1D539	
1D53B	1D53E	
1D540	1D544	
1D546		
1D54A	1D550	
1D56C	1D585	
1D5A0	1D5B9	
1D5D4	1D5ED	
1D608	1D621	
1D63C	1D655	
1D670	1D689	
1D6A8	1D6C0	
1D6E2	1D6FA	
1D71C	1D734	
1D756	1D76E	
1D790	1D7A8	
END
};
1;

package unicore::lib::selfloader::M;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{M}
# 	\p{M} (and fuzzy permutations)
# 
# Meaning: Major Category 'M'
#
return <<'END';
0300	036F	
0483	0486	
0488	0489	
0591	05B9	
05BB	05BD	
05BF		
05C1	05C2	
05C4	05C5	
05C7		
0610	0615	
064B	065E	
0670		
06D6	06DC	
06DE	06E4	
06E7	06E8	
06EA	06ED	
0711		
0730	074A	
07A6	07B0	
0901	0903	
093C		
093E	094D	
0951	0954	
0962	0963	
0981	0983	
09BC		
09BE	09C4	
09C7	09C8	
09CB	09CD	
09D7		
09E2	09E3	
0A01	0A03	
0A3C		
0A3E	0A42	
0A47	0A48	
0A4B	0A4D	
0A70	0A71	
0A81	0A83	
0ABC		
0ABE	0AC5	
0AC7	0AC9	
0ACB	0ACD	
0AE2	0AE3	
0B01	0B03	
0B3C		
0B3E	0B43	
0B47	0B48	
0B4B	0B4D	
0B56	0B57	
0B82		
0BBE	0BC2	
0BC6	0BC8	
0BCA	0BCD	
0BD7		
0C01	0C03	
0C3E	0C44	
0C46	0C48	
0C4A	0C4D	
0C55	0C56	
0C82	0C83	
0CBC		
0CBE	0CC4	
0CC6	0CC8	
0CCA	0CCD	
0CD5	0CD6	
0D02	0D03	
0D3E	0D43	
0D46	0D48	
0D4A	0D4D	
0D57		
0D82	0D83	
0DCA		
0DCF	0DD4	
0DD6		
0DD8	0DDF	
0DF2	0DF3	
0E31		
0E34	0E3A	
0E47	0E4E	
0EB1		
0EB4	0EB9	
0EBB	0EBC	
0EC8	0ECD	
0F18	0F19	
0F35		
0F37		
0F39		
0F3E	0F3F	
0F71	0F84	
0F86	0F87	
0F90	0F97	
0F99	0FBC	
0FC6		
102C	1032	
1036	1039	
1056	1059	
135F		
1712	1714	
1732	1734	
1752	1753	
1772	1773	
17B6	17D3	
17DD		
180B	180D	
18A9		
1920	192B	
1930	193B	
19B0	19C0	
19C8	19C9	
1A17	1A1B	
1DC0	1DC3	
20D0	20EB	
302A	302F	
3099	309A	
A802		
A806		
A80B		
A823	A827	
FB1E		
FE00	FE0F	
FE20	FE23	
10A01	10A03	
10A05	10A06	
10A0C	10A0F	
10A38	10A3A	
10A3F		
1D165	1D169	
1D16D	1D172	
1D17B	1D182	
1D185	1D18B	
1D1AA	1D1AD	
1D242	1D244	
E0100	E01EF	
END
};
1;

package unicore::lib::selfloader::Math;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Math} (and fuzzy permutations)
# 
# Meaning: [\p{Sm}\p{OtherMath}]
#
return <<'END';
002B		
003C	003E	
005E		
007C		
007E		
00AC		
00B1		
00D7		
00F7		
03D0	03D2	
03D5		
03F0	03F1	
03F4	03F6	
2016		
2032	2034	
2040		
2044		
2052		
2061	2063	
207A	207E	
208A	208E	
20D0	20DC	
20E1		
20E5	20E6	
2102		
210A	2113	
2115		
2119	211D	
2124		
2128	2129	
212C	212D	
212F	2131	
2133	2138	
213C	2149	
214B		
2190	2194	
219A	219B	
21A0		
21A3		
21A6		
21AE		
21CE	21CF	
21D2		
21D4		
21F4	22FF	
2308	230B	
2320	2321	
237C		
239B	23B3	
23B7		
23D0		
25B7		
25C1		
25F8	25FF	
266F		
27C0	27C6	
27D0	27EB	
27F0	27FF	
2900	2AFF	
FB29		
FE61	FE66	
FE68		
FF0B		
FF1C	FF1E	
FF3C		
FF3E		
FF5C		
FF5E		
FFE2		
FFE9	FFEC	
1D400	1D454	
1D456	1D49C	
1D49E	1D49F	
1D4A2		
1D4A5	1D4A6	
1D4A9	1D4AC	
1D4AE	1D4B9	
1D4BB		
1D4BD	1D4C3	
1D4C5	1D505	
1D507	1D50A	
1D50D	1D514	
1D516	1D51C	
1D51E	1D539	
1D53B	1D53E	
1D540	1D544	
1D546		
1D54A	1D550	
1D552	1D6A3	
1D6A8	1D7C9	
1D7CE	1D7FF	
END
};
1;

package unicore::lib::selfloader::Mc;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Mc}
# 	\p{Mc} (and fuzzy permutations)
# 
# Meaning: General Category 'Mc'
#
return <<'END';
0903		
093E	0940	
0949	094C	
0982	0983	
09BE	09C0	
09C7	09C8	
09CB	09CC	
09D7		
0A03		
0A3E	0A40	
0A83		
0ABE	0AC0	
0AC9		
0ACB	0ACC	
0B02	0B03	
0B3E		
0B40		
0B47	0B48	
0B4B	0B4C	
0B57		
0BBE	0BBF	
0BC1	0BC2	
0BC6	0BC8	
0BCA	0BCC	
0BD7		
0C01	0C03	
0C41	0C44	
0C82	0C83	
0CBE		
0CC0	0CC4	
0CC7	0CC8	
0CCA	0CCB	
0CD5	0CD6	
0D02	0D03	
0D3E	0D40	
0D46	0D48	
0D4A	0D4C	
0D57		
0D82	0D83	
0DCF	0DD1	
0DD8	0DDF	
0DF2	0DF3	
0F3E	0F3F	
0F7F		
102C		
1031		
1038		
1056	1057	
17B6		
17BE	17C5	
17C7	17C8	
1923	1926	
1929	192B	
1930	1931	
1933	1938	
19B0	19C0	
19C8	19C9	
1A19	1A1B	
A802		
A823	A824	
A827		
1D165	1D166	
1D16D	1D172	
END
};
1;

package unicore::lib::selfloader::Me;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Me}
# 	\p{Me} (and fuzzy permutations)
# 
# Meaning: General Category 'Me'
#
return <<'END';
0488	0489	
06DE		
20DD	20E0	
20E2	20E4	
END
};
1;

package unicore::lib::selfloader::Mlym;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Malayalam} (and fuzzy permutations)
# 
# Meaning: Script 'Malayalam'
#
return <<'END';
0D02	0D03	Malayalam
0D05	0D0C	Malayalam
0D0E	0D10	Malayalam
0D12	0D28	Malayalam
0D2A	0D39	Malayalam
0D3E	0D43	Malayalam
0D46	0D48	Malayalam
0D4A	0D4D	Malayalam
0D57		Malayalam
0D60	0D61	Malayalam
0D66	0D6F	Malayalam
END
};
1;

package unicore::lib::selfloader::Mn;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Mn}
# 	\p{Mn} (and fuzzy permutations)
# 
# Meaning: General Category 'Mn'
#
return <<'END';
0300	036F	
0483	0486	
0591	05B9	
05BB	05BD	
05BF		
05C1	05C2	
05C4	05C5	
05C7		
0610	0615	
064B	065E	
0670		
06D6	06DC	
06DF	06E4	
06E7	06E8	
06EA	06ED	
0711		
0730	074A	
07A6	07B0	
0901	0902	
093C		
0941	0948	
094D		
0951	0954	
0962	0963	
0981		
09BC		
09C1	09C4	
09CD		
09E2	09E3	
0A01	0A02	
0A3C		
0A41	0A42	
0A47	0A48	
0A4B	0A4D	
0A70	0A71	
0A81	0A82	
0ABC		
0AC1	0AC5	
0AC7	0AC8	
0ACD		
0AE2	0AE3	
0B01		
0B3C		
0B3F		
0B41	0B43	
0B4D		
0B56		
0B82		
0BC0		
0BCD		
0C3E	0C40	
0C46	0C48	
0C4A	0C4D	
0C55	0C56	
0CBC		
0CBF		
0CC6		
0CCC	0CCD	
0D41	0D43	
0D4D		
0DCA		
0DD2	0DD4	
0DD6		
0E31		
0E34	0E3A	
0E47	0E4E	
0EB1		
0EB4	0EB9	
0EBB	0EBC	
0EC8	0ECD	
0F18	0F19	
0F35		
0F37		
0F39		
0F71	0F7E	
0F80	0F84	
0F86	0F87	
0F90	0F97	
0F99	0FBC	
0FC6		
102D	1030	
1032		
1036	1037	
1039		
1058	1059	
135F		
1712	1714	
1732	1734	
1752	1753	
1772	1773	
17B7	17BD	
17C6		
17C9	17D3	
17DD		
180B	180D	
18A9		
1920	1922	
1927	1928	
1932		
1939	193B	
1A17	1A18	
1DC0	1DC3	
20D0	20DC	
20E1		
20E5	20EB	
302A	302F	
3099	309A	
A806		
A80B		
A825	A826	
FB1E		
FE00	FE0F	
FE20	FE23	
10A01	10A03	
10A05	10A06	
10A0C	10A0F	
10A38	10A3A	
10A3F		
1D167	1D169	
1D17B	1D182	
1D185	1D18B	
1D1AA	1D1AD	
1D242	1D244	
E0100	E01EF	
END
};
1;

package unicore::lib::selfloader::Mong;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Mongolian} (and fuzzy permutations)
# 
# Meaning: Script 'Mongolian'
#
return <<'END';
1800	180E	Mongolian
1810	1819	Mongolian
1820	1877	Mongolian
1880	18A9	Mongolian
END
};
1;

package unicore::lib::selfloader::Mymr;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Myanmar} (and fuzzy permutations)
# 
# Meaning: Script 'Myanmar'
#
return <<'END';
1000	1021	Myanmar
1023	1027	Myanmar
1029	102A	Myanmar
102C	1032	Myanmar
1036	1039	Myanmar
1040	1059	Myanmar
END
};
1;

package unicore::lib::selfloader::N;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{N}
# 	\p{N} (and fuzzy permutations)
# 
# Meaning: Major Category 'N'
#
return <<'END';
0030	0039	
00B2	00B3	
00B9		
00BC	00BE	
0660	0669	
06F0	06F9	
0966	096F	
09E6	09EF	
09F4	09F9	
0A66	0A6F	
0AE6	0AEF	
0B66	0B6F	
0BE6	0BF2	
0C66	0C6F	
0CE6	0CEF	
0D66	0D6F	
0E50	0E59	
0ED0	0ED9	
0F20	0F33	
1040	1049	
1369	137C	
16EE	16F0	
17E0	17E9	
17F0	17F9	
1810	1819	
1946	194F	
19D0	19D9	
2070		
2074	2079	
2080	2089	
2153	2183	
2460	249B	
24EA	24FF	
2776	2793	
2CFD		
3007		
3021	3029	
3038	303A	
3192	3195	
3220	3229	
3251	325F	
3280	3289	
32B1	32BF	
FF10	FF19	
10107	10133	
10140	10178	
1018A		
10320	10323	
1034A		
103D1	103D5	
104A0	104A9	
10A40	10A47	
1D7CE	1D7FF	
END
};
1;

package unicore::lib::selfloader::NChar;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Noncharacter_Code_Point'
#
return <<'END';
FDD0	FDEF	Noncharacter_Code_Point
FFFE	FFFF	Noncharacter_Code_Point
1FFFE	1FFFF	Noncharacter_Code_Point
2FFFE	2FFFF	Noncharacter_Code_Point
3FFFE	3FFFF	Noncharacter_Code_Point
4FFFE	4FFFF	Noncharacter_Code_Point
5FFFE	5FFFF	Noncharacter_Code_Point
6FFFE	6FFFF	Noncharacter_Code_Point
7FFFE	7FFFF	Noncharacter_Code_Point
8FFFE	8FFFF	Noncharacter_Code_Point
9FFFE	9FFFF	Noncharacter_Code_Point
AFFFE	AFFFF	Noncharacter_Code_Point
BFFFE	BFFFF	Noncharacter_Code_Point
CFFFE	CFFFF	Noncharacter_Code_Point
DFFFE	DFFFF	Noncharacter_Code_Point
EFFFE	EFFFF	Noncharacter_Code_Point
FFFFE	FFFFF	Noncharacter_Code_Point
10FFFE	10FFFF	Noncharacter_Code_Point
END
};
1;

package unicore::lib::selfloader::Nd;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Nd}
# 	\p{Nd} (and fuzzy permutations)
# 
# Meaning: General Category 'Nd'
#
return <<'END';
0030	0039	
0660	0669	
06F0	06F9	
0966	096F	
09E6	09EF	
0A66	0A6F	
0AE6	0AEF	
0B66	0B6F	
0BE6	0BEF	
0C66	0C6F	
0CE6	0CEF	
0D66	0D6F	
0E50	0E59	
0ED0	0ED9	
0F20	0F29	
1040	1049	
17E0	17E9	
1810	1819	
1946	194F	
19D0	19D9	
FF10	FF19	
104A0	104A9	
1D7CE	1D7FF	
END
};
1;

package unicore::lib::selfloader::NewTaiLu;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{NewTaiLue} (and fuzzy permutations)
# 
# Meaning: Script 'New_Tai_Lue'
#
return <<'END';
1980	19A9	New_Tai_Lue
19B0	19C9	New_Tai_Lue
19D0	19D9	New_Tai_Lue
19DE	19DF	New_Tai_Lue
END
};
1;

package unicore::lib::selfloader::Nl;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Nl}
# 	\p{Nl} (and fuzzy permutations)
# 
# Meaning: General Category 'Nl'
#
return <<'END';
16EE	16F0	
2160	2183	
3007		
3021	3029	
3038	303A	
10140	10174	
1034A		
103D1	103D5	
END
};
1;

package unicore::lib::selfloader::No;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{No}
# 	\p{No} (and fuzzy permutations)
# 
# Meaning: General Category 'No'
#
return <<'END';
00B2	00B3	
00B9		
00BC	00BE	
09F4	09F9	
0BF0	0BF2	
0F2A	0F33	
1369	137C	
17F0	17F9	
2070		
2074	2079	
2080	2089	
2153	215F	
2460	249B	
24EA	24FF	
2776	2793	
2CFD		
3192	3195	
3220	3229	
3251	325F	
3280	3289	
32B1	32BF	
10107	10133	
10175	10178	
1018A		
10320	10323	
10A40	10A47	
END
};
1;

package unicore::lib::selfloader::Nonchara;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{NoncharacterCodePoint} (and fuzzy permutations)
# 
# Meaning: Extended property 'Noncharacter_Code_Point'
#
return <<'END';
FDD0	FDEF	Noncharacter_Code_Point
FFFE	FFFF	Noncharacter_Code_Point
1FFFE	1FFFF	Noncharacter_Code_Point
2FFFE	2FFFF	Noncharacter_Code_Point
3FFFE	3FFFF	Noncharacter_Code_Point
4FFFE	4FFFF	Noncharacter_Code_Point
5FFFE	5FFFF	Noncharacter_Code_Point
6FFFE	6FFFF	Noncharacter_Code_Point
7FFFE	7FFFF	Noncharacter_Code_Point
8FFFE	8FFFF	Noncharacter_Code_Point
9FFFE	9FFFF	Noncharacter_Code_Point
AFFFE	AFFFF	Noncharacter_Code_Point
BFFFE	BFFFF	Noncharacter_Code_Point
CFFFE	CFFFF	Noncharacter_Code_Point
DFFFE	DFFFF	Noncharacter_Code_Point
EFFFE	EFFFF	Noncharacter_Code_Point
FFFFE	FFFFF	Noncharacter_Code_Point
10FFFE	10FFFF	Noncharacter_Code_Point
END
};
1;

package unicore::lib::selfloader::OAlpha;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Other_Alphabetic'
#
return <<'END';
0345		Other_Alphabetic
05B0	05B9	Other_Alphabetic
05BB	05BD	Other_Alphabetic
05BF		Other_Alphabetic
05C1	05C2	Other_Alphabetic
05C4	05C5	Other_Alphabetic
05C7		Other_Alphabetic
0610	0615	Other_Alphabetic
064B	0657	Other_Alphabetic
0659	065E	Other_Alphabetic
0670		Other_Alphabetic
06D6	06DC	Other_Alphabetic
06E1	06E4	Other_Alphabetic
06E7	06E8	Other_Alphabetic
06ED		Other_Alphabetic
0711		Other_Alphabetic
0730	073F	Other_Alphabetic
07A6	07B0	Other_Alphabetic
0901	0903	Other_Alphabetic
093E	094C	Other_Alphabetic
0962	0963	Other_Alphabetic
0981	0983	Other_Alphabetic
09BE	09C4	Other_Alphabetic
09C7	09C8	Other_Alphabetic
09CB	09CC	Other_Alphabetic
09D7		Other_Alphabetic
09E2	09E3	Other_Alphabetic
0A01	0A03	Other_Alphabetic
0A3E	0A42	Other_Alphabetic
0A47	0A48	Other_Alphabetic
0A4B	0A4C	Other_Alphabetic
0A70	0A71	Other_Alphabetic
0A81	0A83	Other_Alphabetic
0ABE	0AC5	Other_Alphabetic
0AC7	0AC9	Other_Alphabetic
0ACB	0ACC	Other_Alphabetic
0AE2	0AE3	Other_Alphabetic
0B01	0B03	Other_Alphabetic
0B3E	0B43	Other_Alphabetic
0B47	0B48	Other_Alphabetic
0B4B	0B4C	Other_Alphabetic
0B56	0B57	Other_Alphabetic
0B82		Other_Alphabetic
0BBE	0BC2	Other_Alphabetic
0BC6	0BC8	Other_Alphabetic
0BCA	0BCC	Other_Alphabetic
0BD7		Other_Alphabetic
0C01	0C03	Other_Alphabetic
0C3E	0C44	Other_Alphabetic
0C46	0C48	Other_Alphabetic
0C4A	0C4C	Other_Alphabetic
0C55	0C56	Other_Alphabetic
0C82	0C83	Other_Alphabetic
0CBE	0CC4	Other_Alphabetic
0CC6	0CC8	Other_Alphabetic
0CCA	0CCC	Other_Alphabetic
0CD5	0CD6	Other_Alphabetic
0D02	0D03	Other_Alphabetic
0D3E	0D43	Other_Alphabetic
0D46	0D48	Other_Alphabetic
0D4A	0D4C	Other_Alphabetic
0D57		Other_Alphabetic
0D82	0D83	Other_Alphabetic
0DCF	0DD4	Other_Alphabetic
0DD6		Other_Alphabetic
0DD8	0DDF	Other_Alphabetic
0DF2	0DF3	Other_Alphabetic
0E31		Other_Alphabetic
0E34	0E3A	Other_Alphabetic
0E4D		Other_Alphabetic
0EB1		Other_Alphabetic
0EB4	0EB9	Other_Alphabetic
0EBB	0EBC	Other_Alphabetic
0ECD		Other_Alphabetic
0F71	0F81	Other_Alphabetic
0F90	0F97	Other_Alphabetic
0F99	0FBC	Other_Alphabetic
102C	1032	Other_Alphabetic
1036		Other_Alphabetic
1038		Other_Alphabetic
1056	1059	Other_Alphabetic
135F		Other_Alphabetic
1712	1713	Other_Alphabetic
1732	1733	Other_Alphabetic
1752	1753	Other_Alphabetic
1772	1773	Other_Alphabetic
17B6	17C8	Other_Alphabetic
18A9		Other_Alphabetic
1920	192B	Other_Alphabetic
1930	1938	Other_Alphabetic
19B0	19C0	Other_Alphabetic
19C8	19C9	Other_Alphabetic
1A17	1A1B	Other_Alphabetic
24B6	24E9	Other_Alphabetic
A823	A827	Other_Alphabetic
FB1E		Other_Alphabetic
10A01	10A03	Other_Alphabetic
10A05	10A06	Other_Alphabetic
10A0C	10A0F	Other_Alphabetic
END
};
1;

package unicore::lib::selfloader::ODI;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Other_Default_Ignorable_Code_Point'
#
return <<'END';
034F		Other_Default_Ignorable_Code_Point
115F	1160	Other_Default_Ignorable_Code_Point
2064	2069	Other_Default_Ignorable_Code_Point
3164		Other_Default_Ignorable_Code_Point
FFA0		Other_Default_Ignorable_Code_Point
FFF0	FFF8	Other_Default_Ignorable_Code_Point
E0000		Other_Default_Ignorable_Code_Point
E0002	E001F	Other_Default_Ignorable_Code_Point
E0080	E00FF	Other_Default_Ignorable_Code_Point
E01F0	E0FFF	Other_Default_Ignorable_Code_Point
END
};
1;

package unicore::lib::selfloader::Ogam;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Ogham} (and fuzzy permutations)
# 
# Meaning: Script 'Ogham'
#
return <<'END';
1680	169C	Ogham
END
};
1;

package unicore::lib::selfloader::OGrExt;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Other_Grapheme_Extend'
#
return <<'END';
09BE		Other_Grapheme_Extend
09D7		Other_Grapheme_Extend
0B3E		Other_Grapheme_Extend
0B57		Other_Grapheme_Extend
0BBE		Other_Grapheme_Extend
0BD7		Other_Grapheme_Extend
0CC2		Other_Grapheme_Extend
0CD5	0CD6	Other_Grapheme_Extend
0D3E		Other_Grapheme_Extend
0D57		Other_Grapheme_Extend
0DCF		Other_Grapheme_Extend
0DDF		Other_Grapheme_Extend
200C	200D	Other_Grapheme_Extend
1D165		Other_Grapheme_Extend
1D16E	1D172	Other_Grapheme_Extend
END
};
1;

package unicore::lib::selfloader::OIDC;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Other_ID_Continue'
#
return <<'END';
1369	1371	Other_ID_Continue
END
};
1;

package unicore::lib::selfloader::OIDS;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Other_ID_Start'
#
return <<'END';
2118		Other_ID_Start
212E		Other_ID_Start
309B	309C	Other_ID_Start
END
};
1;

package unicore::lib::selfloader::OldItali;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{OldItalic} (and fuzzy permutations)
# 
# Meaning: Script 'Old_Italic'
#
return <<'END';
10300	1031E	Old_Italic
10320	10323	Old_Italic
END
};
1;

package unicore::lib::selfloader::OldPersi;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{OldPersian} (and fuzzy permutations)
# 
# Meaning: Script 'Old_Persian'
#
return <<'END';
103A0	103C3	Old_Persian
103C8	103D5	Old_Persian
END
};
1;

package unicore::lib::selfloader::OLower;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Other_Lowercase'
#
return <<'END';
02B0	02B8	Other_Lowercase
02C0	02C1	Other_Lowercase
02E0	02E4	Other_Lowercase
0345		Other_Lowercase
037A		Other_Lowercase
1D2C	1D61	Other_Lowercase
1D78		Other_Lowercase
1D9B	1DBF	Other_Lowercase
2090	2094	Other_Lowercase
2170	217F	Other_Lowercase
24D0	24E9	Other_Lowercase
END
};
1;

package unicore::lib::selfloader::OMath;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Other_Math'
#
return <<'END';
005E		Other_Math
03D0	03D2	Other_Math
03D5		Other_Math
03F0	03F1	Other_Math
03F4	03F5	Other_Math
2016		Other_Math
2032	2034	Other_Math
2040		Other_Math
2061	2063	Other_Math
207D	207E	Other_Math
208D	208E	Other_Math
20D0	20DC	Other_Math
20E1		Other_Math
20E5	20E6	Other_Math
2102		Other_Math
210A	2113	Other_Math
2115		Other_Math
2119	211D	Other_Math
2124		Other_Math
2128	2129	Other_Math
212C	212D	Other_Math
212F	2131	Other_Math
2133	2138	Other_Math
213C	213F	Other_Math
2145	2149	Other_Math
23B7		Other_Math
23D0		Other_Math
27C5	27C6	Other_Math
27E6	27EB	Other_Math
2983	2998	Other_Math
29D8	29DB	Other_Math
29FC	29FD	Other_Math
FE61		Other_Math
FE63		Other_Math
FE68		Other_Math
FF3C		Other_Math
FF3E		Other_Math
1D400	1D454	Other_Math
1D456	1D49C	Other_Math
1D49E	1D49F	Other_Math
1D4A2		Other_Math
1D4A5	1D4A6	Other_Math
1D4A9	1D4AC	Other_Math
1D4AE	1D4B9	Other_Math
1D4BB		Other_Math
1D4BD	1D4C3	Other_Math
1D4C5	1D505	Other_Math
1D507	1D50A	Other_Math
1D50D	1D514	Other_Math
1D516	1D51C	Other_Math
1D51E	1D539	Other_Math
1D53B	1D53E	Other_Math
1D540	1D544	Other_Math
1D546		Other_Math
1D54A	1D550	Other_Math
1D552	1D6A3	Other_Math
1D6A8	1D6C0	Other_Math
1D6C2	1D6DA	Other_Math
1D6DC	1D6FA	Other_Math
1D6FC	1D714	Other_Math
1D716	1D734	Other_Math
1D736	1D74E	Other_Math
1D750	1D76E	Other_Math
1D770	1D788	Other_Math
1D78A	1D7A8	Other_Math
1D7AA	1D7C2	Other_Math
1D7C4	1D7C9	Other_Math
1D7CE	1D7FF	Other_Math
END
};
1;

package unicore::lib::selfloader::Orya;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Oriya} (and fuzzy permutations)
# 
# Meaning: Script 'Oriya'
#
return <<'END';
0B01	0B03	Oriya
0B05	0B0C	Oriya
0B0F	0B10	Oriya
0B13	0B28	Oriya
0B2A	0B30	Oriya
0B32	0B33	Oriya
0B35	0B39	Oriya
0B3C	0B43	Oriya
0B47	0B48	Oriya
0B4B	0B4D	Oriya
0B56	0B57	Oriya
0B5C	0B5D	Oriya
0B5F	0B61	Oriya
0B66	0B71	Oriya
END
};
1;

package unicore::lib::selfloader::Osma;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Osmanya} (and fuzzy permutations)
# 
# Meaning: Script 'Osmanya'
#
return <<'END';
10480	1049D	Osmanya
104A0	104A9	Osmanya
END
};
1;

package unicore::lib::selfloader::OtherAlp;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{OtherAlphabetic} (and fuzzy permutations)
# 
# Meaning: Extended property 'Other_Alphabetic'
#
return <<'END';
0345		Other_Alphabetic
05B0	05B9	Other_Alphabetic
05BB	05BD	Other_Alphabetic
05BF		Other_Alphabetic
05C1	05C2	Other_Alphabetic
05C4	05C5	Other_Alphabetic
05C7		Other_Alphabetic
0610	0615	Other_Alphabetic
064B	0657	Other_Alphabetic
0659	065E	Other_Alphabetic
0670		Other_Alphabetic
06D6	06DC	Other_Alphabetic
06E1	06E4	Other_Alphabetic
06E7	06E8	Other_Alphabetic
06ED		Other_Alphabetic
0711		Other_Alphabetic
0730	073F	Other_Alphabetic
07A6	07B0	Other_Alphabetic
0901	0903	Other_Alphabetic
093E	094C	Other_Alphabetic
0962	0963	Other_Alphabetic
0981	0983	Other_Alphabetic
09BE	09C4	Other_Alphabetic
09C7	09C8	Other_Alphabetic
09CB	09CC	Other_Alphabetic
09D7		Other_Alphabetic
09E2	09E3	Other_Alphabetic
0A01	0A03	Other_Alphabetic
0A3E	0A42	Other_Alphabetic
0A47	0A48	Other_Alphabetic
0A4B	0A4C	Other_Alphabetic
0A70	0A71	Other_Alphabetic
0A81	0A83	Other_Alphabetic
0ABE	0AC5	Other_Alphabetic
0AC7	0AC9	Other_Alphabetic
0ACB	0ACC	Other_Alphabetic
0AE2	0AE3	Other_Alphabetic
0B01	0B03	Other_Alphabetic
0B3E	0B43	Other_Alphabetic
0B47	0B48	Other_Alphabetic
0B4B	0B4C	Other_Alphabetic
0B56	0B57	Other_Alphabetic
0B82		Other_Alphabetic
0BBE	0BC2	Other_Alphabetic
0BC6	0BC8	Other_Alphabetic
0BCA	0BCC	Other_Alphabetic
0BD7		Other_Alphabetic
0C01	0C03	Other_Alphabetic
0C3E	0C44	Other_Alphabetic
0C46	0C48	Other_Alphabetic
0C4A	0C4C	Other_Alphabetic
0C55	0C56	Other_Alphabetic
0C82	0C83	Other_Alphabetic
0CBE	0CC4	Other_Alphabetic
0CC6	0CC8	Other_Alphabetic
0CCA	0CCC	Other_Alphabetic
0CD5	0CD6	Other_Alphabetic
0D02	0D03	Other_Alphabetic
0D3E	0D43	Other_Alphabetic
0D46	0D48	Other_Alphabetic
0D4A	0D4C	Other_Alphabetic
0D57		Other_Alphabetic
0D82	0D83	Other_Alphabetic
0DCF	0DD4	Other_Alphabetic
0DD6		Other_Alphabetic
0DD8	0DDF	Other_Alphabetic
0DF2	0DF3	Other_Alphabetic
0E31		Other_Alphabetic
0E34	0E3A	Other_Alphabetic
0E4D		Other_Alphabetic
0EB1		Other_Alphabetic
0EB4	0EB9	Other_Alphabetic
0EBB	0EBC	Other_Alphabetic
0ECD		Other_Alphabetic
0F71	0F81	Other_Alphabetic
0F90	0F97	Other_Alphabetic
0F99	0FBC	Other_Alphabetic
102C	1032	Other_Alphabetic
1036		Other_Alphabetic
1038		Other_Alphabetic
1056	1059	Other_Alphabetic
135F		Other_Alphabetic
1712	1713	Other_Alphabetic
1732	1733	Other_Alphabetic
1752	1753	Other_Alphabetic
1772	1773	Other_Alphabetic
17B6	17C8	Other_Alphabetic
18A9		Other_Alphabetic
1920	192B	Other_Alphabetic
1930	1938	Other_Alphabetic
19B0	19C0	Other_Alphabetic
19C8	19C9	Other_Alphabetic
1A17	1A1B	Other_Alphabetic
24B6	24E9	Other_Alphabetic
A823	A827	Other_Alphabetic
FB1E		Other_Alphabetic
10A01	10A03	Other_Alphabetic
10A05	10A06	Other_Alphabetic
10A0C	10A0F	Other_Alphabetic
END
};
1;

package unicore::lib::selfloader::OtherDef;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{OtherDefaultIgnorableCodePoint} (and fuzzy permutations)
# 
# Meaning: Extended property 'Other_Default_Ignorable_Code_Point'
#
return <<'END';
034F		Other_Default_Ignorable_Code_Point
115F	1160	Other_Default_Ignorable_Code_Point
2064	2069	Other_Default_Ignorable_Code_Point
3164		Other_Default_Ignorable_Code_Point
FFA0		Other_Default_Ignorable_Code_Point
FFF0	FFF8	Other_Default_Ignorable_Code_Point
E0000		Other_Default_Ignorable_Code_Point
E0002	E001F	Other_Default_Ignorable_Code_Point
E0080	E00FF	Other_Default_Ignorable_Code_Point
E01F0	E0FFF	Other_Default_Ignorable_Code_Point
END
};
1;

package unicore::lib::selfloader::OtherGra;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{OtherGraphemeExtend} (and fuzzy permutations)
# 
# Meaning: Extended property 'Other_Grapheme_Extend'
#
return <<'END';
09BE		Other_Grapheme_Extend
09D7		Other_Grapheme_Extend
0B3E		Other_Grapheme_Extend
0B57		Other_Grapheme_Extend
0BBE		Other_Grapheme_Extend
0BD7		Other_Grapheme_Extend
0CC2		Other_Grapheme_Extend
0CD5	0CD6	Other_Grapheme_Extend
0D3E		Other_Grapheme_Extend
0D57		Other_Grapheme_Extend
0DCF		Other_Grapheme_Extend
0DDF		Other_Grapheme_Extend
200C	200D	Other_Grapheme_Extend
1D165		Other_Grapheme_Extend
1D16E	1D172	Other_Grapheme_Extend
END
};
1;

package unicore::lib::selfloader::OtherIdC;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{OtherIdContinue} (and fuzzy permutations)
# 
# Meaning: Extended property 'Other_ID_Continue'
#
return <<'END';
1369	1371	Other_ID_Continue
END
};
1;

package unicore::lib::selfloader::OtherIdS;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{OtherIdStart} (and fuzzy permutations)
# 
# Meaning: Extended property 'Other_ID_Start'
#
return <<'END';
2118		Other_ID_Start
212E		Other_ID_Start
309B	309C	Other_ID_Start
END
};
1;

package unicore::lib::selfloader::OtherLow;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{OtherLowercase} (and fuzzy permutations)
# 
# Meaning: Extended property 'Other_Lowercase'
#
return <<'END';
02B0	02B8	Other_Lowercase
02C0	02C1	Other_Lowercase
02E0	02E4	Other_Lowercase
0345		Other_Lowercase
037A		Other_Lowercase
1D2C	1D61	Other_Lowercase
1D78		Other_Lowercase
1D9B	1DBF	Other_Lowercase
2090	2094	Other_Lowercase
2170	217F	Other_Lowercase
24D0	24E9	Other_Lowercase
END
};
1;

package unicore::lib::selfloader::OtherMat;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{OtherMath} (and fuzzy permutations)
# 
# Meaning: Extended property 'Other_Math'
#
return <<'END';
005E		Other_Math
03D0	03D2	Other_Math
03D5		Other_Math
03F0	03F1	Other_Math
03F4	03F5	Other_Math
2016		Other_Math
2032	2034	Other_Math
2040		Other_Math
2061	2063	Other_Math
207D	207E	Other_Math
208D	208E	Other_Math
20D0	20DC	Other_Math
20E1		Other_Math
20E5	20E6	Other_Math
2102		Other_Math
210A	2113	Other_Math
2115		Other_Math
2119	211D	Other_Math
2124		Other_Math
2128	2129	Other_Math
212C	212D	Other_Math
212F	2131	Other_Math
2133	2138	Other_Math
213C	213F	Other_Math
2145	2149	Other_Math
23B7		Other_Math
23D0		Other_Math
27C5	27C6	Other_Math
27E6	27EB	Other_Math
2983	2998	Other_Math
29D8	29DB	Other_Math
29FC	29FD	Other_Math
FE61		Other_Math
FE63		Other_Math
FE68		Other_Math
FF3C		Other_Math
FF3E		Other_Math
1D400	1D454	Other_Math
1D456	1D49C	Other_Math
1D49E	1D49F	Other_Math
1D4A2		Other_Math
1D4A5	1D4A6	Other_Math
1D4A9	1D4AC	Other_Math
1D4AE	1D4B9	Other_Math
1D4BB		Other_Math
1D4BD	1D4C3	Other_Math
1D4C5	1D505	Other_Math
1D507	1D50A	Other_Math
1D50D	1D514	Other_Math
1D516	1D51C	Other_Math
1D51E	1D539	Other_Math
1D53B	1D53E	Other_Math
1D540	1D544	Other_Math
1D546		Other_Math
1D54A	1D550	Other_Math
1D552	1D6A3	Other_Math
1D6A8	1D6C0	Other_Math
1D6C2	1D6DA	Other_Math
1D6DC	1D6FA	Other_Math
1D6FC	1D714	Other_Math
1D716	1D734	Other_Math
1D736	1D74E	Other_Math
1D750	1D76E	Other_Math
1D770	1D788	Other_Math
1D78A	1D7A8	Other_Math
1D7AA	1D7C2	Other_Math
1D7C4	1D7C9	Other_Math
1D7CE	1D7FF	Other_Math
END
};
1;

package unicore::lib::selfloader::OtherUpp;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{OtherUppercase} (and fuzzy permutations)
# 
# Meaning: Extended property 'Other_Uppercase'
#
return <<'END';
2160	216F	Other_Uppercase
24B6	24CF	Other_Uppercase
END
};
1;

package unicore::lib::selfloader::OUpper;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Other_Uppercase'
#
return <<'END';
2160	216F	Other_Uppercase
24B6	24CF	Other_Uppercase
END
};
1;

package unicore::lib::selfloader::P;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{P}
# 	\p{P} (and fuzzy permutations)
# 
# Meaning: Major Category 'P'
#
return <<'END';
0021	0023	
0025	002A	
002C	002F	
003A	003B	
003F	0040	
005B	005D	
005F		
007B		
007D		
00A1		
00AB		
00B7		
00BB		
00BF		
037E		
0387		
055A	055F	
0589	058A	
05BE		
05C0		
05C3		
05C6		
05F3	05F4	
060C	060D	
061B		
061E	061F	
066A	066D	
06D4		
0700	070D	
0964	0965	
0970		
0DF4		
0E4F		
0E5A	0E5B	
0F04	0F12	
0F3A	0F3D	
0F85		
0FD0	0FD1	
104A	104F	
10FB		
1361	1368	
166D	166E	
169B	169C	
16EB	16ED	
1735	1736	
17D4	17D6	
17D8	17DA	
1800	180A	
1944	1945	
19DE	19DF	
1A1E	1A1F	
2010	2027	
2030	2043	
2045	2051	
2053	205E	
207D	207E	
208D	208E	
2329	232A	
23B4	23B6	
2768	2775	
27C5	27C6	
27E6	27EB	
2983	2998	
29D8	29DB	
29FC	29FD	
2CF9	2CFC	
2CFE	2CFF	
2E00	2E17	
2E1C	2E1D	
3001	3003	
3008	3011	
3014	301F	
3030		
303D		
30A0		
30FB		
FD3E	FD3F	
FE10	FE19	
FE30	FE52	
FE54	FE61	
FE63		
FE68		
FE6A	FE6B	
FF01	FF03	
FF05	FF0A	
FF0C	FF0F	
FF1A	FF1B	
FF1F	FF20	
FF3B	FF3D	
FF3F		
FF5B		
FF5D		
FF5F	FF65	
10100	10101	
1039F		
10A50	10A58	
END
};
1;

package unicore::lib::selfloader::PatSyn;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Pattern_Syntax'
#
return <<'END';
0021	002F	Pattern_Syntax
003A	0040	Pattern_Syntax
005B	005E	Pattern_Syntax
0060		Pattern_Syntax
007B	007E	Pattern_Syntax
00A1	00A7	Pattern_Syntax
00A9		Pattern_Syntax
00AB	00AC	Pattern_Syntax
00AE		Pattern_Syntax
00B0	00B1	Pattern_Syntax
00B6		Pattern_Syntax
00BB		Pattern_Syntax
00BF		Pattern_Syntax
00D7		Pattern_Syntax
00F7		Pattern_Syntax
2010	2027	Pattern_Syntax
2030	203E	Pattern_Syntax
2041	2053	Pattern_Syntax
2055	205E	Pattern_Syntax
2190	245F	Pattern_Syntax
2500	2775	Pattern_Syntax
2794	2BFF	Pattern_Syntax
2E00	2E7F	Pattern_Syntax
3001	3003	Pattern_Syntax
3008	3020	Pattern_Syntax
3030		Pattern_Syntax
FD3E	FD3F	Pattern_Syntax
FE45	FE46	Pattern_Syntax
END
};
1;

package unicore::lib::selfloader::PatternS;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{PatternSyntax} (and fuzzy permutations)
# 
# Meaning: Extended property 'Pattern_Syntax'
#
return <<'END';
0021	002F	Pattern_Syntax
003A	0040	Pattern_Syntax
005B	005E	Pattern_Syntax
0060		Pattern_Syntax
007B	007E	Pattern_Syntax
00A1	00A7	Pattern_Syntax
00A9		Pattern_Syntax
00AB	00AC	Pattern_Syntax
00AE		Pattern_Syntax
00B0	00B1	Pattern_Syntax
00B6		Pattern_Syntax
00BB		Pattern_Syntax
00BF		Pattern_Syntax
00D7		Pattern_Syntax
00F7		Pattern_Syntax
2010	2027	Pattern_Syntax
2030	203E	Pattern_Syntax
2041	2053	Pattern_Syntax
2055	205E	Pattern_Syntax
2190	245F	Pattern_Syntax
2500	2775	Pattern_Syntax
2794	2BFF	Pattern_Syntax
2E00	2E7F	Pattern_Syntax
3001	3003	Pattern_Syntax
3008	3020	Pattern_Syntax
3030		Pattern_Syntax
FD3E	FD3F	Pattern_Syntax
FE45	FE46	Pattern_Syntax
END
};
1;

package unicore::lib::selfloader::PatternW;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{PatternWhiteSpace} (and fuzzy permutations)
# 
# Meaning: Extended property 'Pattern_White_Space'
#
return <<'END';
0009	000D	Pattern_White_Space
0020		Pattern_White_Space
0085		Pattern_White_Space
200E	200F	Pattern_White_Space
2028	2029	Pattern_White_Space
END
};
1;

package unicore::lib::selfloader::PatWS;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Pattern_White_Space'
#
return <<'END';
0009	000D	Pattern_White_Space
0020		Pattern_White_Space
0085		Pattern_White_Space
200E	200F	Pattern_White_Space
2028	2029	Pattern_White_Space
END
};
1;

package unicore::lib::selfloader::Pc;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Pc}
# 	\p{Pc} (and fuzzy permutations)
# 
# Meaning: General Category 'Pc'
#
return <<'END';
005F		
203F	2040	
2054		
FE33	FE34	
FE4D	FE4F	
FF3F		
END
};
1;

package unicore::lib::selfloader::Pd;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Pd}
# 	\p{Pd} (and fuzzy permutations)
# 
# Meaning: General Category 'Pd'
#
return <<'END';
002D		
058A		
1806		
2010	2015	
2E17		
301C		
3030		
30A0		
FE31	FE32	
FE58		
FE63		
FF0D		
END
};
1;

package unicore::lib::selfloader::Pe;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Pe}
# 	\p{Pe} (and fuzzy permutations)
# 
# Meaning: General Category 'Pe'
#
return <<'END';
0029		
005D		
007D		
0F3B		
0F3D		
169C		
2046		
207E		
208E		
232A		
23B5		
2769		
276B		
276D		
276F		
2771		
2773		
2775		
27C6		
27E7		
27E9		
27EB		
2984		
2986		
2988		
298A		
298C		
298E		
2990		
2992		
2994		
2996		
2998		
29D9		
29DB		
29FD		
3009		
300B		
300D		
300F		
3011		
3015		
3017		
3019		
301B		
301E	301F	
FD3F		
FE18		
FE36		
FE38		
FE3A		
FE3C		
FE3E		
FE40		
FE42		
FE44		
FE48		
FE5A		
FE5C		
FE5E		
FF09		
FF3D		
FF5D		
FF60		
FF63		
END
};
1;

package unicore::lib::selfloader::Pf;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Pf}
# 	\p{Pf} (and fuzzy permutations)
# 
# Meaning: General Category 'Pf'
#
return <<'END';
00BB		
2019		
201D		
203A		
2E03		
2E05		
2E0A		
2E0D		
2E1D		
END
};
1;

package unicore::lib::selfloader::Pi;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Pi}
# 	\p{Pi} (and fuzzy permutations)
# 
# Meaning: General Category 'Pi'
#
return <<'END';
00AB		
2018		
201B	201C	
201F		
2039		
2E02		
2E04		
2E09		
2E0C		
2E1C		
END
};
1;

package unicore::lib::selfloader::Po;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Po}
# 	\p{Po} (and fuzzy permutations)
# 
# Meaning: General Category 'Po'
#
return <<'END';
0021	0023	
0025	0027	
002A		
002C		
002E	002F	
003A	003B	
003F	0040	
005C		
00A1		
00B7		
00BF		
037E		
0387		
055A	055F	
0589		
05BE		
05C0		
05C3		
05C6		
05F3	05F4	
060C	060D	
061B		
061E	061F	
066A	066D	
06D4		
0700	070D	
0964	0965	
0970		
0DF4		
0E4F		
0E5A	0E5B	
0F04	0F12	
0F85		
0FD0	0FD1	
104A	104F	
10FB		
1361	1368	
166D	166E	
16EB	16ED	
1735	1736	
17D4	17D6	
17D8	17DA	
1800	1805	
1807	180A	
1944	1945	
19DE	19DF	
1A1E	1A1F	
2016	2017	
2020	2027	
2030	2038	
203B	203E	
2041	2043	
2047	2051	
2053		
2055	205E	
23B6		
2CF9	2CFC	
2CFE	2CFF	
2E00	2E01	
2E06	2E08	
2E0B		
2E0E	2E16	
3001	3003	
303D		
30FB		
FE10	FE16	
FE19		
FE30		
FE45	FE46	
FE49	FE4C	
FE50	FE52	
FE54	FE57	
FE5F	FE61	
FE68		
FE6A	FE6B	
FF01	FF03	
FF05	FF07	
FF0A		
FF0C		
FF0E	FF0F	
FF1A	FF1B	
FF1F	FF20	
FF3C		
FF61		
FF64	FF65	
10100	10101	
1039F		
10A50	10A58	
END
};
1;

package unicore::lib::selfloader::Print;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Print}
# 
# Meaning: [[:Print:]]
#
return <<'END';
0009	000D	
0020	007E	
0085		
00A0	0241	
0250	036F	
0374	0375	
037A		
037E		
0384	038A	
038C		
038E	03A1	
03A3	03CE	
03D0	0486	
0488	04CE	
04D0	04F9	
0500	050F	
0531	0556	
0559	055F	
0561	0587	
0589	058A	
0591	05B9	
05BB	05C7	
05D0	05EA	
05F0	05F4	
0600	0603	
060B	0615	
061B		
061E	061F	
0621	063A	
0640	065E	
0660	070D	
070F	074A	
074D	076D	
0780	07B1	
0901	0939	
093C	094D	
0950	0954	
0958	0970	
097D		
0981	0983	
0985	098C	
098F	0990	
0993	09A8	
09AA	09B0	
09B2		
09B6	09B9	
09BC	09C4	
09C7	09C8	
09CB	09CE	
09D7		
09DC	09DD	
09DF	09E3	
09E6	09FA	
0A01	0A03	
0A05	0A0A	
0A0F	0A10	
0A13	0A28	
0A2A	0A30	
0A32	0A33	
0A35	0A36	
0A38	0A39	
0A3C		
0A3E	0A42	
0A47	0A48	
0A4B	0A4D	
0A59	0A5C	
0A5E		
0A66	0A74	
0A81	0A83	
0A85	0A8D	
0A8F	0A91	
0A93	0AA8	
0AAA	0AB0	
0AB2	0AB3	
0AB5	0AB9	
0ABC	0AC5	
0AC7	0AC9	
0ACB	0ACD	
0AD0		
0AE0	0AE3	
0AE6	0AEF	
0AF1		
0B01	0B03	
0B05	0B0C	
0B0F	0B10	
0B13	0B28	
0B2A	0B30	
0B32	0B33	
0B35	0B39	
0B3C	0B43	
0B47	0B48	
0B4B	0B4D	
0B56	0B57	
0B5C	0B5D	
0B5F	0B61	
0B66	0B71	
0B82	0B83	
0B85	0B8A	
0B8E	0B90	
0B92	0B95	
0B99	0B9A	
0B9C		
0B9E	0B9F	
0BA3	0BA4	
0BA8	0BAA	
0BAE	0BB9	
0BBE	0BC2	
0BC6	0BC8	
0BCA	0BCD	
0BD7		
0BE6	0BFA	
0C01	0C03	
0C05	0C0C	
0C0E	0C10	
0C12	0C28	
0C2A	0C33	
0C35	0C39	
0C3E	0C44	
0C46	0C48	
0C4A	0C4D	
0C55	0C56	
0C60	0C61	
0C66	0C6F	
0C82	0C83	
0C85	0C8C	
0C8E	0C90	
0C92	0CA8	
0CAA	0CB3	
0CB5	0CB9	
0CBC	0CC4	
0CC6	0CC8	
0CCA	0CCD	
0CD5	0CD6	
0CDE		
0CE0	0CE1	
0CE6	0CEF	
0D02	0D03	
0D05	0D0C	
0D0E	0D10	
0D12	0D28	
0D2A	0D39	
0D3E	0D43	
0D46	0D48	
0D4A	0D4D	
0D57		
0D60	0D61	
0D66	0D6F	
0D82	0D83	
0D85	0D96	
0D9A	0DB1	
0DB3	0DBB	
0DBD		
0DC0	0DC6	
0DCA		
0DCF	0DD4	
0DD6		
0DD8	0DDF	
0DF2	0DF4	
0E01	0E3A	
0E3F	0E5B	
0E81	0E82	
0E84		
0E87	0E88	
0E8A		
0E8D		
0E94	0E97	
0E99	0E9F	
0EA1	0EA3	
0EA5		
0EA7		
0EAA	0EAB	
0EAD	0EB9	
0EBB	0EBD	
0EC0	0EC4	
0EC6		
0EC8	0ECD	
0ED0	0ED9	
0EDC	0EDD	
0F00	0F47	
0F49	0F6A	
0F71	0F8B	
0F90	0F97	
0F99	0FBC	
0FBE	0FCC	
0FCF	0FD1	
1000	1021	
1023	1027	
1029	102A	
102C	1032	
1036	1039	
1040	1059	
10A0	10C5	
10D0	10FC	
1100	1159	
115F	11A2	
11A8	11F9	
1200	1248	
124A	124D	
1250	1256	
1258		
125A	125D	
1260	1288	
128A	128D	
1290	12B0	
12B2	12B5	
12B8	12BE	
12C0		
12C2	12C5	
12C8	12D6	
12D8	1310	
1312	1315	
1318	135A	
135F	137C	
1380	1399	
13A0	13F4	
1401	1676	
1680	169C	
16A0	16F0	
1700	170C	
170E	1714	
1720	1736	
1740	1753	
1760	176C	
176E	1770	
1772	1773	
1780	17DD	
17E0	17E9	
17F0	17F9	
1800	180E	
1810	1819	
1820	1877	
1880	18A9	
1900	191C	
1920	192B	
1930	193B	
1940		
1944	196D	
1970	1974	
1980	19A9	
19B0	19C9	
19D0	19D9	
19DE	1A1B	
1A1E	1A1F	
1D00	1DC3	
1E00	1E9B	
1EA0	1EF9	
1F00	1F15	
1F18	1F1D	
1F20	1F45	
1F48	1F4D	
1F50	1F57	
1F59		
1F5B		
1F5D		
1F5F	1F7D	
1F80	1FB4	
1FB6	1FC4	
1FC6	1FD3	
1FD6	1FDB	
1FDD	1FEF	
1FF2	1FF4	
1FF6	1FFE	
2000	2063	
206A	2071	
2074	208E	
2090	2094	
20A0	20B5	
20D0	20EB	
2100	214C	
2153	2183	
2190	23DB	
2400	2426	
2440	244A	
2460	269C	
26A0	26B1	
2701	2704	
2706	2709	
270C	2727	
2729	274B	
274D		
274F	2752	
2756		
2758	275E	
2761	2794	
2798	27AF	
27B1	27BE	
27C0	27C6	
27D0	27EB	
27F0	2B13	
2C00	2C2E	
2C30	2C5E	
2C80	2CEA	
2CF9	2D25	
2D30	2D65	
2D6F		
2D80	2D96	
2DA0	2DA6	
2DA8	2DAE	
2DB0	2DB6	
2DB8	2DBE	
2DC0	2DC6	
2DC8	2DCE	
2DD0	2DD6	
2DD8	2DDE	
2E00	2E17	
2E1C	2E1D	
2E80	2E99	
2E9B	2EF3	
2F00	2FD5	
2FF0	2FFB	
3000	303F	
3041	3096	
3099	30FF	
3105	312C	
3131	318E	
3190	31B7	
31C0	31CF	
31F0	321E	
3220	3243	
3250	32FE	
3300	4DB5	
4DC0	9FBB	
A000	A48C	
A490	A4C6	
A700	A716	
A800	A82B	
AC00	D7A3	
E000	FA2D	
FA30	FA6A	
FA70	FAD9	
FB00	FB06	
FB13	FB17	
FB1D	FB36	
FB38	FB3C	
FB3E		
FB40	FB41	
FB43	FB44	
FB46	FBB1	
FBD3	FD3F	
FD50	FD8F	
FD92	FDC7	
FDF0	FDFD	
FE00	FE19	
FE20	FE23	
FE30	FE52	
FE54	FE66	
FE68	FE6B	
FE70	FE74	
FE76	FEFC	
FEFF		
FF01	FFBE	
FFC2	FFC7	
FFCA	FFCF	
FFD2	FFD7	
FFDA	FFDC	
FFE0	FFE6	
FFE8	FFEE	
FFF9	FFFD	
10000	1000B	
1000D	10026	
10028	1003A	
1003C	1003D	
1003F	1004D	
10050	1005D	
10080	100FA	
10100	10102	
10107	10133	
10137	1018A	
10300	1031E	
10320	10323	
10330	1034A	
10380	1039D	
1039F	103C3	
103C8	103D5	
10400	1049D	
104A0	104A9	
10800	10805	
10808		
1080A	10835	
10837	10838	
1083C		
1083F		
10A00	10A03	
10A05	10A06	
10A0C	10A13	
10A15	10A17	
10A19	10A33	
10A38	10A3A	
10A3F	10A47	
10A50	10A58	
1D000	1D0F5	
1D100	1D126	
1D12A	1D1DD	
1D200	1D245	
1D300	1D356	
1D400	1D454	
1D456	1D49C	
1D49E	1D49F	
1D4A2		
1D4A5	1D4A6	
1D4A9	1D4AC	
1D4AE	1D4B9	
1D4BB		
1D4BD	1D4C3	
1D4C5	1D505	
1D507	1D50A	
1D50D	1D514	
1D516	1D51C	
1D51E	1D539	
1D53B	1D53E	
1D540	1D544	
1D546		
1D54A	1D550	
1D552	1D6A5	
1D6A8	1D7C9	
1D7CE	1D7FF	
20000	2A6D6	
2F800	2FA1D	
E0001		
E0020	E007F	
E0100	E01EF	
F0000	FFFFD	
100000	10FFFD	
END
};
1;

package unicore::lib::selfloader::Ps;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Ps}
# 	\p{Ps} (and fuzzy permutations)
# 
# Meaning: General Category 'Ps'
#
return <<'END';
0028		
005B		
007B		
0F3A		
0F3C		
169B		
201A		
201E		
2045		
207D		
208D		
2329		
23B4		
2768		
276A		
276C		
276E		
2770		
2772		
2774		
27C5		
27E6		
27E8		
27EA		
2983		
2985		
2987		
2989		
298B		
298D		
298F		
2991		
2993		
2995		
2997		
29D8		
29DA		
29FC		
3008		
300A		
300C		
300E		
3010		
3014		
3016		
3018		
301A		
301D		
FD3E		
FE17		
FE35		
FE37		
FE39		
FE3B		
FE3D		
FE3F		
FE41		
FE43		
FE47		
FE59		
FE5B		
FE5D		
FF08		
FF3B		
FF5B		
FF5F		
FF62		
END
};
1;

package unicore::lib::selfloader::Punct;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Punct}
# 
# Meaning: [[:Punct:]]
#
return <<'END';
0021	0023	
0025	002A	
002C	002F	
003A	003B	
003F	0040	
005B	005D	
005F		
007B		
007D		
00A1		
00AB		
00B7		
00BB		
00BF		
037E		
0387		
055A	055F	
0589	058A	
05BE		
05C0		
05C3		
05C6		
05F3	05F4	
060C	060D	
061B		
061E	061F	
066A	066D	
06D4		
0700	070D	
0964	0965	
0970		
0DF4		
0E4F		
0E5A	0E5B	
0F04	0F12	
0F3A	0F3D	
0F85		
0FD0	0FD1	
104A	104F	
10FB		
1361	1368	
166D	166E	
169B	169C	
16EB	16ED	
1735	1736	
17D4	17D6	
17D8	17DA	
1800	180A	
1944	1945	
19DE	19DF	
1A1E	1A1F	
2010	2027	
2030	2043	
2045	2051	
2053	205E	
207D	207E	
208D	208E	
2329	232A	
23B4	23B6	
2768	2775	
27C5	27C6	
27E6	27EB	
2983	2998	
29D8	29DB	
29FC	29FD	
2CF9	2CFC	
2CFE	2CFF	
2E00	2E17	
2E1C	2E1D	
3001	3003	
3008	3011	
3014	301F	
3030		
303D		
30A0		
30FB		
FD3E	FD3F	
FE10	FE19	
FE30	FE52	
FE54	FE61	
FE63		
FE68		
FE6A	FE6B	
FF01	FF03	
FF05	FF0A	
FF0C	FF0F	
FF1A	FF1B	
FF1F	FF20	
FF3B	FF3D	
FF3F		
FF5B		
FF5D		
FF5F	FF65	
10100	10101	
1039F		
10A50	10A58	
END
};
1;

package unicore::lib::selfloader::Qaai;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Inherited} (and fuzzy permutations)
# 
# Meaning: Script 'Inherited'
#
return <<'END';
0300	036F	Inherited
064B	0655	Inherited
0670		Inherited
1DC0	1DC3	Inherited
200C	200D	Inherited
20D0	20EB	Inherited
302A	302F	Inherited
3099	309A	Inherited
FE00	FE0F	Inherited
FE20	FE23	Inherited
1D167	1D169	Inherited
1D17B	1D182	Inherited
1D185	1D18B	Inherited
1D1AA	1D1AD	Inherited
E0100	E01EF	Inherited
END
};
1;

package unicore::lib::selfloader::QMark;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Quotation_Mark'
#
return <<'END';
0022		Quotation_Mark
0027		Quotation_Mark
00AB		Quotation_Mark
00BB		Quotation_Mark
2018	201F	Quotation_Mark
2039	203A	Quotation_Mark
300C	300F	Quotation_Mark
301D	301F	Quotation_Mark
FE41	FE44	Quotation_Mark
FF02		Quotation_Mark
FF07		Quotation_Mark
FF62	FF63	Quotation_Mark
END
};
1;

package unicore::lib::selfloader::Quotatio;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{QuotationMark} (and fuzzy permutations)
# 
# Meaning: Extended property 'Quotation_Mark'
#
return <<'END';
0022		Quotation_Mark
0027		Quotation_Mark
00AB		Quotation_Mark
00BB		Quotation_Mark
2018	201F	Quotation_Mark
2039	203A	Quotation_Mark
300C	300F	Quotation_Mark
301D	301F	Quotation_Mark
FE41	FE44	Quotation_Mark
FF02		Quotation_Mark
FF07		Quotation_Mark
FF62	FF63	Quotation_Mark
END
};
1;

package unicore::lib::selfloader::Radical;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Radical'
#
return <<'END';
2E80	2E99	Radical
2E9B	2EF3	Radical
2F00	2FD5	Radical
END
};
1;

package unicore::lib::selfloader::Radical2;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Radical} (and fuzzy permutations)
# 
# Meaning: Extended property 'Radical'
#
return <<'END';
2E80	2E99	Radical
2E9B	2EF3	Radical
2F00	2FD5	Radical
END
};
1;

package unicore::lib::selfloader::Runr;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Runic} (and fuzzy permutations)
# 
# Meaning: Script 'Runic'
#
return <<'END';
16A0	16EA	Runic
16EE	16F0	Runic
END
};
1;

package unicore::lib::selfloader::S;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{S}
# 	\p{S} (and fuzzy permutations)
# 
# Meaning: Major Category 'S'
#
return <<'END';
0024		
002B		
003C	003E	
005E		
0060		
007C		
007E		
00A2	00A9	
00AC		
00AE	00B1	
00B4		
00B6		
00B8		
00D7		
00F7		
02C2	02C5	
02D2	02DF	
02E5	02ED	
02EF	02FF	
0374	0375	
0384	0385	
03F6		
0482		
060B		
060E	060F	
06E9		
06FD	06FE	
09F2	09F3	
09FA		
0AF1		
0B70		
0BF3	0BFA	
0E3F		
0F01	0F03	
0F13	0F17	
0F1A	0F1F	
0F34		
0F36		
0F38		
0FBE	0FC5	
0FC7	0FCC	
0FCF		
1360		
1390	1399	
17DB		
1940		
19E0	19FF	
1FBD		
1FBF	1FC1	
1FCD	1FCF	
1FDD	1FDF	
1FED	1FEF	
1FFD	1FFE	
2044		
2052		
207A	207C	
208A	208C	
20A0	20B5	
2100	2101	
2103	2106	
2108	2109	
2114		
2116	2118	
211E	2123	
2125		
2127		
2129		
212E		
2132		
213A	213B	
2140	2144	
214A	214C	
2190	2328	
232B	23B3	
23B7	23DB	
2400	2426	
2440	244A	
249C	24E9	
2500	269C	
26A0	26B1	
2701	2704	
2706	2709	
270C	2727	
2729	274B	
274D		
274F	2752	
2756		
2758	275E	
2761	2767	
2794		
2798	27AF	
27B1	27BE	
27C0	27C4	
27D0	27E5	
27F0	2982	
2999	29D7	
29DC	29FB	
29FE	2B13	
2CE5	2CEA	
2E80	2E99	
2E9B	2EF3	
2F00	2FD5	
2FF0	2FFB	
3004		
3012	3013	
3020		
3036	3037	
303E	303F	
309B	309C	
3190	3191	
3196	319F	
31C0	31CF	
3200	321E	
322A	3243	
3250		
3260	327F	
328A	32B0	
32C0	32FE	
3300	33FF	
4DC0	4DFF	
A490	A4C6	
A700	A716	
A828	A82B	
FB29		
FDFC	FDFD	
FE62		
FE64	FE66	
FE69		
FF04		
FF0B		
FF1C	FF1E	
FF3E		
FF40		
FF5C		
FF5E		
FFE0	FFE6	
FFE8	FFEE	
FFFC	FFFD	
10102		
10137	1013F	
10179	10189	
103D0		
1D000	1D0F5	
1D100	1D126	
1D12A	1D164	
1D16A	1D16C	
1D183	1D184	
1D18C	1D1A9	
1D1AE	1D1DD	
1D200	1D241	
1D245		
1D300	1D356	
1D6C1		
1D6DB		
1D6FB		
1D715		
1D735		
1D74F		
1D76F		
1D789		
1D7A9		
1D7C3		
END
};
1;

package unicore::lib::selfloader::Sc;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Sc}
# 	\p{Sc} (and fuzzy permutations)
# 
# Meaning: General Category 'Sc'
#
return <<'END';
0024		
00A2	00A5	
060B		
09F2	09F3	
0AF1		
0BF9		
0E3F		
17DB		
20A0	20B5	
FDFC		
FE69		
FF04		
FFE0	FFE1	
FFE5	FFE6	
END
};
1;

package unicore::lib::selfloader::SD;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Soft_Dotted'
#
return <<'END';
0069	006A	Soft_Dotted
012F		Soft_Dotted
0268		Soft_Dotted
029D		Soft_Dotted
02B2		Soft_Dotted
03F3		Soft_Dotted
0456		Soft_Dotted
0458		Soft_Dotted
1D62		Soft_Dotted
1D96		Soft_Dotted
1DA4		Soft_Dotted
1DA8		Soft_Dotted
1E2D		Soft_Dotted
1ECB		Soft_Dotted
2071		Soft_Dotted
2148	2149	Soft_Dotted
1D422	1D423	Soft_Dotted
1D456	1D457	Soft_Dotted
1D48A	1D48B	Soft_Dotted
1D4BE	1D4BF	Soft_Dotted
1D4F2	1D4F3	Soft_Dotted
1D526	1D527	Soft_Dotted
1D55A	1D55B	Soft_Dotted
1D58E	1D58F	Soft_Dotted
1D5C2	1D5C3	Soft_Dotted
1D5F6	1D5F7	Soft_Dotted
1D62A	1D62B	Soft_Dotted
1D65E	1D65F	Soft_Dotted
1D692	1D693	Soft_Dotted
END
};
1;

package unicore::lib::selfloader::Shaw;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Shavian} (and fuzzy permutations)
# 
# Meaning: Script 'Shavian'
#
return <<'END';
10450	1047F	Shavian
END
};
1;

package unicore::lib::selfloader::Sinh;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Sinhala} (and fuzzy permutations)
# 
# Meaning: Script 'Sinhala'
#
return <<'END';
0D82	0D83	Sinhala
0D85	0D96	Sinhala
0D9A	0DB1	Sinhala
0DB3	0DBB	Sinhala
0DBD		Sinhala
0DC0	0DC6	Sinhala
0DCA		Sinhala
0DCF	0DD4	Sinhala
0DD6		Sinhala
0DD8	0DDF	Sinhala
0DF2	0DF4	Sinhala
END
};
1;

package unicore::lib::selfloader::Sk;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Sk}
# 	\p{Sk} (and fuzzy permutations)
# 
# Meaning: General Category 'Sk'
#
return <<'END';
005E		
0060		
00A8		
00AF		
00B4		
00B8		
02C2	02C5	
02D2	02DF	
02E5	02ED	
02EF	02FF	
0374	0375	
0384	0385	
1FBD		
1FBF	1FC1	
1FCD	1FCF	
1FDD	1FDF	
1FED	1FEF	
1FFD	1FFE	
309B	309C	
A700	A716	
FF3E		
FF40		
FFE3		
END
};
1;

package unicore::lib::selfloader::Sm;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Sm}
# 	\p{Sm} (and fuzzy permutations)
# 
# Meaning: General Category 'Sm'
#
return <<'END';
002B		
003C	003E	
007C		
007E		
00AC		
00B1		
00D7		
00F7		
03F6		
2044		
2052		
207A	207C	
208A	208C	
2140	2144	
214B		
2190	2194	
219A	219B	
21A0		
21A3		
21A6		
21AE		
21CE	21CF	
21D2		
21D4		
21F4	22FF	
2308	230B	
2320	2321	
237C		
239B	23B3	
25B7		
25C1		
25F8	25FF	
266F		
27C0	27C4	
27D0	27E5	
27F0	27FF	
2900	2982	
2999	29D7	
29DC	29FB	
29FE	2AFF	
FB29		
FE62		
FE64	FE66	
FF0B		
FF1C	FF1E	
FF5C		
FF5E		
FFE2		
FFE9	FFEC	
1D6C1		
1D6DB		
1D6FB		
1D715		
1D735		
1D74F		
1D76F		
1D789		
1D7A9		
1D7C3		
END
};
1;

package unicore::lib::selfloader::So;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{So}
# 	\p{So} (and fuzzy permutations)
# 
# Meaning: General Category 'So'
#
return <<'END';
00A6	00A7	
00A9		
00AE		
00B0		
00B6		
0482		
060E	060F	
06E9		
06FD	06FE	
09FA		
0B70		
0BF3	0BF8	
0BFA		
0F01	0F03	
0F13	0F17	
0F1A	0F1F	
0F34		
0F36		
0F38		
0FBE	0FC5	
0FC7	0FCC	
0FCF		
1360		
1390	1399	
1940		
19E0	19FF	
2100	2101	
2103	2106	
2108	2109	
2114		
2116	2118	
211E	2123	
2125		
2127		
2129		
212E		
2132		
213A	213B	
214A		
214C		
2195	2199	
219C	219F	
21A1	21A2	
21A4	21A5	
21A7	21AD	
21AF	21CD	
21D0	21D1	
21D3		
21D5	21F3	
2300	2307	
230C	231F	
2322	2328	
232B	237B	
237D	239A	
23B7	23DB	
2400	2426	
2440	244A	
249C	24E9	
2500	25B6	
25B8	25C0	
25C2	25F7	
2600	266E	
2670	269C	
26A0	26B1	
2701	2704	
2706	2709	
270C	2727	
2729	274B	
274D		
274F	2752	
2756		
2758	275E	
2761	2767	
2794		
2798	27AF	
27B1	27BE	
2800	28FF	
2B00	2B13	
2CE5	2CEA	
2E80	2E99	
2E9B	2EF3	
2F00	2FD5	
2FF0	2FFB	
3004		
3012	3013	
3020		
3036	3037	
303E	303F	
3190	3191	
3196	319F	
31C0	31CF	
3200	321E	
322A	3243	
3250		
3260	327F	
328A	32B0	
32C0	32FE	
3300	33FF	
4DC0	4DFF	
A490	A4C6	
A828	A82B	
FDFD		
FFE4		
FFE8		
FFED	FFEE	
FFFC	FFFD	
10102		
10137	1013F	
10179	10189	
103D0		
1D000	1D0F5	
1D100	1D126	
1D12A	1D164	
1D16A	1D16C	
1D183	1D184	
1D18C	1D1A9	
1D1AE	1D1DD	
1D200	1D241	
1D245		
1D300	1D356	
END
};
1;

package unicore::lib::selfloader::SoftDott;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{SoftDotted} (and fuzzy permutations)
# 
# Meaning: Extended property 'Soft_Dotted'
#
return <<'END';
0069	006A	Soft_Dotted
012F		Soft_Dotted
0268		Soft_Dotted
029D		Soft_Dotted
02B2		Soft_Dotted
03F3		Soft_Dotted
0456		Soft_Dotted
0458		Soft_Dotted
1D62		Soft_Dotted
1D96		Soft_Dotted
1DA4		Soft_Dotted
1DA8		Soft_Dotted
1E2D		Soft_Dotted
1ECB		Soft_Dotted
2071		Soft_Dotted
2148	2149	Soft_Dotted
1D422	1D423	Soft_Dotted
1D456	1D457	Soft_Dotted
1D48A	1D48B	Soft_Dotted
1D4BE	1D4BF	Soft_Dotted
1D4F2	1D4F3	Soft_Dotted
1D526	1D527	Soft_Dotted
1D55A	1D55B	Soft_Dotted
1D58E	1D58F	Soft_Dotted
1D5C2	1D5C3	Soft_Dotted
1D5F6	1D5F7	Soft_Dotted
1D62A	1D62B	Soft_Dotted
1D65E	1D65F	Soft_Dotted
1D692	1D693	Soft_Dotted
END
};
1;

package unicore::lib::selfloader::Space;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Space}
# 
# Meaning: [[:Space:]]
#
return <<'END';
0009	000D	
0020		
0085		
00A0		
1680		
180E		
2000	200A	
2028	2029	
202F		
205F		
3000		
END
};
1;

package unicore::lib::selfloader::SpacePer;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{SpacePerl}
# 
# Meaning: \s
#
return <<'END';
0009	000A	
000C	000D	
0020		
0085		
00A0		
1680		
180E		
2000	200A	
2028	2029	
202F		
205F		
3000		
END
};
1;

package unicore::lib::selfloader::STerm;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'STerm'
#
return <<'END';
0021		STerm
002E		STerm
003F		STerm
055C		STerm
055E		STerm
0589		STerm
061F		STerm
06D4		STerm
0700	0702	STerm
0964	0965	STerm
104A	104B	STerm
1362		STerm
1367	1368	STerm
166E		STerm
1803		STerm
1809		STerm
1944	1945	STerm
203C	203D	STerm
2047	2049	STerm
3002		STerm
FE52		STerm
FE56	FE57	STerm
FF01		STerm
FF0E		STerm
FF1F		STerm
FF61		STerm
END
};
1;

package unicore::lib::selfloader::Sterm2;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Sterm} (and fuzzy permutations)
# 
# Meaning: Extended property 'STerm'
#
return <<'END';
0021		STerm
002E		STerm
003F		STerm
055C		STerm
055E		STerm
0589		STerm
061F		STerm
06D4		STerm
0700	0702	STerm
0964	0965	STerm
104A	104B	STerm
1362		STerm
1367	1368	STerm
166E		STerm
1803		STerm
1809		STerm
1944	1945	STerm
203C	203D	STerm
2047	2049	STerm
3002		STerm
FE52		STerm
FE56	FE57	STerm
FF01		STerm
FF0E		STerm
FF1F		STerm
FF61		STerm
END
};
1;

package unicore::lib::selfloader::SylotiNa;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{SylotiNagri} (and fuzzy permutations)
# 
# Meaning: Script 'Syloti_Nagri'
#
return <<'END';
A800	A82B	Syloti_Nagri
END
};
1;

package unicore::lib::selfloader::Syrc;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Syriac} (and fuzzy permutations)
# 
# Meaning: Script 'Syriac'
#
return <<'END';
0700	070D	Syriac
070F	074A	Syriac
074D	074F	Syriac
END
};
1;

package unicore::lib::selfloader::Tagb;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Tagbanwa} (and fuzzy permutations)
# 
# Meaning: Script 'Tagbanwa'
#
return <<'END';
1760	176C	Tagbanwa
176E	1770	Tagbanwa
1772	1773	Tagbanwa
END
};
1;

package unicore::lib::selfloader::TaiLe;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{TaiLe} (and fuzzy permutations)
# 
# Meaning: Script 'Tai_Le'
#
return <<'END';
1950	196D	Tai_Le
1970	1974	Tai_Le
END
};
1;

package unicore::lib::selfloader::Taml;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Tamil} (and fuzzy permutations)
# 
# Meaning: Script 'Tamil'
#
return <<'END';
0B82	0B83	Tamil
0B85	0B8A	Tamil
0B8E	0B90	Tamil
0B92	0B95	Tamil
0B99	0B9A	Tamil
0B9C		Tamil
0B9E	0B9F	Tamil
0BA3	0BA4	Tamil
0BA8	0BAA	Tamil
0BAE	0BB9	Tamil
0BBE	0BC2	Tamil
0BC6	0BC8	Tamil
0BCA	0BCD	Tamil
0BD7		Tamil
0BE6	0BFA	Tamil
END
};
1;

package unicore::lib::selfloader::Telu;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Telugu} (and fuzzy permutations)
# 
# Meaning: Script 'Telugu'
#
return <<'END';
0C01	0C03	Telugu
0C05	0C0C	Telugu
0C0E	0C10	Telugu
0C12	0C28	Telugu
0C2A	0C33	Telugu
0C35	0C39	Telugu
0C3E	0C44	Telugu
0C46	0C48	Telugu
0C4A	0C4D	Telugu
0C55	0C56	Telugu
0C60	0C61	Telugu
0C66	0C6F	Telugu
END
};
1;

package unicore::lib::selfloader::Term;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Terminal_Punctuation'
#
return <<'END';
0021		Terminal_Punctuation
002C		Terminal_Punctuation
002E		Terminal_Punctuation
003A	003B	Terminal_Punctuation
003F		Terminal_Punctuation
037E		Terminal_Punctuation
0387		Terminal_Punctuation
0589		Terminal_Punctuation
05C3		Terminal_Punctuation
060C		Terminal_Punctuation
061B		Terminal_Punctuation
061F		Terminal_Punctuation
06D4		Terminal_Punctuation
0700	070A	Terminal_Punctuation
070C		Terminal_Punctuation
0964	0965	Terminal_Punctuation
0E5A	0E5B	Terminal_Punctuation
0F08		Terminal_Punctuation
0F0D	0F12	Terminal_Punctuation
104A	104B	Terminal_Punctuation
1361	1368	Terminal_Punctuation
166D	166E	Terminal_Punctuation
16EB	16ED	Terminal_Punctuation
17D4	17D6	Terminal_Punctuation
17DA		Terminal_Punctuation
1802	1805	Terminal_Punctuation
1808	1809	Terminal_Punctuation
1944	1945	Terminal_Punctuation
203C	203D	Terminal_Punctuation
2047	2049	Terminal_Punctuation
3001	3002	Terminal_Punctuation
FE50	FE52	Terminal_Punctuation
FE54	FE57	Terminal_Punctuation
FF01		Terminal_Punctuation
FF0C		Terminal_Punctuation
FF0E		Terminal_Punctuation
FF1A	FF1B	Terminal_Punctuation
FF1F		Terminal_Punctuation
FF61		Terminal_Punctuation
FF64		Terminal_Punctuation
END
};
1;

package unicore::lib::selfloader::Terminal;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{TerminalPunctuation} (and fuzzy permutations)
# 
# Meaning: Extended property 'Terminal_Punctuation'
#
return <<'END';
0021		Terminal_Punctuation
002C		Terminal_Punctuation
002E		Terminal_Punctuation
003A	003B	Terminal_Punctuation
003F		Terminal_Punctuation
037E		Terminal_Punctuation
0387		Terminal_Punctuation
0589		Terminal_Punctuation
05C3		Terminal_Punctuation
060C		Terminal_Punctuation
061B		Terminal_Punctuation
061F		Terminal_Punctuation
06D4		Terminal_Punctuation
0700	070A	Terminal_Punctuation
070C		Terminal_Punctuation
0964	0965	Terminal_Punctuation
0E5A	0E5B	Terminal_Punctuation
0F08		Terminal_Punctuation
0F0D	0F12	Terminal_Punctuation
104A	104B	Terminal_Punctuation
1361	1368	Terminal_Punctuation
166D	166E	Terminal_Punctuation
16EB	16ED	Terminal_Punctuation
17D4	17D6	Terminal_Punctuation
17DA		Terminal_Punctuation
1802	1805	Terminal_Punctuation
1808	1809	Terminal_Punctuation
1944	1945	Terminal_Punctuation
203C	203D	Terminal_Punctuation
2047	2049	Terminal_Punctuation
3001	3002	Terminal_Punctuation
FE50	FE52	Terminal_Punctuation
FE54	FE57	Terminal_Punctuation
FF01		Terminal_Punctuation
FF0C		Terminal_Punctuation
FF0E		Terminal_Punctuation
FF1A	FF1B	Terminal_Punctuation
FF1F		Terminal_Punctuation
FF61		Terminal_Punctuation
FF64		Terminal_Punctuation
END
};
1;

package unicore::lib::selfloader::Tfng;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Tifinagh} (and fuzzy permutations)
# 
# Meaning: Script 'Tifinagh'
#
return <<'END';
2D30	2D65	Tifinagh
2D6F		Tifinagh
END
};
1;

package unicore::lib::selfloader::Tglg;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Tagalog} (and fuzzy permutations)
# 
# Meaning: Script 'Tagalog'
#
return <<'END';
1700	170C	Tagalog
170E	1714	Tagalog
END
};
1;

package unicore::lib::selfloader::Thaa;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Thaana} (and fuzzy permutations)
# 
# Meaning: Script 'Thaana'
#
return <<'END';
0780	07B1	Thaana
END
};
1;

package unicore::lib::selfloader::Thai;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Thai} (and fuzzy permutations)
# 
# Meaning: Script 'Thai'
#
return <<'END';
0E01	0E3A	Thai
0E40	0E5B	Thai
END
};
1;

package unicore::lib::selfloader::Tibt;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Tibetan} (and fuzzy permutations)
# 
# Meaning: Script 'Tibetan'
#
return <<'END';
0F00	0F47	Tibetan
0F49	0F6A	Tibetan
0F71	0F8B	Tibetan
0F90	0F97	Tibetan
0F99	0FBC	Tibetan
0FBE	0FCC	Tibetan
0FCF	0FD1	Tibetan
END
};
1;

package unicore::lib::selfloader::Title;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Title}
# 
# Meaning: [[:Title:]]
#
return <<'END';
01C5		
01C8		
01CB		
01F2		
1F88	1F8F	
1F98	1F9F	
1FA8	1FAF	
1FBC		
1FCC		
1FFC		
END
};
1;

package unicore::lib::selfloader::Ugar;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Ugaritic} (and fuzzy permutations)
# 
# Meaning: Script 'Ugaritic'
#
return <<'END';
10380	1039D	Ugaritic
1039F		Ugaritic
END
};
1;

package unicore::lib::selfloader::UIdeo;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Unified_Ideograph'
#
return <<'END';
3400	4DB5	Unified_Ideograph
4E00	9FBB	Unified_Ideograph
FA0E	FA0F	Unified_Ideograph
FA11		Unified_Ideograph
FA13	FA14	Unified_Ideograph
FA1F		Unified_Ideograph
FA21		Unified_Ideograph
FA23	FA24	Unified_Ideograph
FA27	FA29	Unified_Ideograph
20000	2A6D6	Unified_Ideograph
END
};
1;

package unicore::lib::selfloader::UnifiedI;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{UnifiedIdeograph} (and fuzzy permutations)
# 
# Meaning: Extended property 'Unified_Ideograph'
#
return <<'END';
3400	4DB5	Unified_Ideograph
4E00	9FBB	Unified_Ideograph
FA0E	FA0F	Unified_Ideograph
FA11		Unified_Ideograph
FA13	FA14	Unified_Ideograph
FA1F		Unified_Ideograph
FA21		Unified_Ideograph
FA23	FA24	Unified_Ideograph
FA27	FA29	Unified_Ideograph
20000	2A6D6	Unified_Ideograph
END
};
1;

package unicore::lib::selfloader::Upper;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Upper}
# 
# Meaning: [[:Upper:]]
#
return <<'END';
0041	005A	
00C0	00D6	
00D8	00DE	
0100		
0102		
0104		
0106		
0108		
010A		
010C		
010E		
0110		
0112		
0114		
0116		
0118		
011A		
011C		
011E		
0120		
0122		
0124		
0126		
0128		
012A		
012C		
012E		
0130		
0132		
0134		
0136		
0139		
013B		
013D		
013F		
0141		
0143		
0145		
0147		
014A		
014C		
014E		
0150		
0152		
0154		
0156		
0158		
015A		
015C		
015E		
0160		
0162		
0164		
0166		
0168		
016A		
016C		
016E		
0170		
0172		
0174		
0176		
0178	0179	
017B		
017D		
0181	0182	
0184		
0186	0187	
0189	018B	
018E	0191	
0193	0194	
0196	0198	
019C	019D	
019F	01A0	
01A2		
01A4		
01A6	01A7	
01A9		
01AC		
01AE	01AF	
01B1	01B3	
01B5		
01B7	01B8	
01BC		
01C4		
01C7		
01CA		
01CD		
01CF		
01D1		
01D3		
01D5		
01D7		
01D9		
01DB		
01DE		
01E0		
01E2		
01E4		
01E6		
01E8		
01EA		
01EC		
01EE		
01F1		
01F4		
01F6	01F8	
01FA		
01FC		
01FE		
0200		
0202		
0204		
0206		
0208		
020A		
020C		
020E		
0210		
0212		
0214		
0216		
0218		
021A		
021C		
021E		
0220		
0222		
0224		
0226		
0228		
022A		
022C		
022E		
0230		
0232		
023A	023B	
023D	023E	
0241		
0386		
0388	038A	
038C		
038E	038F	
0391	03A1	
03A3	03AB	
03D2	03D4	
03D8		
03DA		
03DC		
03DE		
03E0		
03E2		
03E4		
03E6		
03E8		
03EA		
03EC		
03EE		
03F4		
03F7		
03F9	03FA	
03FD	042F	
0460		
0462		
0464		
0466		
0468		
046A		
046C		
046E		
0470		
0472		
0474		
0476		
0478		
047A		
047C		
047E		
0480		
048A		
048C		
048E		
0490		
0492		
0494		
0496		
0498		
049A		
049C		
049E		
04A0		
04A2		
04A4		
04A6		
04A8		
04AA		
04AC		
04AE		
04B0		
04B2		
04B4		
04B6		
04B8		
04BA		
04BC		
04BE		
04C0	04C1	
04C3		
04C5		
04C7		
04C9		
04CB		
04CD		
04D0		
04D2		
04D4		
04D6		
04D8		
04DA		
04DC		
04DE		
04E0		
04E2		
04E4		
04E6		
04E8		
04EA		
04EC		
04EE		
04F0		
04F2		
04F4		
04F6		
04F8		
0500		
0502		
0504		
0506		
0508		
050A		
050C		
050E		
0531	0556	
10A0	10C5	
1E00		
1E02		
1E04		
1E06		
1E08		
1E0A		
1E0C		
1E0E		
1E10		
1E12		
1E14		
1E16		
1E18		
1E1A		
1E1C		
1E1E		
1E20		
1E22		
1E24		
1E26		
1E28		
1E2A		
1E2C		
1E2E		
1E30		
1E32		
1E34		
1E36		
1E38		
1E3A		
1E3C		
1E3E		
1E40		
1E42		
1E44		
1E46		
1E48		
1E4A		
1E4C		
1E4E		
1E50		
1E52		
1E54		
1E56		
1E58		
1E5A		
1E5C		
1E5E		
1E60		
1E62		
1E64		
1E66		
1E68		
1E6A		
1E6C		
1E6E		
1E70		
1E72		
1E74		
1E76		
1E78		
1E7A		
1E7C		
1E7E		
1E80		
1E82		
1E84		
1E86		
1E88		
1E8A		
1E8C		
1E8E		
1E90		
1E92		
1E94		
1EA0		
1EA2		
1EA4		
1EA6		
1EA8		
1EAA		
1EAC		
1EAE		
1EB0		
1EB2		
1EB4		
1EB6		
1EB8		
1EBA		
1EBC		
1EBE		
1EC0		
1EC2		
1EC4		
1EC6		
1EC8		
1ECA		
1ECC		
1ECE		
1ED0		
1ED2		
1ED4		
1ED6		
1ED8		
1EDA		
1EDC		
1EDE		
1EE0		
1EE2		
1EE4		
1EE6		
1EE8		
1EEA		
1EEC		
1EEE		
1EF0		
1EF2		
1EF4		
1EF6		
1EF8		
1F08	1F0F	
1F18	1F1D	
1F28	1F2F	
1F38	1F3F	
1F48	1F4D	
1F59		
1F5B		
1F5D		
1F5F		
1F68	1F6F	
1FB8	1FBB	
1FC8	1FCB	
1FD8	1FDB	
1FE8	1FEC	
1FF8	1FFB	
2102		
2107		
210B	210D	
2110	2112	
2115		
2119	211D	
2124		
2126		
2128		
212A	212D	
2130	2131	
2133		
213E	213F	
2145		
2C00	2C2E	
2C80		
2C82		
2C84		
2C86		
2C88		
2C8A		
2C8C		
2C8E		
2C90		
2C92		
2C94		
2C96		
2C98		
2C9A		
2C9C		
2C9E		
2CA0		
2CA2		
2CA4		
2CA6		
2CA8		
2CAA		
2CAC		
2CAE		
2CB0		
2CB2		
2CB4		
2CB6		
2CB8		
2CBA		
2CBC		
2CBE		
2CC0		
2CC2		
2CC4		
2CC6		
2CC8		
2CCA		
2CCC		
2CCE		
2CD0		
2CD2		
2CD4		
2CD6		
2CD8		
2CDA		
2CDC		
2CDE		
2CE0		
2CE2		
FF21	FF3A	
10400	10427	
1D400	1D419	
1D434	1D44D	
1D468	1D481	
1D49C		
1D49E	1D49F	
1D4A2		
1D4A5	1D4A6	
1D4A9	1D4AC	
1D4AE	1D4B5	
1D4D0	1D4E9	
1D504	1D505	
1D507	1D50A	
1D50D	1D514	
1D516	1D51C	
1D538	1D539	
1D53B	1D53E	
1D540	1D544	
1D546		
1D54A	1D550	
1D56C	1D585	
1D5A0	1D5B9	
1D5D4	1D5ED	
1D608	1D621	
1D63C	1D655	
1D670	1D689	
1D6A8	1D6C0	
1D6E2	1D6FA	
1D71C	1D734	
1D756	1D76E	
1D790	1D7A8	
END
};
1;

package unicore::lib::selfloader::Uppercas;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Uppercase} (and fuzzy permutations)
# 
# Meaning: [\p{Lu}\p{Other_Uppercase}]
#
return <<'END';
0041	005A	
00C0	00D6	
00D8	00DE	
0100		
0102		
0104		
0106		
0108		
010A		
010C		
010E		
0110		
0112		
0114		
0116		
0118		
011A		
011C		
011E		
0120		
0122		
0124		
0126		
0128		
012A		
012C		
012E		
0130		
0132		
0134		
0136		
0139		
013B		
013D		
013F		
0141		
0143		
0145		
0147		
014A		
014C		
014E		
0150		
0152		
0154		
0156		
0158		
015A		
015C		
015E		
0160		
0162		
0164		
0166		
0168		
016A		
016C		
016E		
0170		
0172		
0174		
0176		
0178	0179	
017B		
017D		
0181	0182	
0184		
0186	0187	
0189	018B	
018E	0191	
0193	0194	
0196	0198	
019C	019D	
019F	01A0	
01A2		
01A4		
01A6	01A7	
01A9		
01AC		
01AE	01AF	
01B1	01B3	
01B5		
01B7	01B8	
01BC		
01C4		
01C7		
01CA		
01CD		
01CF		
01D1		
01D3		
01D5		
01D7		
01D9		
01DB		
01DE		
01E0		
01E2		
01E4		
01E6		
01E8		
01EA		
01EC		
01EE		
01F1		
01F4		
01F6	01F8	
01FA		
01FC		
01FE		
0200		
0202		
0204		
0206		
0208		
020A		
020C		
020E		
0210		
0212		
0214		
0216		
0218		
021A		
021C		
021E		
0220		
0222		
0224		
0226		
0228		
022A		
022C		
022E		
0230		
0232		
023A	023B	
023D	023E	
0241		
0386		
0388	038A	
038C		
038E	038F	
0391	03A1	
03A3	03AB	
03D2	03D4	
03D8		
03DA		
03DC		
03DE		
03E0		
03E2		
03E4		
03E6		
03E8		
03EA		
03EC		
03EE		
03F4		
03F7		
03F9	03FA	
03FD	042F	
0460		
0462		
0464		
0466		
0468		
046A		
046C		
046E		
0470		
0472		
0474		
0476		
0478		
047A		
047C		
047E		
0480		
048A		
048C		
048E		
0490		
0492		
0494		
0496		
0498		
049A		
049C		
049E		
04A0		
04A2		
04A4		
04A6		
04A8		
04AA		
04AC		
04AE		
04B0		
04B2		
04B4		
04B6		
04B8		
04BA		
04BC		
04BE		
04C0	04C1	
04C3		
04C5		
04C7		
04C9		
04CB		
04CD		
04D0		
04D2		
04D4		
04D6		
04D8		
04DA		
04DC		
04DE		
04E0		
04E2		
04E4		
04E6		
04E8		
04EA		
04EC		
04EE		
04F0		
04F2		
04F4		
04F6		
04F8		
0500		
0502		
0504		
0506		
0508		
050A		
050C		
050E		
0531	0556	
10A0	10C5	
1E00		
1E02		
1E04		
1E06		
1E08		
1E0A		
1E0C		
1E0E		
1E10		
1E12		
1E14		
1E16		
1E18		
1E1A		
1E1C		
1E1E		
1E20		
1E22		
1E24		
1E26		
1E28		
1E2A		
1E2C		
1E2E		
1E30		
1E32		
1E34		
1E36		
1E38		
1E3A		
1E3C		
1E3E		
1E40		
1E42		
1E44		
1E46		
1E48		
1E4A		
1E4C		
1E4E		
1E50		
1E52		
1E54		
1E56		
1E58		
1E5A		
1E5C		
1E5E		
1E60		
1E62		
1E64		
1E66		
1E68		
1E6A		
1E6C		
1E6E		
1E70		
1E72		
1E74		
1E76		
1E78		
1E7A		
1E7C		
1E7E		
1E80		
1E82		
1E84		
1E86		
1E88		
1E8A		
1E8C		
1E8E		
1E90		
1E92		
1E94		
1EA0		
1EA2		
1EA4		
1EA6		
1EA8		
1EAA		
1EAC		
1EAE		
1EB0		
1EB2		
1EB4		
1EB6		
1EB8		
1EBA		
1EBC		
1EBE		
1EC0		
1EC2		
1EC4		
1EC6		
1EC8		
1ECA		
1ECC		
1ECE		
1ED0		
1ED2		
1ED4		
1ED6		
1ED8		
1EDA		
1EDC		
1EDE		
1EE0		
1EE2		
1EE4		
1EE6		
1EE8		
1EEA		
1EEC		
1EEE		
1EF0		
1EF2		
1EF4		
1EF6		
1EF8		
1F08	1F0F	
1F18	1F1D	
1F28	1F2F	
1F38	1F3F	
1F48	1F4D	
1F59		
1F5B		
1F5D		
1F5F		
1F68	1F6F	
1FB8	1FBB	
1FC8	1FCB	
1FD8	1FDB	
1FE8	1FEC	
1FF8	1FFB	
2102		
2107		
210B	210D	
2110	2112	
2115		
2119	211D	
2124		
2126		
2128		
212A	212D	
2130	2131	
2133		
213E	213F	
2145		
2160	216F	
24B6	24CF	
2C00	2C2E	
2C80		
2C82		
2C84		
2C86		
2C88		
2C8A		
2C8C		
2C8E		
2C90		
2C92		
2C94		
2C96		
2C98		
2C9A		
2C9C		
2C9E		
2CA0		
2CA2		
2CA4		
2CA6		
2CA8		
2CAA		
2CAC		
2CAE		
2CB0		
2CB2		
2CB4		
2CB6		
2CB8		
2CBA		
2CBC		
2CBE		
2CC0		
2CC2		
2CC4		
2CC6		
2CC8		
2CCA		
2CCC		
2CCE		
2CD0		
2CD2		
2CD4		
2CD6		
2CD8		
2CDA		
2CDC		
2CDE		
2CE0		
2CE2		
FF21	FF3A	
10400	10427	
1D400	1D419	
1D434	1D44D	
1D468	1D481	
1D49C		
1D49E	1D49F	
1D4A2		
1D4A5	1D4A6	
1D4A9	1D4AC	
1D4AE	1D4B5	
1D4D0	1D4E9	
1D504	1D505	
1D507	1D50A	
1D50D	1D514	
1D516	1D51C	
1D538	1D539	
1D53B	1D53E	
1D540	1D544	
1D546		
1D54A	1D550	
1D56C	1D585	
1D5A0	1D5B9	
1D5D4	1D5ED	
1D608	1D621	
1D63C	1D655	
1D670	1D689	
1D6A8	1D6C0	
1D6E2	1D6FA	
1D71C	1D734	
1D756	1D76E	
1D790	1D7A8	
END
};
1;

package unicore::lib::selfloader::Variatio;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{VariationSelector} (and fuzzy permutations)
# 
# Meaning: Extended property 'Variation_Selector'
#
return <<'END';
180B	180D	Variation_Selector
FE00	FE0F	Variation_Selector
E0100	E01EF	Variation_Selector
END
};
1;

package unicore::lib::selfloader::VS;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'Variation_Selector'
#
return <<'END';
180B	180D	Variation_Selector
FE00	FE0F	Variation_Selector
E0100	E01EF	Variation_Selector
END
};
1;

package unicore::lib::selfloader::WhiteSpa;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{WhiteSpace} (and fuzzy permutations)
# 
# Meaning: Extended property 'White_Space'
#
return <<'END';
0009	000D	White_Space
0020		White_Space
0085		White_Space
00A0		White_Space
1680		White_Space
180E		White_Space
2000	200A	White_Space
2028	2029	White_Space
202F		White_Space
205F		White_Space
3000		White_Space
END
};
1;

package unicore::lib::selfloader::Word;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Word}
# 
# Meaning: [[:Word:]]
#
return <<'END';
0030	0039	
0041	005A	
005F		
0061	007A	
00AA		
00B2	00B3	
00B5		
00B9	00BA	
00BC	00BE	
00C0	00D6	
00D8	00F6	
00F8	0241	
0250	02C1	
02C6	02D1	
02E0	02E4	
02EE		
0300	036F	
037A		
0386		
0388	038A	
038C		
038E	03A1	
03A3	03CE	
03D0	03F5	
03F7	0481	
0483	0486	
0488	04CE	
04D0	04F9	
0500	050F	
0531	0556	
0559		
0561	0587	
0591	05B9	
05BB	05BD	
05BF		
05C1	05C2	
05C4	05C5	
05C7		
05D0	05EA	
05F0	05F2	
0610	0615	
0621	063A	
0640	065E	
0660	0669	
066E	06D3	
06D5	06DC	
06DE	06E8	
06EA	06FC	
06FF		
0710	074A	
074D	076D	
0780	07B1	
0901	0939	
093C	094D	
0950	0954	
0958	0963	
0966	096F	
097D		
0981	0983	
0985	098C	
098F	0990	
0993	09A8	
09AA	09B0	
09B2		
09B6	09B9	
09BC	09C4	
09C7	09C8	
09CB	09CE	
09D7		
09DC	09DD	
09DF	09E3	
09E6	09F1	
09F4	09F9	
0A01	0A03	
0A05	0A0A	
0A0F	0A10	
0A13	0A28	
0A2A	0A30	
0A32	0A33	
0A35	0A36	
0A38	0A39	
0A3C		
0A3E	0A42	
0A47	0A48	
0A4B	0A4D	
0A59	0A5C	
0A5E		
0A66	0A74	
0A81	0A83	
0A85	0A8D	
0A8F	0A91	
0A93	0AA8	
0AAA	0AB0	
0AB2	0AB3	
0AB5	0AB9	
0ABC	0AC5	
0AC7	0AC9	
0ACB	0ACD	
0AD0		
0AE0	0AE3	
0AE6	0AEF	
0B01	0B03	
0B05	0B0C	
0B0F	0B10	
0B13	0B28	
0B2A	0B30	
0B32	0B33	
0B35	0B39	
0B3C	0B43	
0B47	0B48	
0B4B	0B4D	
0B56	0B57	
0B5C	0B5D	
0B5F	0B61	
0B66	0B6F	
0B71		
0B82	0B83	
0B85	0B8A	
0B8E	0B90	
0B92	0B95	
0B99	0B9A	
0B9C		
0B9E	0B9F	
0BA3	0BA4	
0BA8	0BAA	
0BAE	0BB9	
0BBE	0BC2	
0BC6	0BC8	
0BCA	0BCD	
0BD7		
0BE6	0BF2	
0C01	0C03	
0C05	0C0C	
0C0E	0C10	
0C12	0C28	
0C2A	0C33	
0C35	0C39	
0C3E	0C44	
0C46	0C48	
0C4A	0C4D	
0C55	0C56	
0C60	0C61	
0C66	0C6F	
0C82	0C83	
0C85	0C8C	
0C8E	0C90	
0C92	0CA8	
0CAA	0CB3	
0CB5	0CB9	
0CBC	0CC4	
0CC6	0CC8	
0CCA	0CCD	
0CD5	0CD6	
0CDE		
0CE0	0CE1	
0CE6	0CEF	
0D02	0D03	
0D05	0D0C	
0D0E	0D10	
0D12	0D28	
0D2A	0D39	
0D3E	0D43	
0D46	0D48	
0D4A	0D4D	
0D57		
0D60	0D61	
0D66	0D6F	
0D82	0D83	
0D85	0D96	
0D9A	0DB1	
0DB3	0DBB	
0DBD		
0DC0	0DC6	
0DCA		
0DCF	0DD4	
0DD6		
0DD8	0DDF	
0DF2	0DF3	
0E01	0E3A	
0E40	0E4E	
0E50	0E59	
0E81	0E82	
0E84		
0E87	0E88	
0E8A		
0E8D		
0E94	0E97	
0E99	0E9F	
0EA1	0EA3	
0EA5		
0EA7		
0EAA	0EAB	
0EAD	0EB9	
0EBB	0EBD	
0EC0	0EC4	
0EC6		
0EC8	0ECD	
0ED0	0ED9	
0EDC	0EDD	
0F00		
0F18	0F19	
0F20	0F33	
0F35		
0F37		
0F39		
0F3E	0F47	
0F49	0F6A	
0F71	0F84	
0F86	0F8B	
0F90	0F97	
0F99	0FBC	
0FC6		
1000	1021	
1023	1027	
1029	102A	
102C	1032	
1036	1039	
1040	1049	
1050	1059	
10A0	10C5	
10D0	10FA	
10FC		
1100	1159	
115F	11A2	
11A8	11F9	
1200	1248	
124A	124D	
1250	1256	
1258		
125A	125D	
1260	1288	
128A	128D	
1290	12B0	
12B2	12B5	
12B8	12BE	
12C0		
12C2	12C5	
12C8	12D6	
12D8	1310	
1312	1315	
1318	135A	
135F		
1369	137C	
1380	138F	
13A0	13F4	
1401	166C	
166F	1676	
1681	169A	
16A0	16EA	
16EE	16F0	
1700	170C	
170E	1714	
1720	1734	
1740	1753	
1760	176C	
176E	1770	
1772	1773	
1780	17B3	
17B6	17D3	
17D7		
17DC	17DD	
17E0	17E9	
17F0	17F9	
180B	180D	
1810	1819	
1820	1877	
1880	18A9	
1900	191C	
1920	192B	
1930	193B	
1946	196D	
1970	1974	
1980	19A9	
19B0	19C9	
19D0	19D9	
1A00	1A1B	
1D00	1DC3	
1E00	1E9B	
1EA0	1EF9	
1F00	1F15	
1F18	1F1D	
1F20	1F45	
1F48	1F4D	
1F50	1F57	
1F59		
1F5B		
1F5D		
1F5F	1F7D	
1F80	1FB4	
1FB6	1FBC	
1FBE		
1FC2	1FC4	
1FC6	1FCC	
1FD0	1FD3	
1FD6	1FDB	
1FE0	1FEC	
1FF2	1FF4	
1FF6	1FFC	
203F	2040	
2054		
2070	2071	
2074	2079	
207F	2089	
2090	2094	
20D0	20EB	
2102		
2107		
210A	2113	
2115		
2119	211D	
2124		
2126		
2128		
212A	212D	
212F	2131	
2133	2139	
213C	213F	
2145	2149	
2153	2183	
2460	249B	
24EA	24FF	
2776	2793	
2C00	2C2E	
2C30	2C5E	
2C80	2CE4	
2CFD		
2D00	2D25	
2D30	2D65	
2D6F		
2D80	2D96	
2DA0	2DA6	
2DA8	2DAE	
2DB0	2DB6	
2DB8	2DBE	
2DC0	2DC6	
2DC8	2DCE	
2DD0	2DD6	
2DD8	2DDE	
3005	3007	
3021	302F	
3031	3035	
3038	303C	
3041	3096	
3099	309A	
309D	309F	
30A1	30FA	
30FC	30FF	
3105	312C	
3131	318E	
3192	3195	
31A0	31B7	
31F0	31FF	
3220	3229	
3251	325F	
3280	3289	
32B1	32BF	
3400	4DB5	
4E00	9FBB	
A000	A48C	
A800	A827	
AC00	D7A3	
F900	FA2D	
FA30	FA6A	
FA70	FAD9	
FB00	FB06	
FB13	FB17	
FB1D	FB28	
FB2A	FB36	
FB38	FB3C	
FB3E		
FB40	FB41	
FB43	FB44	
FB46	FBB1	
FBD3	FD3D	
FD50	FD8F	
FD92	FDC7	
FDF0	FDFB	
FE00	FE0F	
FE20	FE23	
FE33	FE34	
FE4D	FE4F	
FE70	FE74	
FE76	FEFC	
FF10	FF19	
FF21	FF3A	
FF3F		
FF41	FF5A	
FF66	FFBE	
FFC2	FFC7	
FFCA	FFCF	
FFD2	FFD7	
FFDA	FFDC	
10000	1000B	
1000D	10026	
10028	1003A	
1003C	1003D	
1003F	1004D	
10050	1005D	
10080	100FA	
10107	10133	
10140	10178	
1018A		
10300	1031E	
10320	10323	
10330	1034A	
10380	1039D	
103A0	103C3	
103C8	103CF	
103D1	103D5	
10400	1049D	
104A0	104A9	
10800	10805	
10808		
1080A	10835	
10837	10838	
1083C		
1083F		
10A00	10A03	
10A05	10A06	
10A0C	10A13	
10A15	10A17	
10A19	10A33	
10A38	10A3A	
10A3F	10A47	
1D165	1D169	
1D16D	1D172	
1D17B	1D182	
1D185	1D18B	
1D1AA	1D1AD	
1D242	1D244	
1D400	1D454	
1D456	1D49C	
1D49E	1D49F	
1D4A2		
1D4A5	1D4A6	
1D4A9	1D4AC	
1D4AE	1D4B9	
1D4BB		
1D4BD	1D4C3	
1D4C5	1D505	
1D507	1D50A	
1D50D	1D514	
1D516	1D51C	
1D51E	1D539	
1D53B	1D53E	
1D540	1D544	
1D546		
1D54A	1D550	
1D552	1D6A5	
1D6A8	1D6C0	
1D6C2	1D6DA	
1D6DC	1D6FA	
1D6FC	1D714	
1D716	1D734	
1D736	1D74E	
1D750	1D76E	
1D770	1D788	
1D78A	1D7A8	
1D7AA	1D7C2	
1D7C4	1D7C9	
1D7CE	1D7FF	
20000	2A6D6	
2F800	2FA1D	
E0100	E01EF	
END
};
1;

package unicore::lib::selfloader::WSpace;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# Binary property 'White_Space'
#
return <<'END';
0009	000D	White_Space
0020		White_Space
0085		White_Space
00A0		White_Space
1680		White_Space
180E		White_Space
2000	200A	White_Space
2028	2029	White_Space
202F		White_Space
205F		White_Space
3000		White_Space
END
};
1;

package unicore::lib::selfloader::XDigit;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{XDigit}
# 
# Meaning: [[:XDigit:]]
#
return <<'END';
0030	0039	
0041	0046	
0061	0066	
END
};
1;

package unicore::lib::selfloader::Yiii;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Yi} (and fuzzy permutations)
# 
# Meaning: Script 'Yi'
#
return <<'END';
A000	A48C	Yi
A490	A4C6	Yi
END
};
1;

package unicore::lib::selfloader::Z;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Z}
# 	\p{Z} (and fuzzy permutations)
# 
# Meaning: Major Category 'Z'
#
return <<'END';
0020		
00A0		
1680		
180E		
2000	200A	
2028	2029	
202F		
205F		
3000		
END
};
1;

package unicore::lib::selfloader::Zl;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Zl}
# 	\p{Zl} (and fuzzy permutations)
# 
# Meaning: General Category 'Zl'
#
return <<'END';
2028		
END
};
1;

package unicore::lib::selfloader::Zp;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Zp}
# 	\p{Zp} (and fuzzy permutations)
# 
# Meaning: General Category 'Zp'
#
return <<'END';
2029		
END
};
1;

package unicore::lib::selfloader::Zs;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Zs}
# 	\p{Zs} (and fuzzy permutations)
# 
# Meaning: General Category 'Zs'
#
return <<'END';
0020		
00A0		
1680		
180E		
2000	200A	
202F		
205F		
3000		
END
};
1;

package unicore::lib::selfloader::Zyyy;
sub getstrings() {
# !!!!!!!   DO NOT EDIT THIS FILE   !!!!!!! 
# This file is built by mktables from e.g. UnicodeData.txt.
# Any changes made here will be lost!

#
# This file supports:
# 	\p{Common} (and fuzzy permutations)
# 
# Meaning: Script 'Common'
#
return <<'END';
0000	0040	Common
005B	0060	Common
007B	00A9	Common
00AB	00B9	Common
00BB	00BF	Common
00D7		Common
00F7		Common
02B9	02DF	Common
02E5	02FF	Common
037E		Common
0387		Common
0589		Common
0600	0603	Common
060C		Common
061B		Common
061F		Common
0640		Common
0660	0669	Common
06DD		Common
0964	0965	Common
0970		Common
0E3F		Common
10FB		Common
16EB	16ED	Common
1735	1736	Common
2000	200B	Common
200E	2063	Common
206A	2070	Common
2074	207E	Common
2080	208E	Common
20A0	20B5	Common
2100	2125	Common
2127	2129	Common
212C	214C	Common
2153	2183	Common
2190	23DB	Common
2400	2426	Common
2440	244A	Common
2460	269C	Common
26A0	26B1	Common
2701	2704	Common
2706	2709	Common
270C	2727	Common
2729	274B	Common
274D		Common
274F	2752	Common
2756		Common
2758	275E	Common
2761	2794	Common
2798	27AF	Common
27B1	27BE	Common
27C0	27C6	Common
27D0	27EB	Common
27F0	27FF	Common
2900	2B13	Common
2E00	2E17	Common
2E1C	2E1D	Common
2FF0	2FFB	Common
3000	3004	Common
3006		Common
3008	3020	Common
3030	3037	Common
303C	303F	Common
309B	309C	Common
30A0		Common
30FB	30FC	Common
3190	319F	Common
31C0	31CF	Common
3220	3243	Common
3250	325F	Common
327E	32FE	Common
3300	33FF	Common
4DC0	4DFF	Common
A700	A716	Common
E000	F8FF	Common
FD3E	FD3F	Common
FDFD		Common
FE10	FE19	Common
FE30	FE52	Common
FE54	FE66	Common
FE68	FE6B	Common
FEFF		Common
FF01	FF20	Common
FF3B	FF40	Common
FF5B	FF65	Common
FF70		Common
FF9E	FF9F	Common
FFE0	FFE6	Common
FFE8	FFEE	Common
FFF9	FFFD	Common
10100	10102	Common
10107	10133	Common
10137	1013F	Common
1D000	1D0F5	Common
1D100	1D126	Common
1D12A	1D166	Common
1D16A	1D17A	Common
1D183	1D184	Common
1D18C	1D1A9	Common
1D1AE	1D1DD	Common
1D300	1D356	Common
1D400	1D454	Common
1D456	1D49C	Common
1D49E	1D49F	Common
1D4A2		Common
1D4A5	1D4A6	Common
1D4A9	1D4AC	Common
1D4AE	1D4B9	Common
1D4BB		Common
1D4BD	1D4C3	Common
1D4C5	1D505	Common
1D507	1D50A	Common
1D50D	1D514	Common
1D516	1D51C	Common
1D51E	1D539	Common
1D53B	1D53E	Common
1D540	1D544	Common
1D546		Common
1D54A	1D550	Common
1D552	1D6A5	Common
1D6A8	1D7C9	Common
1D7CE	1D7FF	Common
E0001		Common
E0020	E007F	Common
F0000	FFFFD	Common
100000	10FFFD	Common
END
};
1;

