# $Id: Lookup.pm 3688 2006-08-31 11:27:11Z kankri $

use strict;
use warnings;

package FS::PhishBL::Lookup;

use FS::Logger;
our $FS_Logger_facility = "phlookup";

use Carp;
use LWP::UserAgent;
use MIME::Base64 qw(decode_base64);

use vars qw($VERSION);

$VERSION = qw/$Revision: 3688 $/[1];

=head1 NAME

FS::PhishBL::Lookup - Provides interface to the network lookup protocol

=head1 SYNOPSIS

use FS::PhishBL::Lookup;

=cut

=item new ( CONFIGHASHREF )

Construct FS::PhishBL::Lookup object

=cut

my $prev_proxy = ""; # just for logging

sub new {
	my $class = @_ ? shift : 'FS::PhishBL::Lookup';
	croak "Usage: new $class ( CONFIGHASHREF )\n"
	    unless @_ == 1;
	my ($config) = @_;

	debug(">>FS::PhishBL::Lookup::new()");

	my $self = {};
	$self->{config} = $config;

	# create user agent
	$self->{ua} = new LWP::UserAgent(keep_alive => 5, timeout => 5);

	# set user agent string to use when communicating with server
	$self->{ua}->agent("F-SecurePhishURLLookup/$VERSION");
	
	bless $self, $class;

	$self->refresh_settings($config);

	debug("<<FS::PhishBL::Lookup::new()");
	return $self;
}

=item refresh_settings ( CONFIGHASHREF )

Refresh FS::PhishBL::Lookup object after config settings have been changed

=cut

sub refresh_settings {
	my ($self, $config) = @_;

	return unless $config->{proxy};

	$self->{ua}->proxy(['http', 'ftp'], $config->{'proxy'});

	info("using HTTP proxy $config->{proxy}")
	    if $prev_proxy ne $config->{proxy};

	$prev_proxy = $config->{proxy};
}


=item query ( URLSTRING )

Queries the lookup server and returns an arrayref to entries of the form

  {regex => $regex, classes => \%classes}

In case of query errors, dies with an HTTP::Response object.
Catch with eval{}.

=cut

sub query {
    my ($self, $url) = @_;

    debug("query($url)");
 
    my ($reply, $response) = $self->get($self->{config}->{query_url} . "$url");

    die $response if defined $response && $response->code != 404;

    return [] unless defined $reply;

    my @patterns = ();

    for (split /\n/, $reply) {
	chomp;
	s/\015$//;

	next unless /^re\t/;
	my ($regex, %classes);
	
	# we have tab separated elements: "re" regex key value key value...
	unless (((undef, $regex, %classes) = split(/\t/)) % 2 == 0) {
	    error("incomplete key/value pair in '$_'");
	}

	# convert class names canonically to lowercase
	my $i = 0;
	%classes = map { $i++ % 2 ? $_ : lc($_) } %classes;	

	push(@patterns, {regex => $regex, classes => \%classes});
    }

    return \@patterns;
}

sub get {
    my ($self, $url) = @_;

    debug("get: query URL: $url");
    my $request = HTTP::Request->new(GET => $url);
    my $response = $self->{ua}->request($request);
    debug("get: response " . $response->status_line());
    return $response->is_success ?
	($response->content, undef) : (undef, $response);
}

=head1 COPYRIGHT

Copyright (C) 2006 F-Secure Corp, all rights reserved

=cut

1;
